//###################################################################################
// ~ International Laser Display Association ~
// ~     ILDA File Format Driver in C++      ~
// ~             Built for EZFB              ~
//
// by James Lehman
// Extra Stimulus Inc.
// james@akrobiz.com
//
// began: October 2003
//
//###################################################################################
#ifndef __EZ_3D_SHORT_DEFINITIONS__
#define __EZ_3D_SHORT_DEFINITIONS__

//###################################################################################
#include "ez_ilda_includes.hpp"

//###################################################################################
class ez_3D_short
{
public:
    ez_3D_short( const short& _x = 0,
                 const short& _y = 0,
                 const short& _z = 0
               )
               : x(_x), y(_y), z(_z)                    {};
    //-------------------------------------------------------------------------------
    ez_3D_short(const ez_3D_short& p)
               : x(p.x), y(p.y), z(p.z)                 {};
    //-------------------------------------------------------------------------------
    ez_3D_short(const int& s)
               : x((short)s), y((short)s), z((short)s)  {};
    //-------------------------------------------------------------------------------
    inline ez_3D_short& operator = (const ez_3D_short& p)
                                {
                                    x = p.x;
                                    y = p.y;
                                    z = p.z;
                                    return *this;
                                };
    //-------------------------------------------------------------------------------
    inline bool operator == (const ez_3D_short& p) const
                        {   
                            if(    (x != p.x)
                                || (y != p.y)
                                || (z != p.z)
                              )
                                return false;
                            return true;
                        };
    //-------------------------------------------------------------------------------
    inline bool operator != (const ez_3D_short& p) const
                        {   
                            if(    (x == p.x)
                                && (y == p.y)
                                && (z == p.z)
                              )
                                return false;
                            return true;
                        };
    //-------------------------------------------------------------------------------
    inline bool operator > (const ez_3D_short& p) const // distance from the origin
                        {   
                            return(   ((float)  x *   x + (float)  y *   y + (float)  z *   z)
                                    > ((float)p.x * p.x + (float)p.y * p.y + (float)p.z * p.z)
                                  );
                        };
    //-------------------------------------------------------------------------------
    inline bool operator >= (const ez_3D_short& p) const // distance from the origin
                        {   
                            return(    ((float)  x *   x + (float)  y *   y + (float)  z *   z)
                                    >= ((float)p.x * p.x + (float)p.y * p.y + (float)p.z * p.z)
                                  );
                        };
    //-------------------------------------------------------------------------------
    inline bool operator < (const ez_3D_short& p) const // distance from the origin
                        {   
                            return(   ((float)  x *   x + (float)  y *   y + (float)  z *   z)
                                    < ((float)p.x * p.x + (float)p.y * p.y + (float)p.z * p.z)
                                  );
                        };
    //-------------------------------------------------------------------------------
    inline bool operator <= (const ez_3D_short& p) const // distance from the origin
                        {   
                            return(    ((float)  x *   x + (float)  y *   y + (float)  z *   z)
                                    <= ((float)p.x * p.x + (float)p.y * p.y + (float)p.z * p.z)
                                  );
                        };
    //-------------------------------------------------------------------------------
    inline ez_3D_short  operator -  (const ez_3D_short& p)
                                {   
                                    return ez_3D_short(x - p.x, y - p.y, z - p.z);
                                };
    //-------------------------------------------------------------------------------
    inline ez_3D_short& operator -= (const ez_3D_short& p)
                                {   
                                    x -= p.x;
                                    y -= p.y;
                                    z -= p.z;
                                    return *this;
                                };
    //-------------------------------------------------------------------------------
    inline ez_3D_short  operator +  (const ez_3D_short& p) const
                                {   
                                    return ez_3D_short(x + p.x, y + p.y, z + p.z);
                                };
    //-------------------------------------------------------------------------------
    inline ez_3D_short& operator += (const ez_3D_short& p)
                                {   
                                    x += p.x;
                                    y += p.y;
                                    z += p.z;
                                    return *this;
                                };
    //-------------------------------------------------------------------------------
    inline double operator | (const ez_3D_short& p) const // distance from each other
                            {   
                                double dx = abs(x - p.x),
                                       dy = abs(y - p.y),
                                       dz = abs(z - p.z);
                                return sqrt(dx*dx + dy*dy + dz*dz);
                            };
    //-------------------------------------------------------------------------------
    inline ez_3D_short  operator - ()
                            {
                                ez_3D_short negative;
                                negative.x = -x;
                                negative.y = -y;
                                negative.z = -z;
                                return negative;
                            };
    //-------------------------------------------------------------------------------
    inline ez_3D_short    abs_3D() {    return ez_3D_short(abs(x), abs(y), abs(z));    };
    //-------------------------------------------------------------------------------
    inline int            sum()    {    return (int)x + (int)y + (int)z;    };
    //-------------------------------------------------------------------------------
    short x, y, z;
};

//###################################################################################
#endif

//###################################################################################
/////////////////////////////////////////////////////////////////////////////////////
//###################################################################################


