//############################################################################
// ~ International Laser Display Association ~
// ~     ILDA file format driver in C++      ~
// ~             built for EZFB              ~
//
// by James Lehman
// Extra Stimulus Inc.
// james@akrobiz.com
//
// began: October 2003
//
//############################################################################
#include "ez_ilda.hpp"

//############################################################################
ez_ilda*     ez_ilda_pointer = NULL; // the one and only!
SDL_Surface* screen          = NULL; // an SDL graphics window
SDL_Event    event                 ;

//############################################################################
ez_ilda::ez_ilda(u_int xres, u_int yres)
          : ez_ilda_frame_set       (          ),
            ez_palette_set          (          ),
            bmp                     (          ),
            background              (          ),
            wave_sample_rate        (48000     ), // Hz
            lit_dwell_overhang      (2         ),
            font_size_factor        (1         ),
            move_points_per_key     (100       ), // pixels
            output_bmp_size         (1000      ), // pixels square
            view_angle              (          ),
            view_scale              (1.0       ),
            view_offset             (          ),
            distance_delta_max      (600       ), // points
            longest_dwell_microsec  (333       ),
            insignificant_distance  (60        ), // points
            zero_angle_tolerance    (one_degree), // radians
            recolor_span_factor     (2.0       ),
            rendered_arc_angle      (6.0       ), // degrees
            frames_per_second       (16        ), // 0 == every frame is shown exactly once
            still_frame_duration    (10.0      ), // seconds
            show_points             (false     ),
            show_blanking           (false     ),
            show_cursors            (false     ),
            show_palettes           (true      ),
            show_stats              (true      ),
            show_fixed_bounds       (true      ),
            show_fixed_origin       (false     ),
            show_floating_axis      (false     ),
            show_floating_bounds    (false     ),
            show_ui_message         (true      ),
            show_arbitrary_point    (false     ),
            show_effects_generation (true      ),
            auto_scale_dxf          (true      ),
            maintain_dxf_origin     (false     ),
            loop_animation          (true      ),
            approximate_frame_rate  (true      ),
            black_dwell_points      (true      ),
            auto_flatten_z          (true      ),
            auto_optimize           (true      )
{
    if(ez_ilda_pointer == NULL)
    {
        ez_ilda_pointer = this;

        if(SDL_Init(SDL_INIT_VIDEO) < 0)
            SDL_bailout(2);
        if((screen = SDL_SetVideoMode( xres,
                                       yres,
                                       0,
                                         SDL_HWSURFACE
                                       //| SDL_FULLSCREEN
                                       //| SDL_NOFRAME
                                       | SDL_ANYFORMAT
                                     )
           ) == NULL
          )
            SDL_bailout(2);
            
        SDL_EnableUNICODE(1);
        //SDL_EnableKeyRepeat(SDL_DEFAULT_REPEAT_INTERVAL, SDL_DEFAULT_REPEAT_DELAY);
        SDL_EnableKeyRepeat(250, 1);
        
        font_size_factor = screen->h / 480;
        if(font_size_factor == 0)
            font_size_factor = 1;

        SDL_EventState(SDL_MOUSEMOTION, SDL_IGNORE);
        SDL_EventState(SDL_PRESSED    , SDL_IGNORE);
        SDL_EventState(SDL_RELEASED   , SDL_IGNORE);
        SDL_EventState(SDL_SYSWMEVENT , SDL_IGNORE);

        SDL_WM_SetCaption("LaserBoy Multiplatform by James Lehman : Extra Stimulus Inc.", NULL);
        
        if(SDL_LockSurface(screen) < 0)
            SDL_bailout(2);
        
        bmp_clone_SDL_Surface(&bmp, screen, 0);
        bmp_copy(&background, &bmp);
        bmp_to_SDL_Surface(&bmp, screen);

        load_effects();
    }
}

//############################################################################
ez_ilda::~ez_ilda()
{
    SDL_Quit     ();
    bmp_free     (&bmp);
    bmp_free     (&background);
    ez_ilda_pointer = NULL;
}

//############################################################################
void ez_ilda::sync_rgb_and_palette()
{
    int             i                         = 0           ,
                    j                                       ,
                    super_palette_first_frame = 0           ,
                    super_palette_last_frame  = 0           ,
                    current_frame_index       = frame_index ;
    ez_palette_set  reduced_palette_set                     ;
    ez_palette      super_palette                           ;
    char            super_palette_name[7]                   ;
    static int      super_palette_number      = 0           ;
    //------------------------------------------------------------------------
    for(j = 0; j < number_of_frames(); j++)
        frame(j).sync_rgb_and_palette();
    //------------------------------------------------------------------------
    while(i < number_of_frames())
    {
        if(frame_picker(i).palette_index != EZ_TRUE_COLOR)
        {
            //--------------------------------------------------------------------
            super_palette_first_frame = i;
            super_palette = palette_picker(frame_picker(i).palette_index);
            sprintf(super_palette_name, "%06d", super_palette_number);
            super_palette.name = string("SP") + super_palette_name;
            super_palette_number++;
            //--------------------------------------------------------------------
            while(    i < number_of_frames()
                   && super_palette.unite(palette_picker(frame_picker(i).palette_index))
                 )
            i++;
            //--------------------------------------------------------------------
            reduced_palette_set.push_back(super_palette);
            super_palette_last_frame  = i - 1;
            //--------------------------------------------------------------------
            for(j = super_palette_first_frame; j <= super_palette_last_frame; j++)
            {
                frame(j).find_rgb_in_palette(super_palette);
                frame(j).palette_index = reduced_palette_set.number_of_palettes() - 1;
            }
            //--------------------------------------------------------------------
#if EZ_ILDA_DEBUG_MESSAGES
            cout << "super_palette_first_frame : "
                 << super_palette_first_frame
                 << " | super_palette_last_frame  : "
                 << super_palette_last_frame
                 << endl;
#endif
        }
        else
            i++;
    }
    //------------------------------------------------------------------------
    frame(current_frame_index);
    replace_palette_set(reduced_palette_set);
    palette_index = current_frame().palette_index;
    return;
}

//############################################################################
void ez_ilda::display(string ui_message)
{
    if(palette_index != current_frame().palette_index)
    {
        palette_index = current_frame().palette_index;
        if(current_frame().palette_index != EZ_TRUE_COLOR)
            cout << "palette switched to " << palette_picker(palette_index).name << endl;
        else
            cout << "palette switched to TRUE_COLOR" << endl;
    }
    //------------------------------------------------------------------------
    current_frame().render_in_ezfb_bitmap(ui_message);
    bmp_to_SDL_Surface(&bmp, screen);
    return;
}

//############################################################################
string ez_ilda::display_prompt_and_echo_name(string prompt)
{
    bool   done = false;

    int    x_off  = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off  = BMP.yres - 24 * font_size_factor,
           x_gap  = 8 * font_size_factor,
           y_line = 8 * font_size_factor;

    string input;

    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.black
                , 0
                , font_size_factor
                , prompt.c_str()
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.make_rgb(0,255,255)
                , BMP.black
                , 0
                , font_size_factor
                , "_"
              );
    bmp_to_SDL_Surface(&bmp, screen);
      
              
    while(!done)
    {
        do
        {
            SDL_WaitEvent(&event);
        } while(    event.type != SDL_KEYDOWN
                 && event.type != SDL_QUIT
               );               
        if(event.type == SDL_QUIT)
            return "";
        if(event.key.keysym.sym == SDLK_ESCAPE)
            return "";
        if(event.key.keysym.sym == SDLK_RETURN)
            done = true;
        else
        {
            switch(event.key.keysym.sym)
            {
                //-------------------------------------------------    
                case SDLK_LEFT:
                case SDLK_DELETE:
                case SDLK_BACKSPACE:
                    if(input.size())
                    {
                        input = input.substr(0, input.size() - 1);
                        x_off -= x_gap;
                        bmp_printf(   &BMP
                                    , x_off
                                    , y_off
                                    , BMP.make_rgb(0,255,255)
                                    , BMP.black
                                    , 0
                                    , font_size_factor
                                    , " "
                                  );
                        bmp_to_SDL_Surface(&bmp, screen);
                    }                         
                    break;
                //-------------------------------------------------    
                default:
                    if(    isalnum(event.key.keysym.sym)
                        || event.key.keysym.unicode == '_'
                      )
                    {
                        bmp_printf(   &BMP
                                    , x_off
                                    , y_off
                                    , BMP.make_rgb(0,255,255)
                                    , BMP.black
                                    , 0
                                    , font_size_factor
                                    , "%c"
                                    , event.key.keysym.unicode
                                  );
                        bmp_to_SDL_Surface(&bmp, screen);
                        input += event.key.keysym.unicode;
                        x_off += x_gap;
                    }                            
                    break;                         
                //-------------------------------------------------    
            } // end switch(event.key.keysym.sym)
        } // end else of if(event.key.keysym.sym == SDLK_RETURN)
    } // end while(!done)    
    bmp_printf(   &BMP
                , x_off - x_gap * input.size()
                , y_off
                , BMP.make_rgb(255,0,0)
                , BMP.black
                , 0
                , font_size_factor
                , input.c_str()
              );
    bmp_to_SDL_Surface(&bmp, screen);
    return input;
}

//############################################################################
string ez_ilda::display_prompt_and_echo_string(string prompt)
{
    bool   done = false;

    int    x_off  = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off  = BMP.yres - 24 * font_size_factor,
           x_gap  = 8 * font_size_factor,
           y_line = 8 * font_size_factor;

    string input;

    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.black
                , 0
                , font_size_factor
                , prompt.c_str()
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.make_rgb(0,255,255)
                , BMP.black
                , 0
                , font_size_factor
                , "_"
              );
    bmp_to_SDL_Surface(&bmp, screen);
      
              
    while(!done)
    {
        do
        {
            SDL_WaitEvent(&event);
        } while(    event.type != SDL_KEYDOWN
                 && event.type != SDL_QUIT
               );               
        if(event.type == SDL_QUIT)
            return "";
        if(event.key.keysym.sym == SDLK_ESCAPE)
            return "";
        if(event.key.keysym.sym == SDLK_RETURN)
            done = true;
        else
        {
            switch(event.key.keysym.sym)
            {
                //-------------------------------------------------    
                case SDLK_LEFT:
                case SDLK_DELETE:
                case SDLK_BACKSPACE:
                    if(input.size())
                    {
                        input = input.substr(0, input.size() - 1);
                        x_off -= x_gap;
                        bmp_printf(   &BMP
                                    , x_off
                                    , y_off
                                    , BMP.make_rgb(0,255,255)
                                    , BMP.black
                                    , 0
                                    , font_size_factor
                                    , " "
                                  );
                        bmp_to_SDL_Surface(&bmp, screen);
                    }                         
                    break;
                //-------------------------------------------------    
                default:
                    if(    isgraph(event.key.keysym.sym)
                        || event.key.keysym.sym == ' '
                      )
                    {
                        bmp_printf(   &BMP
                                    , x_off
                                    , y_off
                                    , BMP.make_rgb(0,255,255)
                                    , BMP.black
                                    , 0
                                    , font_size_factor
                                    , "%c"
                                    , event.key.keysym.unicode
                                  );
                        bmp_to_SDL_Surface(&bmp, screen);
                        input += event.key.keysym.unicode;
                        x_off += x_gap;
                    }                            
                    break;                         
                //-------------------------------------------------    
            } // end switch(event.key.keysym.sym)
        } // end else of if(event.key.keysym.sym == SDLK_RETURN)
    } // end while(!done)
    
    bmp_printf(   &BMP
                , x_off - x_gap * input.size()
                , y_off
                , BMP.make_rgb(255,0,0)
                , BMP.black
                , 0
                , font_size_factor
                , input.c_str()
              );
    bmp_to_SDL_Surface(&bmp, screen);
    return input;
}

//############################################################################
float ez_ilda::display_prompt_and_echo_float(string prompt)
{
    bool   done = false;

    int    x_off  = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off  = BMP.yres - 24 * font_size_factor,
           x_gap  = 8 * font_size_factor,
           y_line = 8 * font_size_factor;

    float  result;

    string input;

    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.black
                , 0
                , font_size_factor
                , prompt.c_str()
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.make_rgb(0,255,255)
                , BMP.black
                , 0
                , font_size_factor
                , "_"
              );
    bmp_to_SDL_Surface(&bmp, screen);

    while(!done)
    {
        do
        {
            SDL_WaitEvent(&event);
        } while(    event.type != SDL_KEYDOWN
                 && event.type != SDL_QUIT
               );               
        if(event.type == SDL_QUIT)
            return 0.0;
        if(event.key.keysym.sym == SDLK_ESCAPE)
            return 0.0;
        if(event.key.keysym.sym == SDLK_RETURN)
            done = true;
        else
        {
            switch(event.key.keysym.sym)
            {
                //-------------------------------------------------    
                case SDLK_LEFT:
                case SDLK_DELETE:
                case SDLK_BACKSPACE:
                    if(input.size())
                    {
                        input = input.substr(0, input.size() - 1);
                        x_off -= x_gap;
                        bmp_printf(   &BMP
                                    , x_off
                                    , y_off
                                    , BMP.make_rgb(0,255,255)
                                    , BMP.black
                                    , 0
                                    , font_size_factor
                                    , " "
                                  );
                        bmp_to_SDL_Surface(&bmp, screen);
                    }                         
                    break;
                //-------------------------------------------------    
                default:
                    if(    isdigit(event.key.keysym.sym)
                        || (    event.key.keysym.sym == '.'
                             && input.find('.') == string::npos
                           )
                        || (    event.key.keysym.sym == '+'
                             && !input.size()
                           )
                        || (    event.key.keysym.sym == '-'
                             && !input.size()
                           )
                      )                            
                    {
                        bmp_printf(   &BMP
                                    , x_off
                                    , y_off
                                    , BMP.make_rgb(0,255,255)
                                    , BMP.black
                                    , 0
                                    , font_size_factor
                                    , "%c"
                                    , event.key.keysym.sym
                                  );
                        bmp_to_SDL_Surface(&bmp, screen);
                        input += event.key.keysym.sym;
                        x_off += x_gap;
                    }                            
                    break;                         
                //-------------------------------------------------    
            } // end switch(event.key.keysym.sym)
        } // end else of if(event.key.keysym.sym == SDLK_RETURN)
    } // end while(!done)
    
    bmp_printf(   &BMP
                , x_off - x_gap * input.size()
                , y_off
                , BMP.make_rgb(255,0,0)
                , BMP.black
                , 0
                , font_size_factor
                , input.c_str()
              );
    bmp_to_SDL_Surface(&bmp, screen);

    sscanf(input.c_str(), "%f", &result);
    return result;
}

//############################################################################
int ez_ilda::display_prompt_and_echo_int(string prompt)
{
    bool   done = false;
    
    int    x_off  = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off  = BMP.yres - 24 * font_size_factor,
           x_gap  = 8 * font_size_factor,
           y_line = 8 * font_size_factor,
           result;

    string input;
    
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.black
                , 0
                , font_size_factor
                , prompt.c_str()
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.make_rgb(0,255,255)
                , BMP.black
                , 0
                , font_size_factor
                , "_"
              );
    bmp_to_SDL_Surface(&bmp, screen);

    while(!done)
    {
        do
        {
            SDL_WaitEvent(&event);
        } while(    event.type != SDL_KEYDOWN
                 && event.type != SDL_QUIT
               );
        if(event.type == SDL_QUIT) // clicked window X
            return 0;
        if(event.key.keysym.sym == SDLK_ESCAPE)
            return 0;
        if(event.key.keysym.sym == SDLK_RETURN)
            done = true;
        else
        {
            switch(event.key.keysym.sym)
            {
                //-------------------------------------------------    
                case SDLK_LEFT:
                case SDLK_DELETE:
                case SDLK_BACKSPACE:
                    if(input.size())
                    {
                        input = input.substr(0, input.size() - 1);
                        x_off -= x_gap;
                        bmp_printf(   &BMP
                                    , x_off
                                    , y_off
                                    , BMP.make_rgb(0,255,255)
                                    , BMP.black
                                    , 0
                                    , font_size_factor
                                    , " "
                                  );
                        bmp_to_SDL_Surface(&bmp, screen);
                    }                         
                    break;
                //-------------------------------------------------    
                default:
                    if(    isdigit(event.key.keysym.sym)
                        || (    event.key.keysym.sym == '+'
                             && !input.size()
                           )
                        || (    event.key.keysym.sym == '-'
                             && !input.size()
                           )
                      )                    
                    {
                        bmp_printf(   &BMP
                                    , x_off
                                    , y_off
                                    , BMP.make_rgb(0,255,255)
                                    , BMP.black
                                    , 0
                                    , font_size_factor
                                    , "%c"
                                    , event.key.keysym.sym
                                  );
                        bmp_to_SDL_Surface(&bmp, screen);
                        input += event.key.keysym.sym;
                        x_off += x_gap;
                    }                            
                    break;                         
                //-------------------------------------------------    
            } // end switch(event.key.keysym.sym)
        } // end else of if(event.key.keysym.sym == SDLK_RETURN)
    } // end while(!done)

    bmp_printf(   &BMP
                , x_off - x_gap * input.size()
                , y_off
                , BMP.make_rgb(255,0,0)
                , BMP.black
                , 0
                , font_size_factor
                , input.c_str()
              );
    bmp_to_SDL_Surface(&bmp, screen);

    sscanf(input.c_str(), "%d", &result);    
    return result;
}

//############################################################################
bool ez_ilda::display_prompt_and_echo_bool(string prompt)
{
    int    x_off  = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
           y_off  = BMP.yres - 24 * font_size_factor;

    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.black
                , 0
                , font_size_factor
                , prompt.c_str()
              );
    bmp_to_SDL_Surface(&bmp, screen);    
    do
    {
        SDL_WaitEvent(&event);
    } while(event.type != SDL_KEYDOWN);
    if(event.key.keysym.sym == 'y')
        return true;
    return false;
}

//############################################################################
void ez_ilda::display_program_menu()
{
    int y_line = 8 * font_size_factor,
        x_off  = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
        y_off  = x_off;

    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.black
                , 0
                , font_size_factor
                , "Welcome to LaserBoy! 04:20:2008 build 1"
              );

    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "        i open file in"
              );

    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "        o save file out"       
              );

    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "        u user interface visuals"
              );

    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "        v view"
              );

    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "      ` ~ +- show"
              );

    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "      - _ +- spin"
              );

    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "      = + +- show & spin"
              );

    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "      %c %c +- frame"
                , 0x1b
                , 0x1a
              );

    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "      %c %c +- target palette"
                , 0x18
                , 0x19
              );

    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "        9 first frame"
              );

    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "        0 last frame"
              );

    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "        ( select frames-( start"
              );

    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "        ) select )-frames end"
              );

    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "  [Space] select frame"
              );

    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "[Bkspace] delete frame"
              );

    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "        p palette transforms"
              );

    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "        h frame transforms"
              );

    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "        j frame set transforms"
              );

    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "        k draw, color, blank"
              );

    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "        l move, size, spin"
              );

    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "      / ? frame | set stats"
              );

    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    [Tab] system settings"
              );

    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    [Esc] quit saving OUT.ild"
              );

    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "  [Enter] +- this help"
              );

    bmp_to_SDL_Surface(&bmp, screen);
    return;
}

//############################################################################
void ez_ilda::display_please_wait()
{
    int x_off  = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
        y_line = 8 * font_size_factor,
        y_off  = y_line;

    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.black
                , 0
                , font_size_factor
                , "please wait..."
              );

    bmp_to_SDL_Surface(&bmp, screen);
    return;
}

//############################################################################
void ez_ilda::display_in_file_type_menu()
{
    int y_line = 8 * font_size_factor,
        x_off  = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
        y_off  = x_off;

    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.black
                , 0
                , font_size_factor
                , "file input type"
              );

    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    1 ILDA"
              );

    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    2 DXF"
              );

    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    3 wave"
              );

    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    4 bitmap"
              );

    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    5 ascii palette"
              );

    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.black
                , 0
                , font_size_factor
                , "[Esc] cancel action"
              );

    bmp_to_SDL_Surface(&bmp, screen);
    return;
}

//############################################################################
void ez_ilda::display_in_ild_action_menu()
{
    int y_line = 8 * font_size_factor,
        x_off  = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
        y_off  = x_off;

    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.black
                , 0
                , font_size_factor
                , "ILDA file input action"
              );

    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    1 replace frame set"
              );

    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    2 prepend to frame set"
              );

    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    3 insert after current frame"
              );

    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    4 append to frame set"
              );

    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    5 superimpose frame(s) into frame set"
              );

    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.black
                , 0
                , font_size_factor
                , "[Esc] cancel action"
              );

    bmp_to_SDL_Surface(&bmp, screen);
    return;
}

//############################################################################
void ez_ilda::display_in_dxf_action_menu()
{
    int y_line = 8 * font_size_factor,
        x_off  = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
        y_off  = x_off;

    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.black
                , 0
                , font_size_factor
                , "DXF file input action"
              );

    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    1 replace frame set with DXF file"
              );

    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    2 append  frame set with DXF file"
              );

    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    3 replace frame set with DXF directory"
              );

    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    4 append  frame set with DXF directory"
              );

    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.black
                , 0
                , font_size_factor
                , "[Esc] cancel action"
              );

    bmp_to_SDL_Surface(&bmp, screen);
    return;
}

//############################################################################
void ez_ilda::display_in_wave_action_menu()
{
    int y_line = 8 * font_size_factor,
        x_off  = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
        y_off  = x_off;

    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.black
                , 0
                , font_size_factor
                , "wave file input action"
              );

    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    1 replace current frame set"
              );

    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    2 append to current frame set"
              );

    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.black
                , 0
                , font_size_factor
                , "[Esc] cancel action"
              );

    bmp_to_SDL_Surface(&bmp, screen);
    return;
}

//############################################################################
void ez_ilda::display_in_bmp_action_menu()
{
    int y_line = 8 * font_size_factor,
        x_off  = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
        y_off  = x_off;

    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.black
                , 0
                , font_size_factor
                , "bitmap file input action"
              );

    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    1 open as frame set color map"
              );

    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    2 open into frame background"
              );

    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    3 clear frame background"
              );

    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.black
                , 0
                , font_size_factor
                , "[Esc] cancel action"
              );

    bmp_to_SDL_Surface(&bmp, screen);
    return;
}

//############################################################################
//void ez_ilda::display_in_pal_action_menu()
//{
//     return;
//}

//############################################################################
void ez_ilda::display_out_file_type_menu()
{
    int y_line = 8 * font_size_factor,
        x_off  = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
        y_off  = x_off;

    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.black
                , 0
                , font_size_factor
                , "Output file type"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    1 ILDA current frame"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    2 ILDA selected frames"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    3 ILDA all frames"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    4 ILDA frame effect"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    5 ILDA frame set effect"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    6 DXF save current frame"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    7 DXF save directory selected frames"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    8 DXF save directory all frames"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    9 wave %6.2f second still of current frame"
                , still_frame_duration
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    0 wave animation %3.3f fps of all frames"
                , frames_per_second
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    a C++ save current frame"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    b C++ save selected frames"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    c C++ save all frames"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    d bitmap %d square pixels of current frame"
                , output_bmp_size
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    e bitmap %d square pixels directory of selected frames"
                , output_bmp_size
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    f bitmap %d square pixels directory of all frames"
                , output_bmp_size
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    g ASCII palette of current frame"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    h ASCII palette of target palette"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.black
                , 0
                , font_size_factor
                , "[ESC] cancel action"
              );
    bmp_to_SDL_Surface(&bmp, screen);    
    return;
}

//############################################################################
void ez_ilda::display_settings_menu()
{
    int y_line = 8 * font_size_factor,
        x_off  = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
        y_off  = x_off;

    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.black
                , 0
                , font_size_factor
                , "LaserBoy settings"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    1 zero angle tolorance in degrees....... %f"
                , zero_angle_tolerance * one_radian
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    2 distance delta max in points.......... %f"
                , distance_delta_max
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    3 maximum corner dwell in microseconds.. %f"
                , longest_dwell_microsec
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    4 insignificant distance in points...... %f"
                , insignificant_distance
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    5 rendered arc segment angle in degrees. %f"
                , rendered_arc_angle
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    6 move points per key hit............... %d"
                , move_points_per_key
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    7 recolor palette span factor........... %f"
                , recolor_span_factor
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    8 animation frames per second........... %f"
                , frames_per_second
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    9 still frame durration in seconds...... %f"
                , still_frame_duration
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    0 output wave sample rate............... %d"
                , wave_sample_rate
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    a display font size factor.............. %d"
                , font_size_factor
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    b output bitmap size in square pixels... %d"
                , output_bmp_size
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.black
                , 0
                , font_size_factor
                , "[ESC] cancel action"
              );
    bmp_to_SDL_Surface(&bmp, screen);    
    return;
}

//############################################################################
void ez_ilda::display_veiw_menu()
{
    int x_off  = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
        y_off  = x_off,
        y_line =  8 * font_size_factor,
        x_tab  = 27 * y_line,
        x_tab2 =  9 * y_line;

    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.black
                , 0
                , font_size_factor
                , "press and hold view hotkeys"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.make_rgb(255,255,0)
                , BMP.black
                , 0
                , font_size_factor
                , "1 !  +- spin X"
              );
    bmp_printf(   &BMP
                , x_tab
                , y_off
                , BMP.make_rgb(0,255,255)
                , BMP.black
                , 0
                , font_size_factor
                , "5 %%  front | back"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.make_rgb(255,255,0)
                , BMP.black
                , 0
                , font_size_factor
                , "2 @  +- spin Y"
              );
    bmp_printf(   &BMP
                , x_tab
                , y_off
                , BMP.make_rgb(0,255,255)
                , BMP.black
                , 0
                , font_size_factor
                , "6 ^   left | right"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.make_rgb(255,255,0)
                , BMP.black
                , 0
                , font_size_factor
                , "3 #  +- spin Z"
              );
    bmp_printf(   &BMP
                , x_tab
                , y_off
                , BMP.make_rgb(0,255,255)
                , BMP.black
                , 0
                , font_size_factor
                , "7 &    top | bottom"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.make_rgb(255,255,0)
                , BMP.black
                , 0
                , font_size_factor
                , "4 $  +- spin XYZ"
              );
    bmp_printf(   &BMP
                , x_tab
                , y_off
                , BMP.make_rgb(0,255,255)
                , BMP.black
                , 0
                , font_size_factor
                , "8 *  +- spin & frame"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.make_rgb(255,0,255)
                , BMP.black
                , 0
                , font_size_factor
                , "q Q  +- pan X"
              );
    bmp_printf(   &BMP
                , x_tab
                , y_off
                , BMP.make_rgb(0,255,0)
                , BMP.black
                , 0
                , font_size_factor
                , "a A  +- size X"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.make_rgb(255,0,255)
                , BMP.black
                , 0
                , font_size_factor
                , "w W  +- pan Y"
              );
    bmp_printf(   &BMP
                , x_tab
                , y_off
                , BMP.make_rgb(0,255,0)
                , BMP.black
                , 0
                , font_size_factor
                , "s S  +- size Y"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.make_rgb(255,0,255)
                , BMP.black
                , 0
                , font_size_factor
                , "e E  +- pan Z"
              );
    bmp_printf(   &BMP
                , x_tab
                , y_off
                , BMP.make_rgb(0,255,0)
                , BMP.black
                , 0
                , font_size_factor
                , "d D  +- size Z"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.make_rgb(255,0,255)
                , BMP.black
                , 0
                , font_size_factor
                , "r R  +- pan XYZ"
              );
    bmp_printf(   &BMP
                , x_tab
                , y_off
                , BMP.make_rgb(0,255,0)
                , BMP.black
                , 0
                , font_size_factor
                , "f F  +- size XYZ"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.make_rgb(255,0,255)
                , BMP.black
                , 0
                , font_size_factor
                , "t T  set to center"
              );
    bmp_printf(   &BMP
                , x_tab
                , y_off
                , BMP.make_rgb(0,255,0)
                , BMP.black
                , 0
                , font_size_factor
                , "g G  set to 1:1:1"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.black
                , 0
                , font_size_factor
                , "current view values"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_tab2
                , y_off
                , BMP.make_rgb(255,0,0)
                , BMP.black
                , 0
                , font_size_factor
                , "X"
              );
    bmp_printf(   &BMP
                , x_tab2 * 2
                , y_off
                , BMP.make_rgb(0,255,0)
                , BMP.black
                , 0
                , font_size_factor
                , "Y"
              );
    bmp_printf(   &BMP
                , x_tab2 * 3
                , y_off
                , BMP.make_rgb(0,0,255)
                , BMP.black
                , 0
                , font_size_factor
                , "Z"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.make_rgb(255,255,0)
                , BMP.black
                , 0
                , font_size_factor
                , " angle           ,         ,           degrees"
              );
    bmp_printf(   &BMP
                , x_tab2
                , y_off
                , BMP.make_rgb(255,255,0)
                , BMP.make_rgb(80,0,0)
                , 0
                , font_size_factor
                , "%-3.4f"
                , view_angle.x * one_radian
              );
    bmp_printf(   &BMP
                , x_tab2 * 2
                , y_off
                , BMP.make_rgb(255,255,0)
                , BMP.make_rgb(0,80,0)
                , 0
                , font_size_factor
                , "%-3.4f"
                , view_angle.y * one_radian
              );
    bmp_printf(   &BMP
                , x_tab2 * 3
                , y_off
                , BMP.make_rgb(255,255,0)
                , BMP.make_rgb(0,0,80)
                , 0
                , font_size_factor
                , "%-3.4f"
                , view_angle.z * one_radian
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.make_rgb(255,0,255)
                , BMP.black
                , 0
                , font_size_factor
                , "offset           ,         ,           points"
              );
    bmp_printf(   &BMP
                , x_tab2
                , y_off
                , BMP.make_rgb(255,0,255)
                , BMP.make_rgb(80,0,0)
                , 0
                , font_size_factor
                , "%-8.0f"
                , view_offset.x
              );
    bmp_printf(   &BMP
                , x_tab2 * 2
                , y_off
                , BMP.make_rgb(255,0,255)
                , BMP.make_rgb(0,80,0)
                , 0
                , font_size_factor
                , "%-8.0f"
                , view_offset.y
              );
    bmp_printf(   &BMP
                , x_tab2 * 3
                , y_off
                , BMP.make_rgb(255,0,255)
                , BMP.make_rgb(0,0,80)
                , 0
                , font_size_factor
                , "%-8.0f"
                , view_offset.z
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.make_rgb(0,255,0)
                , BMP.black
                , 0
                , font_size_factor
                , " scale           ,         ,"
              );
    bmp_printf(   &BMP
                , x_tab2
                , y_off
                , BMP.make_rgb(0,255,0)
                , BMP.make_rgb(80,0,0)
                , 0
                , font_size_factor
                , "%-3.4f"
                , view_scale.x
              );
    bmp_printf(   &BMP
                , x_tab2 * 2
                , y_off
                , BMP.make_rgb(0,255,0)
                , BMP.make_rgb(0,80,0)
                , 0
                , font_size_factor
                , "%-3.4f"
                , view_scale.y
              );
    bmp_printf(   &BMP
                , x_tab2 * 3
                , y_off
                , BMP.make_rgb(0,255,0)
                , BMP.make_rgb(0,0,80)
                , 0
                , font_size_factor
                , "%-3.4f"
                , view_scale.z
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.make_rgb(255,255,0)
                , BMP.black
                , 0
                , font_size_factor
                , "      1 apply view angle  to frame data"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.make_rgb(255,0,255)
                , BMP.black
                , 0
                , font_size_factor
                , "      2 apply view offset to frame data"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.make_rgb(0,255,0)
                , BMP.black
                , 0
                , font_size_factor
                , "      3 apply view scale  to frame data"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.black
                , 0
                , font_size_factor
                , "[Shift] 1, 2, 3 apply to all frames"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.black
                , 0
                , font_size_factor
                , "  [Esc] cancel action"
              );
    bmp_to_SDL_Surface(&bmp, screen);
    return;
}

//############################################################################
void ez_ilda::display_visuals_menu()
{
    int y_line = 8 * font_size_factor,
        x_off  = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
        y_off  = x_off;

    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.black
                , 0
                , font_size_factor
                , "user interface visuals"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , ((show_points)?("->  1 points"):("    1 points"))
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , ((show_blanking)?("->  2 blanking"):("    2 blanking"))
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , ((show_palettes)?("->  3 palettes"):("    3 palettes"))
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , ((show_stats)?("->  4 stats"):("    4 stats"))
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , ((show_fixed_bounds)?("->  5 fixed bounds"):("    5 fixed bounds"))
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , ((show_fixed_origin)?("->  6 fixed origin"):("    6 fixed origin"))
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , ((show_floating_axis)?("->  7 floating axis"):("    7 floating axis"))
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , ((show_floating_bounds)?("->  8 floating bounds"):("    8 floating bounds"))
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    9 show all visuals"
                , ((show_ui_message)?("->  9 UI message"):("    9 UI message"))
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    0 show | hide all visuals"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.black
                , 0
                , font_size_factor
                , "toggle settings"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , ((show_effects_generation)?("->  a show live frame set rendering"):("    a show live frame set rendering"))
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , ((auto_flatten_z)?("->  b flatten Z for wave output"):("    b flatten Z for wave output"))
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , ((auto_optimize)?("->  c auto optimize before output"):("    c auto optimize before output"))
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , ((loop_animation)?("->  d loop animation"):("    d loop animation"))
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , ((approximate_frame_rate)?("->  e animate at approximated frame rate"):("    e animate at approximated frame rate"))
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , ((auto_scale_dxf)?("->  f auto scale DXF import"):("    f auto scale DXF import"))
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , ((maintain_dxf_origin)?("->  g maintain DXF origin"):("    g maintain DXF origin"))
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , ((black_dwell_points)?("->  h black dwell"):("    h black dwell"))
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.black
                , 0
                , font_size_factor
                , "[Esc] cancel"
              );
    bmp_to_SDL_Surface(&bmp, screen);
    return;
}

//############################################################################
void ez_ilda::display_palette_tranz_menu()
{
    int y_line = 8 * font_size_factor,
        x_off  = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
        y_off  = x_off;

    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.black
                , 0
                , font_size_factor
                , "palette transforms"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "      1 span X with palette"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "      2 span Y with palette"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "      3 span Z with palette"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "      4 span vectors with palette"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "      5 index vectors with palette"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "      6 span lit segments with palette"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "      7 index lit segments with palette"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "      8 random color vectors with palette"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "      9 random color lit segments with palette"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "      0 set to target palette by index"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "      a best match to target palette"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "      b convert true color to palette (if possible)"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "      c bit reduce true color to palette"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "      d promote to true color"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "[Shift] select to apply to all frames"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.black
                , 0
                , font_size_factor
                , "  [Esc] cancel action"
              );
    bmp_to_SDL_Surface(&bmp, screen);
    return;
}

//############################################################################
void ez_ilda::display_frame_set_tranz_menu()
{
    int y_line = 8 * font_size_factor,
        x_off  = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
        y_off  = x_off;

    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.black
                , 0
                , font_size_factor
                , "frame set transforms"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "      1 delete selected frames"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "      2 trim to selected frames"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "      3 copy selected frames to begining of frame set"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "      4 move selected frames to begining of frame set"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "      5 copy selected frames to end of frame set"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "      6 move selected frames to end of frame set"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "      7 delete redundant frames from wave import"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "      8 add empty frame to begining of frame set"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "      9 add empty frame after current frame in frame set"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "      0 add empty frame to end of frame set"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "      a reverse frame set order"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "      > explode current frame into segment frame set"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "      < collapse selected frames into one frame"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.black
                , 0
                , font_size_factor
                , "  [Esc] cancel action"
              );
    bmp_to_SDL_Surface(&bmp, screen);
    return;
}

//############################################################################
void ez_ilda::display_frame_tranz_menu()
{
    int y_line = 8 * font_size_factor,
        x_off  = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
        y_off  = x_off;

    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.black
                , 0
                , font_size_factor
                , "frame transforms"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "      1 flatten Z (force to 2D)"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "      2 Z order points"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "      3 reverse point order"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "      4 mirror"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "      5 flip"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "      6 quarter turn clockwise"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "      7 quarter turn counter clockwise"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "      8 remove blank span"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "      9 remove angle dwell"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "      0 reduce vectors"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "      a bond segments (reorders segments)"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "      b minimize points (8,9,0,a)"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "      c fracture segments"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "      d reorder segments"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "      e add angle dwell"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "      f add distance span"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "      g optimize"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "[Shift] select to apply to all frames:"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.black
                , 0
                , font_size_factor
                , "  [Esc] cancel action"
              );
    bmp_to_SDL_Surface(&bmp, screen);
    return;
}

//############################################################################
void ez_ilda::display_draw_color_blank_menu()
{
    int y_line = 8 * font_size_factor,
        x_off  = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
        y_off  = x_off;

    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.black
                , 0
                , font_size_factor
                , "draw, color, blank mode entered"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "        \\ | start to first | end to last"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "        [ { start  +- one vertex"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "        ] } end    +- one vertex"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "        i I start  +- lit segment"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "        o O end    +- lit segment"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "        u U select +- lit segment"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "        a A unblank all points | frames"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "        b B blank | unblank select"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "        p P +- palette color index"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "          c color select with palette index"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "          C set palette index to start color"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "        x X move   start X +- 100"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "        y Y move   start Y +- 100"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "        z Z move   start Z +- 100"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "          q set    start X"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "          w set    start Y"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "          e set    start Z"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "        f F rotate start +- 1 degree around Z axis"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "        g G rotate start +- 1 degree around Y axis"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "        m M +-1%% magnitude of start"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "          l draw line from start to end"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "          t render text on line from start to end"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "          d %3.2f deg circular  arc, start to end radius"
                , rendered_arc_angle
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "          D %3.2f deg eliptical arc, start to end radius"
                , rendered_arc_angle
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "          k draw polygon at start center end radius"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "          . split vector in line or add vector to end"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "          s add blank point at start (cut segment)"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "[Backspace] delete vertex at start"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "        / ? show frame | set stats"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "%c %c %c %c and view spin keys are active"
                , 0x1b
                , 0x1a
                , 0x18
                , 0x19
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    [Enter] +- this help"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.black
                , 0
                , font_size_factor
                , "      [Esc] exit draw, color blank mode"
              );
    bmp_to_SDL_Surface(&bmp, screen);
    return;
}

//############################################################################
void ez_ilda::display_move_size_spin_menu()
{
    int y_line = 8 * font_size_factor,
        x_off  = EZ_STATS_DISPLAY_OFFSET * font_size_factor,
        y_off  = x_off;

    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.black
                , 0
                , font_size_factor
                , "move, size, spin mode entered"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    \\ | start to first | end to last"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    [ { start  +- one vertex"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    ] } end    +- one vertex"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    i I start  +- lit segment"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    o O end    +- lit segment"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    u U select +- lit segment"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    x X move   select X +- 100"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    y Y move   select Y +- 100"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    z Z move   select Z +- 100"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "      q set    select center X"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "      w set    select center Y"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "      e set    select center Z"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    f F rotate select +- 1' Z axis on center"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    g G rotate select +- 1' Y axis on center"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    8 * scale  select X   +- 1%% on center"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    9 ( scale  select  Y  +- 1%% on center"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    0 ) scale  select   Z +- 1%% on center"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    - _ scale  select XYZ +- 1%% on center"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    a A move   a point X +- 100"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    s S move   a point Y +- 100"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    d D move   a point Z +- 100"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "      j set    a point X"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "      k set    a point Y"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "      l set    a point Z"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    r R rotate select +- 1' Z axis on a point"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    t T rotate select +- 1' Y axis on a point"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    v V scale  select X   +- 1%% from a point"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    b B scale  select  Y  +- 1%% from a point"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    n N scale  select   Z +- 1%% from a point"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    m M scale  select XYZ +- 1%% from a point"
              );
    y_off += y_line;
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "      c copy   select (in place) to end"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "    / ? frame | set stats"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "%c %c %c %c and view spin keys are active"
                , 0x1b
                , 0x1a
                , 0x18
                , 0x19
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , font_size_factor
                , "[Enter] +- this help"
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.black
                , 0
                , font_size_factor
                , "  [ESC] exit move, size, spin mode"
              );
    bmp_to_SDL_Surface(&bmp, screen);
    return;
}

//############################################################################
void ez_ilda::list_files(string directory)
{
    bool            found_one = false;
    int             i,
                    y_line = 8 * font_size_factor,
                    x_off  = BMP.yres + EZ_STATS_DISPLAY_OFFSET * font_size_factor,
                    y_off  = y_line;
    DIR            *pDIR;
    struct dirent  *pDirEnt;

    pDIR = opendir(directory.c_str());
    if(pDIR == NULL)
        return;
    bmp_printf(   &BMP
                , x_off
                , 0
                , BMP.white
                , BMP.black
                , 0
                , font_size_factor
                , "                  "
              );
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.black
                , 0
                , font_size_factor
                , "                  "
              );
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.black
                , 0
                , font_size_factor
                , "Palette set stats "
              );
    pDirEnt = readdir(pDIR);  
    while(pDirEnt != NULL)
    {
        if(    strcmp(pDirEnt->d_name, ".")  // not the dot
            && strcmp(pDirEnt->d_name, "..") // not the dot dot
          )
        {
            bmp_printf(   &BMP
                        , x_off
                        , y_off
                        , next_hue()
                        , BMP.black
                        , 0
                        , font_size_factor
                        , "%s                  "
                        , pDirEnt->d_name
                      );
            y_off += y_line;
            found_one = true;
        }
        pDirEnt = readdir(pDIR);
    }
    closedir(pDIR);    
    if(!found_one)
    {
        bmp_printf(   &BMP
                    , x_off
                    , y_off
                    , next_hue()
                    , BMP.black
                    , 0
                    , font_size_factor
                    , "none found        "
                  );
        y_off += y_line;
    }    
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.black
                , 0
                , font_size_factor
                , "                  "
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.black
                , 0
                , font_size_factor
                , "                  "
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.black
                , 0
                , font_size_factor
                , "                  "
              );
    y_off += y_line;
    bmp_to_SDL_Surface(&bmp, screen);
    return;
}

//############################################################################
void ez_ilda::list_directories(string directory)
{
    bool            found_one = false;
    int             i,
                    y_line = 8 * font_size_factor,
                    x_off  = BMP.yres + EZ_STATS_DISPLAY_OFFSET * font_size_factor,
                    y_off  = y_line;
    DIR            *pDIR;
    struct dirent  *pDirEnt;

    pDIR = opendir(directory.c_str());
    if(pDIR == NULL)
        return;
    bmp_printf(   &BMP
                , x_off
                , 0
                , BMP.white
                , BMP.black
                , 0
                , font_size_factor
                , "                  "
              );
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.black
                , 0
                , font_size_factor
                , "                  "
              );
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.black
                , 0
                , font_size_factor
                , "Palette set stats "
              );
    pDirEnt = readdir(pDIR);  
    while(pDirEnt != NULL)
    {
        if(!strstr(pDirEnt->d_name, ".")) // if there is no dot in the name
        {
            bmp_printf(   &BMP
                        , x_off
                        , y_off
                        , next_hue()
                        , BMP.black
                        , 0
                        , font_size_factor
                        , "%s                  "
                        , pDirEnt->d_name
                      );
            y_off += y_line;
            found_one = true;
        }
        pDirEnt = readdir(pDIR);
    }
    closedir(pDIR);
    if(!found_one)
    {
        bmp_printf(   &BMP
                    , x_off
                    , y_off
                    , next_hue()
                    , BMP.black
                    , 0
                    , font_size_factor
                    , "none found        "
                  );
        y_off += y_line;
    }
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.black
                , 0
                , font_size_factor
                , "                  "
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.black
                , 0
                , font_size_factor
                , "                  "
              );
    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.black
                , 0
                , font_size_factor
                , "                  "
              );
    y_off += y_line;
    bmp_to_SDL_Surface(&bmp, screen);
    return;
}

//############################################################################
void ez_ilda::load_background_bitmap(struct ezfb_bitmap* bmp)
{
    bmp_flip(bmp);
    bmp_into_bmp(&BACKGROUND, bmp);
    return;
}

//############################################################################
void ez_ilda::load_background_bitmap(string file)
{
    char file_name[81];
    struct ezfb_bitmap bmp;
    if(    file.size() < 4
        || file.substr(file.size() - 4) != ".bmp"
      )
        file += ".bmp";
    strcpy(file_name, (EZ_ILDA_BMP_SHARE + file).c_str());
    if(bmp_from_file(&bmp, file_name))
    {
        load_background_bitmap(&bmp);
        bmp_free(&bmp);
    }
    return;
}

//############################################################################
void ez_ilda::clear_background_bitmap()
{
    bmp_clear(&BACKGROUND, 0x00);    
    return;
}

//############################################################################
void ez_ilda::display_frame_effects()
{
    int i,
        y_line = 8 * font_size_factor,
        x_off  = BMP.yres + EZ_STATS_DISPLAY_OFFSET * font_size_factor,
        y_off  = EZ_STATS_DISPLAY_OFFSET * font_size_factor;
    
    bmp_printf(   &BMP
                , x_off
                , 0
                , BMP.white
                , BMP.black
                , 0
                , font_size_factor
                , "                  "
              );
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.black
                , 0
                , font_size_factor
                , "frame effects"
              );
    if(frame_effect_names.size() > 0)
    {
        for(i = 0; i < (int)frame_effect_names.size(); i++)
        {
            y_off += y_line;
            bmp_printf(   &BMP
                        , x_off
                        , y_off
                        , next_hue()
                        , BMP.black
                        , 0
                        , font_size_factor
                        , "%s          "
                        , frame_effect_names[i].c_str()
                      );
        }
    }
    else
    {
        y_off += y_line;
        bmp_printf(   &BMP
                    , x_off
                    , y_off
                    , next_hue()
                    , BMP.black
                    , 0
                    , font_size_factor
                    , "no frame effects loaded"
                  );
    }        
    bmp_to_SDL_Surface(&BMP, screen);
    return;
}

//############################################################################
void ez_ilda::display_frame_set_effects()
{
    int i,
        y_line = 8 * font_size_factor,
        x_off  = BMP.yres + EZ_STATS_DISPLAY_OFFSET * font_size_factor,
        y_off  = EZ_STATS_DISPLAY_OFFSET * font_size_factor;
    
    bmp_printf(   &BMP
                , x_off
                , 0
                , BMP.white
                , BMP.black
                , 0
                , font_size_factor
                , "                  "
              );
    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.black
                , 0
                , font_size_factor
                , "frame set effects"
              );
    if(frame_set_effect_names.size() > 0)
    {
        for(i = 0; i < (int)frame_set_effect_names.size(); i++)
        {
            y_off += y_line;
            bmp_printf(   &BMP
                        , x_off
                        , y_off
                        , next_hue()
                        , BMP.black
                        , 0
                        , font_size_factor
                        , "%s          "
                        , frame_set_effect_names[i].c_str()
                      );
        }
    }
    else
    {
        y_off += y_line;
        bmp_printf(   &BMP
                    , x_off
                    , y_off
                    , next_hue()
                    , BMP.black
                    , 0
                    , font_size_factor
                    , "no frame set effects loaded"
                  );
    }        
    bmp_to_SDL_Surface(&BMP, screen);
    return;
}

//############################################################################
void ez_ilda::call_frame_effect(string name)
{
    int i;
    for(i = 0; i < (int)frame_effect_names.size(); i++)
        if(frame_effect_names[i] == name)
            break;
    if(i < (int)frame_effect_names.size())
        frame_effects[i](&(current_frame()));
    return;
}

//############################################################################
void ez_ilda::call_frame_set_effect(string name)
{
    int i;
    for(i = 0; i < (int)frame_set_effect_names.size(); i++)
        if(frame_set_effect_names[i] == name)
            break; 
    if(i < (int)frame_set_effect_names.size())
        frame_set_effects[i](this);
    return;
}

//############################################################################
//////////////////////////////////////////////////////////////////////////////
//############################################################################
