//############################################################################
// ~ International Laser Display Association ~
// ~     ILDA file format driver in C++      ~
// ~             built for EZFB              ~
//
// by James Lehman
// Extra Stimulus Inc.
// james@akrobiz.com
//
// began: October 2003
//
//############################################################################
#include "ez_bmp.hpp"
#include "ez_ilda.hpp"
#include "ez_wave.hpp"

//############################################################################
ez_ilda_frame::ez_ilda_frame(ez_ilda_point(*F)(int, int), int points_per_frame, int frame_index)
{
    is_selected          = false                 ;
    is_unique            = true                  ;
    palette_index        = EZ_EZFB_HUES          ;
    error                = EZ_ILDA_OK            ;
    format               = EZ_ILDA_3D_FRAME_DATA ;
    ez_ilda_point point  = F(0, frame_index)     ;
    point.blank();
    push_back(point);
    for(int p = 0; p < points_per_frame; p++)
        push_back(F(p, frame_index));
    push_back(point);
    quantity = (u_short)quantity;
}

//############################################################################
void ez_ilda_frame::save_as_ilda(string file)
{
    if(    file.size() < 4
        || file.substr(file.size() - 4) != ".ild"
      )
        file += ".ild";

    ofstream out((EZ_ILDA_FRAME_SHARE + file).c_str(), ios::out | ios::binary);
    if(out.is_open())
    {
        to_ofstream(out);
        out.close();
    }
    return;
}

//############################################################################
void ez_ilda_frame::optimize_points()
{
    //double                   best_distance           ;
    //int                      best_optimized_index = 0;
    //vector<ez_ilda_segment>  reordered_segments      ;
    //vector<double>           distance                ;
                             //angle                   ;
    //------------------------------------------------------------------------
    //remove_added_points  ();
    //fracture_segments    ();
    minimize_blank_points();
    remove_dwell_points  ();
    minimize_vectors     ();
    bond_segments        ();
    /*
    //------------------------------------------------------------------------
    reordered_segments.push_back(shortest_path_of_segments()); // [0]
    distance.push_back(reordered_segments.back().total_distance());
    //angle.push_back(reordered_segments.back().total_angle());
    best_distance = distance.back();
    //------------------------------------------------------------------------
    reordered_segments.push_back(reordered_segments.back().shortest_path_of_segments()); // [1]
    distance.push_back(reordered_segments.back().total_distance());
    //angle.push_back(reordered_segments.back().total_angle());
    //------------------------------------------------------------------------
    if(distance[0] != distance[1])
    {
        bool    first_match  = false ,
                second_match = false ;
        int     i                    ;
        //--------------------------------------------------------------------
        if(best_distance > distance.back())
        {
            best_distance = distance.back();
            best_optimized_index = 1;
        }
        //--------------------------------------------------------------------
        do
        {
            reordered_segments.push_back(reordered_segments.back().shortest_path_of_segments());
            distance.push_back(reordered_segments.back().total_distance());
            //angle.push_back(reordered_segments.back().total_angle());
            //----------------------------------------------------------------
            for(i = 0; i < (int)distance.size() - 1; i++)
                if(distance.back() == distance[i])
                    break;
            //----------------------------------------------------------------
            if(i < (int)distance.size() - 1) // we have matching distances
            {
                if(first_match)
                    second_match = true;
                first_match = true;
            }
            else if(best_distance > distance.back())
            {
                best_distance = distance.back();
                best_optimized_index = distance.size() - 1;
            }
        } while(!second_match);
    } // end if(min_distance[0] != min_distance[1])
    //------------------------------------------------------------------------
    *this = reordered_segments[best_optimized_index];
    */
    add_dwell_points();
    add_span_points();
    //------------------------------------------------------------------------
    return;
}

//############################################################################
int ez_ilda_frame::true_color_from_ifstream(ifstream& in, const ez_ilda_header& header)
{
    int error = EZ_ILDA_OK;
    *this     = header;
    clear();
    //------------------------------------------------------------------------
    if(quantity)
    {
        int             i;
        ez_color        rgb;
        ez_ilda_point   point;
        ez_ilda_header  next_header;
        //--------------------------------------------------------------------
        for(i = 0; i < (int)quantity; i++)
        {   
            error |= rgb.from_ifstream(in);
            if(!error)
                push_back((ez_ilda_point)rgb);
            else
            {
                ez_ilda_pointer->print_ez_ilda_error(error);
                return error;
            }
        }
        //--------------------------------------------------------------------
        next_header.from_ifstream(in);

        if(quantity != next_header.quantity)
            return EZ_COLOR_POINT_MISSMATCH;

        *this = next_header;

        for(i = 0; i < (int)quantity; i++)
        {   
            error |= point.from_ifstream(in, next_header.format);
            if(!error)
            {
                at(i).x = point.x;
                at(i).y = point.y;
                at(i).z = point.z;
                at(i).k = point.k;
                at(i).c = point.c;
            }
            else
            {
                ez_ilda_pointer->print_ez_ilda_error(error);
                return error;
            }
        }
        //--------------------------------------------------------------------
        front().blank();
        //--------------------------------------------------------------------
        palette_index = EZ_TRUE_COLOR;
        sync_rgb_and_palette();
        //--------------------------------------------------------------------
    }
    //------------------------------------------------------------------------
    return error;
}

//############################################################################
int ez_ilda_frame::from_ifstream(ifstream& in, const ez_ilda_header& header)
{
    int            i,
                   error = EZ_ILDA_OK;
    ez_ilda_point  point;
    //------------------------------------------------------------------------
    *this = header;
    clear();
    //------------------------------------------------------------------------
    if(quantity)
    {
        for(i = 0; i < (int)quantity; i++)
        {   
            error |= point.from_ifstream(in, format);
            if(!error)
                push_back(point);
            else
            {
                ez_ilda_pointer->print_ez_ilda_error(error);
                return error;
            }
        }
        //--------------------------------------------------------------------
        front().blank();
    }
    return error;
}

//############################################################################
int ez_ilda_frame::from_ifstream_dxf(ifstream& in, const ez_ilda_header& header)
{
    int             group_code                ,
                    arc_points    = 0         ,
                    first_segment_point_index ,
                    group_70_flag = 0         ,
                    group_71_flag = 0         ,
                    group_72_flag = 0         ,
                    group_73_flag = 0         ;
       
    string          entity_string             ,
                    font                      ,
                    text                      ;
    
    u_char          color                     ,
                    prev_color                ;

    float           float_value               ,
                    radius                    ,
                    arc_start                 ,
                    arc_end                   ,
                    arc_angle                 ,
                    arc_step                  ,
                    real_size  = 0            ,
                    real_scale = 1            ;

    ez_3D_float     float_3D_1                ,
                    float_3D_2                ,
                    float_3D_3                ,
                    float_3D_4                ,
                    real_min                  ,
                    real_max                  ,
                    real_offset               ;

    ez_ilda_point   point                     ;

    ez_real_segment _real_points              ,
                    real_points               ;

    vector<float>   vertex_x                  ,
                    vertex_y                  ;

    bool            closed_polyline           ;
    //------------------------------------------------------------------------
    while(!in.eof() && entity_string != "ENTITIES") // ignore everthing up to ENTITIES
        in >> entity_string;
      //------------------------------------------------------------------------
    if(!in.eof())
    {
        while(!in.eof())
        {
            color                      = ez_ilda_pointer->palette_picker(EZ_DXF).white;
            prev_color                 = ez_ilda_pointer->palette_picker(EZ_DXF).white;
            first_segment_point_index  = -1,
            closed_polyline            = false;
            group_70_flag              = 0;
            radius                     = 0;
            arc_points                 = 0;
            arc_start                  = 0;
            arc_end                    = 360;
            group_code                 = -1;
            vertex_x.clear();
            vertex_y.clear();
            //----------------------------------------------------------------
            if(entity_string == "LINE")
            {
                while(get_dxf_pair(in, group_code, entity_string))
                {
                    sscanf(entity_string.c_str(), "%f", &float_value);
                    switch(group_code)
                    {
                        case 10: // x1
                                    float_3D_1.x = float_value;
                                    break;
                        case 20: // y1
                                    float_3D_1.y = float_value;
                                    break;
                        case 30: // z1
                                    float_3D_1.z = float_value;
                                    break;
                        case 11: // x2
                                    float_3D_2.x = float_value;
                                    break;
                        case 21: // y2
                                    float_3D_2.y = float_value;
                                    break;
                        case 31: // z2
                                    float_3D_2.z = float_value;
                                    break;
                        case 62: // color
                                    color = (u_char)float_value;
                                    break;
                    }
                }
                real_points.push_back(ez_real_point(float_3D_1, color).blank());
                real_points.push_back(ez_real_point(float_3D_2, color).unblank());
            }
            //------------------------------------------------------------
            else if(    entity_string == "CIRCLE"
                     || entity_string == "ARC"
                   )
            {
                while(get_dxf_pair(in, group_code, entity_string))
                {
                    sscanf(entity_string.c_str(), "%f", &float_value);
                    switch(group_code)
                    {
                        case 10: // x1
                                    float_3D_1.x = float_value;
                                    break;
                        case 20: // y1
                                    float_3D_1.y = float_value;
                                    break;
                        case 30: // z1
                                    float_3D_1.z = float_value;
                                    break;
                        case 40: // radius
                                    radius = float_value;
                                    break;
                        case 50: // start angle
                                    arc_start = float_value;
                                    break;
                        case 51: // end angle
                                    arc_end = float_value;
                                    break;
                        case 62: // color
                                    color = (u_char)float_value;
                                    break;
                    }
                }

                if(arc_start >= arc_end)
                    arc_end += 360.0;
                arc_angle = arc_end - arc_start;
                arc_points = (int)ceil(arc_angle / ez_ilda_pointer->rendered_arc_angle);
                arc_step   = arc_angle / arc_points;
                arc_start *= one_degree;
                arc_end   *= one_degree;
                arc_step  *= one_degree;

                float_3D_2.x = radius * cos(arc_start);
                float_3D_2.y = radius * sin(arc_start);
                float_3D_2.z = float_3D_1.z;

                real_points.push_back(ez_real_point(float_3D_1 + float_3D_2, color).blank());

                for(int a = 1; a <= arc_points; a++)
                {
                    float_3D_2.x = radius * cos(a * arc_step + arc_start);
                    float_3D_2.y = radius * sin(a * arc_step + arc_start);
                    float_3D_2.z = float_3D_1.z;

                    real_points.push_back(ez_real_point(float_3D_1 + float_3D_2, color).unblank());
                }
            }
            //------------------------------------------------------------
            else if(entity_string == "ELLIPSE")
            {
                while(get_dxf_pair(in, group_code, entity_string))
                {
                    sscanf(entity_string.c_str(), "%f", &float_value);
                    switch(group_code)
                    {
                        case 10: // x1
                                    float_3D_1.x = float_value;
                                    break;
                        case 20: // y1
                                    float_3D_1.y = float_value;
                                    break;
                        case 30: // z1
                                    float_3D_1.z = float_value;
                                    break;
                        case 11: // x2
                                    float_3D_2.x = float_value;
                                    break;
                        case 21: // y2
                                    float_3D_2.y = float_value;
                                    break;
                        case 31: // z2
                                    float_3D_2.z = float_value;
                                    break;
                        case 40: // ratio between major and minor axis
                                    radius = float_value;
                                    break;
                        case 41: // start angle
                                    arc_start = float_value;
                                    break;
                        case 42: // end angle
                                    arc_end = float_value;
                                    break;
                        case 62: // color
                                    color = (u_char)float_value;
                                    break;
                    }
                }
                if(arc_start >= arc_end) arc_end += two_pi;
                arc_angle  = arc_end - arc_start;
                arc_points = (int)ceil(arc_angle / (ez_ilda_pointer->rendered_arc_angle * one_degree));
                arc_step   = arc_angle / arc_points;

                float major_axis = float_3D_2.magnitude(),
                      minor_axis = major_axis * radius;

                float_3D_3   = 0.0;
                float_3D_3.z = -float_3D_2.direction().z;

                _real_points.clear();

                float_3D_4.x = -major_axis * cos(arc_start);
                float_3D_4.y = -minor_axis * sin(arc_start);
                float_3D_4.z = float_3D_1.z;

                _real_points.push_back(ez_real_point(float_3D_4, color).blank());

                for(int a = 1; a <= arc_points; a++)
                {
                    float_3D_4.x = -major_axis * cos(a * arc_step + arc_start);
                    float_3D_4.y = -minor_axis * sin(a * arc_step + arc_start);
                    float_3D_4.z = float_3D_1.z;

                    _real_points.push_back(ez_real_point(float_3D_4, color).unblank());
                }

                if(float_3D_3 != 0.0)
                    for(int point_index = 0; point_index < (int)_real_points.size(); point_index++)
                        _real_points[point_index] = rotate_point(_real_points[point_index], float_3D_3);

                for(int point_index = 0; point_index < (int)_real_points.size(); point_index++)
                    real_points.push_back(_real_points[point_index] + float_3D_1);
            }
            //------------------------------------------------------------
            else if(entity_string == "POLYLINE")
            {
                while(get_dxf_pair(in, group_code, entity_string))
                {                    
                    sscanf(entity_string.c_str(), "%f", &float_value);
                    if(    group_code == 70
                        && ((int)float_value) & 0x01
                      )
                        closed_polyline = true;
                }
                while(entity_string == "VERTEX")
                {
                    while(get_dxf_pair(in, group_code, entity_string))
                    {
                        sscanf(entity_string.c_str(), "%f", &float_value);
                        switch(group_code)
                        {
                            case 10: // x1
                                        float_3D_1.x = float_value;
                                        break;
                            case 20: // y1
                                        float_3D_1.y = float_value;
                                        break;
                            case 30: // z1
                                        float_3D_1.z = float_value;
                                        break;
                            case 62: // color
                                        prev_color = color;
                                        color = (u_char)float_value;
                                        break;
                        }
                    }
                    if(first_segment_point_index == -1)
                    {
                        real_points.push_back(ez_real_point(float_3D_1, color).blank());
                        first_segment_point_index = real_points.size() - 1;
                    }
                    else
                        real_points.push_back(ez_real_point(float_3D_1, prev_color).unblank());
                } // end while(entity_string == "VERTEX")
                if(closed_polyline)
                {
                    real_points.push_back(ez_real_point((ez_3D_float)(real_points[first_segment_point_index]), real_points.back().c).unblank());
                }
            }
            //------------------------------------------------------------
            else if(entity_string == "LWPOLYLINE")
            {
                while(get_dxf_pair(in, group_code, entity_string))
                {
                    sscanf(entity_string.c_str(), "%f", &float_value);
                    switch(group_code)
                    {
                        case 90: // number of verticies
                                    arc_points = (int)float_value;
                                    break;                        
                        case 70: // is closed polyline
                                    group_70_flag = (int)float_value;
                                    break;                        
                        case 62: // color
                                    color = (u_char)float_value;
                                    break;
                        case 10: // x
                                    vertex_x.push_back(float_value);
                                    break;
                        case 20: // y
                                    vertex_y.push_back(float_value);
                                    break;
                    }
                }
                if(    (arc_points > 0)
                    && (arc_points == (int)vertex_x.size())
                    && (arc_points == (int)vertex_y.size())
                  )
                {
                    if(color == ez_ilda_pointer->palette_picker(EZ_DXF).black)
                        color = ez_ilda_pointer->palette_picker(EZ_DXF).white;

                    float_3D_1.x = vertex_x[0];
                    float_3D_1.y = vertex_y[0];
                    float_3D_1.z = 0;

                    real_points.push_back(ez_real_point(float_3D_1, color).blank());

                    for(int a = 0; a < arc_points; a++)
                    {
                        float_3D_1.x = vertex_x[a];
                        float_3D_1.y = vertex_y[a];
                        real_points.push_back(ez_real_point(float_3D_1, color).unblank());
                    }

                    if(group_70_flag & 0x01) // closed polyline
                    {
                        float_3D_1.x = vertex_x[0];
                        float_3D_1.y = vertex_y[0];
                        real_points.push_back(ez_real_point(float_3D_1, color).unblank());
                    }
                }
            }
            //------------------------------------------------------------
            else if(entity_string == "TEXT")
            {   
                float_3D_1 = 0.0;
                float_3D_3 = 0.0;
                while(get_dxf_pair(in, group_code, entity_string))
                {                    
                    sscanf(entity_string.c_str(), "%f", &float_value);
                    switch(group_code)
                    {
                        case  1: // the text
                                    text = entity_string;
                                    break;
                        case  7: // the text
                                    font = entity_string;
                                    break;
                        case 10: // x1
                                    float_3D_1.x = float_value;
                                    break;
                        case 20: // y1
                                    float_3D_1.y = float_value;
                                    break;
                        case 30: // z1
                                    float_3D_1.z = float_value;
                                    break;
                        case 40: // text height
                                    radius = float_value;
                                    break;
                        case 50: // flat rotation angle
                                    float_3D_3.z = float_value * one_degree;
                                    break;
                        case 51: // oblique rotation angle
                                    float_3D_3.y = float_value * one_degree;
                                    break;
                        case 62: // color
                                    color = (u_char)float_value;
                                    break;
                        case 71: // flipped or mirrored
                                    group_71_flag = (int)float_value;
                                    break;
                        case 72: // horizontal justification
                                    group_72_flag = (int)float_value;
                                    break;
                        case 73: // vertical justification
                                    group_73_flag = (int)float_value;
                                    break;
                    }
                }

                ez_ilda_frame_set font_frames;

                if(font == "LUCIDA_CONSOLE")
                    font_frames.from_file("lucida.ild");
                else
                    font_frames.from_file("courier_new.ild");

                _real_points.clear();

                for(int text_index = 0; text_index < (int)text.size(); text_index++)
                {
                    if(text[text_index] >= '!' && text[text_index] <= '~')
                    {
                        for(int point_index = 0; point_index < font_frames[text[text_index] - '!'].quantity; point_index++)
                        {
                            float_3D_2 = font_frames[text[text_index] - '!'].at(point_index);
                            float_3D_2.x = ((float_3D_2.x / (float)24000.0) + (float)(text_index)) * radius;
                            float_3D_2.y = ((float_3D_2.y / (float)24000.0)                      ) * radius;
                            _real_points.push_back(   (font_frames[text[text_index] - '!'].at(point_index).is_blank())
                                                    ? (ez_real_point(float_3D_2, color).blank())
                                                    : (ez_real_point(float_3D_2, color).unblank())
                                                  );
                        }
                    }
                }

                if(group_71_flag & 2) // mirrored
                    _real_points.flip(0);

                if(group_71_flag & 4) // flipped
                    _real_points.flip(1);

                if(float_3D_3 != 0.0)
                    for(int point_index = 0; point_index < (int)_real_points.size(); point_index++)
                        _real_points[point_index] = rotate_point(_real_points[point_index], float_3D_3);

                for(int point_index = 0; point_index < (int)_real_points.size(); point_index++)
                    real_points.push_back(_real_points[point_index] + float_3D_1);
            } // end else if(entity_string == "TEXT")             
            //------------------------------------------------------------
            else if(entity_string == "ENDSEC")
                break;
            //------------------------------------------------------------
            else
                get_dxf_pair(in, group_code, entity_string);
        } // end while(!in.eof())

        if(ez_ilda_pointer->auto_scale_dxf)
        {
            real_min =  FLT_MAX;
            real_max = -FLT_MAX;
            for(int i = 0; i < (int)real_points.size(); i++)
            {
                if(real_points[i].x > real_max.x)    real_max.x = real_points[i].x;
                if(real_points[i].x < real_min.x)    real_min.x = real_points[i].x;
                if(real_points[i].y > real_max.y)    real_max.y = real_points[i].y;
                if(real_points[i].y < real_min.y)    real_min.y = real_points[i].y;
                if(real_points[i].z > real_max.z)    real_max.z = real_points[i].z;
                if(real_points[i].z < real_min.z)    real_min.z = real_points[i].z;
            }
            if(ez_ilda_pointer->maintain_dxf_origin)
            {
                if(fabs(real_max.x) > real_size)    real_size = fabs(real_max.x);
                if(fabs(real_min.x) > real_size)    real_size = fabs(real_min.x);
                if(fabs(real_max.y) > real_size)    real_size = fabs(real_max.y);
                if(fabs(real_min.y) > real_size)    real_size = fabs(real_min.y);
                if(fabs(real_max.z) > real_size)    real_size = fabs(real_max.z);
                if(fabs(real_min.z) > real_size)    real_size = fabs(real_min.z);

                real_size *= 2;
            }
            else // find the new center of the universe
            {
                real_offset.x = (real_max.x - ((real_max.x - real_min.x) / 2));
                real_offset.y = (real_max.y - ((real_max.y - real_min.y) / 2));
                real_offset.z = (real_max.z - ((real_max.z - real_min.z) / 2));

                if(fabs(real_max.x - real_min.x) > real_size)    real_size = fabs(real_max.x - real_min.x);
                if(fabs(real_max.y - real_min.y) > real_size)    real_size = fabs(real_max.y - real_min.y);
                if(fabs(real_max.z - real_min.z) > real_size)    real_size = fabs(real_max.z - real_min.z);
            }
            real_scale = 65000.0 / real_size;
        } // end if(auto_scale_dxf)
        //--------------------------------------------------------------------
        for(int i = 0; i < (int)real_points.size(); i++)
        {
            point.x = (short)((real_points[i].x - real_offset.x) * real_scale);
            point.y = (short)((real_points[i].y - real_offset.y) * real_scale);
            point.z = (short)((real_points[i].z - real_offset.z) * real_scale);
            point.c = real_points[i].c;
            if(real_points[i].is_blank())
                point.blank();
            else
                point.unblank();
            //point.set_unique();
            push_back(point);

        }
        set_rgb_from_palette();
    } // end if(!in.eof())
    //------------------------------------------------------------------------
    else
        error = EZ_ILDA_EOF_ERROR;
    //------------------------------------------------------------------------    
    return error;
}

//############################################################################
int ez_ilda_frame::from_ifstream_shp(ifstream& in, const ez_ilda_header& header)
{ 
/*
    get_shp_number(in);
    get_shp_byte_count(in);
    get_shp_name(in);
    get_shp_op_code(in);
*/
    return error;
}  

//############################################################################
int ez_ilda_frame::move_selection(ez_3D_float d)
{
    if(number_of_points() > 1)
    {
        ez_ilda_frame   frame(*this);
        ez_3D_float     f;
        int             i,
                        out_of_bounds = EZ_IN_BOUNDS;

        for(i = 0; i < (int)size(); i++)
        {
            if(is_point_selected(i))
            {
                f = at(i).position();
                out_of_bounds = ez_ilda_bounds_check(f + d, EZ_ILDA_CUBE);
                if(out_of_bounds)
                    return out_of_bounds;
            }
        }
        //--------------------------------------------------------------------
        for(i = 0; i < (int)size(); i++)
        {
            if(is_point_selected(i))
            {
                f = at(i).position();
                at(i) = ez_ilda_point( f + d,
                                       (ez_color)at(i),
                                       at(i).k,
                                       at(i).c
                                     );
            }
        }
        //--------------------------------------------------------------------
    }
    return EZ_IN_BOUNDS;
}
//############################################################################
int ez_ilda_frame::scale_selection(ez_3D_float m)
{
    if(number_of_points() > 1)
    {
        ez_3D_float     f,
                        center = selected_segment().find_center();

        int             i,
                        out_of_bounds = EZ_IN_BOUNDS;

        for(i = 0; i < (int)size(); i++)
        {
            if(is_point_selected(i))
            {
                f = at(i).position();
                out_of_bounds = ez_ilda_bounds_check(((f - center) * m) + center, EZ_ILDA_CUBE);
                if(out_of_bounds)
                    return out_of_bounds;
            }
        }
        //--------------------------------------------------------------------
        for(i = 0; i < (int)size(); i++)
        {
            if(is_point_selected(i))
            {
                f = at(i).position();
                at(i) = ez_ilda_point( ((f - center) * m) + center,
                                       (ez_color)at(i),
                                       at(i).k,
                                       at(i).c
                                     );
            }
        }
        //--------------------------------------------------------------------
    }
    return EZ_IN_BOUNDS;
}

//############################################################################
int ez_ilda_frame::scale_selection_around_point(ez_3D_float m, ez_3D_float p)
{
    if(number_of_points() > 1)
    {
        ez_3D_float     f;

        int             i,
                        out_of_bounds = EZ_IN_BOUNDS;

        for(i = 0; i < (int)size(); i++)
        {
            if(is_point_selected(i))
            {
                f = at(i).position();
                out_of_bounds = ez_ilda_bounds_check(((f - p) * m) + p, EZ_ILDA_CUBE);
                if(out_of_bounds)
                    return out_of_bounds;
            }
        }
        //--------------------------------------------------------------------
        for(i = 0; i < (int)size(); i++)
        {
            if(is_point_selected(i))
            {
                f = at(i).position();
                at(i) = ez_ilda_point( ((f - p) * m) + p,
                                       (ez_color)at(i),
                                       at(i).k,
                                       at(i).c
                                     );
            }
        }
        //--------------------------------------------------------------------
    }
    return EZ_IN_BOUNDS;
}

//############################################################################
int ez_ilda_frame::scale_selection_around_arbitrary_point(ez_3D_float m)
{
    if(number_of_points() > 1)
    {
        ez_3D_float     f;

        int             i,
                        out_of_bounds = EZ_IN_BOUNDS;

        for(i = 0; i < (int)size(); i++)
        {
            if(is_point_selected(i))
            {
                f = at(i).position();
                out_of_bounds = ez_ilda_bounds_check(((f - ez_ilda_pointer->arbitrary_point) * m) + ez_ilda_pointer->arbitrary_point, EZ_ILDA_CUBE);
                if(out_of_bounds)
                    return out_of_bounds;
            }
        }
        //--------------------------------------------------------------------
        for(i = 0; i < (int)size(); i++)
        {
            if(is_point_selected(i))
            {
                f = at(i).position();
                at(i) = ez_ilda_point( ((f - ez_ilda_pointer->arbitrary_point) * m) + ez_ilda_pointer->arbitrary_point,
                                       (ez_color)at(i),
                                       at(i).k,
                                       at(i).c
                                     );
            }
        }
        //--------------------------------------------------------------------
    }
    return EZ_IN_BOUNDS;
}

//############################################################################
int ez_ilda_frame::rotate_selection(ez_3D_float a)
{
    if(number_of_points() > 1)
    {
        ez_3D_float     f,
                        center = selected_segment().find_center();

        int             i,
                        out_of_bounds = EZ_IN_BOUNDS;

        for(i = 0; i < (int)size(); i++)
        {
            if(is_point_selected(i))
            {
                f = at(i).position();
                out_of_bounds = ez_ilda_bounds_check(rotate_point_around_point(f, center, a), EZ_ILDA_CUBE);
                if(out_of_bounds)
                    return out_of_bounds;
            }
        }
        //--------------------------------------------------------------------
        for(i = 0; i < (int)size(); i++)
        {
            if(is_point_selected(i))
            {
                f = at(i).position();
                at(i) = ez_ilda_point( rotate_point_around_point(f, center, a),
                                       (ez_color)at(i),
                                       at(i).k,
                                       at(i).c
                                     );
            }
        }
        //--------------------------------------------------------------------
    }
    return EZ_IN_BOUNDS;
}

//############################################################################
int ez_ilda_frame::rotate_selection_around_point(ez_3D_float a, ez_3D_float p)
{
    if(number_of_points() > 1)
    {
        ez_3D_float     f;

        int             i,
                        out_of_bounds = EZ_IN_BOUNDS;

        for(i = 0; i < (int)size(); i++)
        {
            if(is_point_selected(i))
            {
                f = at(i).position();
                out_of_bounds = ez_ilda_bounds_check(rotate_point_around_point(f, p, a), EZ_ILDA_CUBE);
                if(out_of_bounds)
                    return out_of_bounds;
            }
        }
        //--------------------------------------------------------------------
        for(i = 0; i < (int)size(); i++)
        {
            if(is_point_selected(i))
            {
                f = at(i).position();
                at(i) = ez_ilda_point( rotate_point_around_point(f, p, a),
                                       (ez_color)at(i),
                                       at(i).k,
                                       at(i).c
                                     );
            }
        }
        //--------------------------------------------------------------------
    }
    return EZ_IN_BOUNDS;
}

//############################################################################
int ez_ilda_frame::rotate_selection_around_arbitrary_point(ez_3D_float a)
{
    if(number_of_points() > 1)
    {
        ez_3D_float     f;

        int             i,
                        out_of_bounds = EZ_IN_BOUNDS;

        for(i = 0; i < (int)size(); i++)
        {
            if(is_point_selected(i))
            {
                f = at(i).position();
                out_of_bounds = ez_ilda_bounds_check(rotate_point_around_point(f, ez_ilda_pointer->arbitrary_point, a), EZ_ILDA_CUBE);
                if(out_of_bounds)
                    return out_of_bounds;
            }
        }
        //--------------------------------------------------------------------
        for(i = 0; i < (int)size(); i++)
        {
            if(is_point_selected(i))
            {
                f = at(i).position();
                at(i) = ez_ilda_point( rotate_point_around_point(f, ez_ilda_pointer->arbitrary_point, a),
                                       (ez_color)at(i),
                                       at(i).k,
                                       at(i).c
                                     );
            }
        }
        //--------------------------------------------------------------------
    }
    return EZ_IN_BOUNDS;
}

//############################################################################
void ez_ilda_frame::next_segment_select()
{
    int i, start, end;
    find_segment(point_select_start, start, end, i);
    find_segment(i + 1, point_select_start, point_select_end);
    return;
}

//############################################################################
void ez_ilda_frame::next_segment_select_start()
{
    int i, start, end;
    find_segment(point_select_start, start, end, i);
    find_segment(i + 1, point_select_start, end);
    return;
}

//############################################################################
void ez_ilda_frame::next_segment_select_end()
{
    int i, start, end;
    find_segment(point_select_end, start, end, i);
    find_segment(i + 1, start, point_select_end);
    return;
}

//############################################################################
void ez_ilda_frame::previous_segment_select()
{
    int i, start, end;
    find_segment(point_select_start, start, end, i);
    find_segment(i - 1, point_select_start, point_select_end);
    return;
}

//############################################################################
void ez_ilda_frame::previous_segment_select_start()
{
    int i, start, end;
    find_segment(point_select_start, start, end, i);
    find_segment(i - 1, point_select_start, end);
    return;
}

//############################################################################
void ez_ilda_frame::previous_segment_select_end()
{
    int i, start, end;
    find_segment(point_select_end, start, end, i);
    find_segment(i - 1, start, point_select_end);
    return;
}

//############################################################################
void ez_ilda_frame::set_points_color_index()
{
    if(quantity > 1)
    {
        int i;
        //--------------------------------------------------------------------                        
        if(palette_index != EZ_TRUE_COLOR)
        {
            for(i = 0; i < number_of_points(); i++)
                if(is_point_selected(i))
                {
                    at(i).c = ez_ilda_pointer->selected_palette_color;
                    at(i).r = ez_ilda_pointer->palette_picker(palette_index)[ez_ilda_pointer->selected_palette_color].r;
                    at(i).g = ez_ilda_pointer->palette_picker(palette_index)[ez_ilda_pointer->selected_palette_color].g;
                    at(i).b = ez_ilda_pointer->palette_picker(palette_index)[ez_ilda_pointer->selected_palette_color].b;
                }
        }
        //--------------------------------------------------------------------                        
        else
        {
            for(i = 0; i < number_of_points(); i++)
                if(is_point_selected(i))
                {
                    at(i).r = ez_ilda_pointer->palette_picker(palette_index)[ez_ilda_pointer->selected_palette_color].r;
                    at(i).g = ez_ilda_pointer->palette_picker(palette_index)[ez_ilda_pointer->selected_palette_color].g;
                    at(i).b = ez_ilda_pointer->palette_picker(palette_index)[ez_ilda_pointer->selected_palette_color].b;
                }            
        }
        //--------------------------------------------------------------------                        
    }
    return;
}

//############################################################################
void ez_ilda_frame::blank_points()
{
    if(quantity > 1)
    {
        int i;
        //--------------------------------------------------------------------
        for(i = 0; i < number_of_points(); i++)
            if(is_point_selected(i))
                at(i).blank();
        //--------------------------------------------------------------------
    }
    return;
}

//############################################################################
void ez_ilda_frame::unblank_points()
{
    if(quantity > 1)
    {
        int  i;
        bool anchor_left_blank = false;
        //--------------------------------------------------------------------
        for(i = 0; i < number_of_points(); i++)
            if(is_point_selected(i))
            {
                if(anchor_left_blank)
                    at(i).unblank();
                anchor_left_blank = true;
            }
        //--------------------------------------------------------------------
    }
    return;
}

//############################################################################
void ez_ilda_frame::unblank_all_points()
{
    if(quantity > 1)
    {
        int  i;
        for(i = 1; i < number_of_points(); i++)
            at(i).unblank();
    }
    return;
}

//############################################################################
ez_ilda_frame& ez_ilda_frame::add_new_point(int point_index)
{
    ez_ilda_segment::add_new_point(point_index);
    quantity = (u_short)number_of_points();
    if(    quantity > 2
        && point_index >= 0
        && point_index <  quantity
      )
    {
        if((at(point_index) | at(point_index + 1)) > ez_ilda_pointer->insignificant_distance)
            point_select_start = point_index + 1;
        else
            point_select_start = point_index;
    }
    else
    {
        point_select_start = point_select_end = point_index = 0;
    }
    return *this;
}

//############################################################################
ez_ilda_frame& ez_ilda_frame::break_segment()
{
    ez_ilda_segment::break_segment(point_select_start);
    quantity = (u_short)number_of_points();
    return *this;
}

//############################################################################
ez_ilda_frame& ez_ilda_frame::connect_the_dots()
{
    ez_ilda_segment::connect_the_dots(point_select_start, point_select_end);
    quantity = (u_short)number_of_points();
    return *this;
}

//############################################################################
void ez_ilda_frame::to_ofstream(ofstream& out)
{
    char number[7];
    //quantity = (u_short)quantity;
    if(palette_index == EZ_TRUE_COLOR)
    {
        ez_ilda_header  header  = (ez_ilda_header)*this;
        ez_palette      palette = as_ez_palette();
        header.identity--;
        sprintf(number, "%06d", header.identity);
        header.format    = EZ_ILDA_TRUE_COLOR_DATA;
        header.name      = "ES" + string(number);
        palette.to_ofstream(out, header);
    }
    else
    {
        sprintf(number, "%06d", identity);
        name = "ES" + string(number);
    }
    ez_ilda_header::to_ofstream(out);
    for(int i = 0; i < (int)quantity; i++)
        at(i).to_ofstream(out, format);
    return;
}

//############################################################################
void ez_ilda_frame::render_in_ezfb_bitmap(string ui_message) const
{
    u_short      i;

    ez_3D_float  _0,
                 _1;

    u_int        select_start_color,
                 select_end_color;

    int          xx,
                 yy,
                 x_off       = BMP.yres + EZ_STATS_DISPLAY_OFFSET * ez_ilda_pointer->font_size_factor,
                 y_off       =   EZ_PALETTE_MAX * ez_ilda_pointer->font_size_factor
                               + (EZ_STATS_DISPLAY_OFFSET + 24) * ez_ilda_pointer->font_size_factor,
                 y_line      = 8 * ez_ilda_pointer->font_size_factor,
                 bmp_center  = BMP.yres / 2;

    float        scale       = BMP.yres / 65536.0;

    bool         rotate_view = (ez_ilda_pointer->view_angle  != 0.0),
                 offset_view = (ez_ilda_pointer->view_offset != 0.0),
                 scale_view  = (ez_ilda_pointer->view_scale  != 1.0);
    //------------------------------------------------------------------------
    bmp_copy(&BMP, &BACKGROUND);
    //------------------------------------------------------------------------
    if(ez_ilda_pointer->show_fixed_bounds)
        bmp_put_rectangle(&BMP, 0, 0, BMP.yres - 1, BMP.yres - 1, BMP.gray, 0);
    //------------------------------------------------------------------------
    if(ez_ilda_pointer->show_fixed_origin)
    {
        bmp_put_line_fast(&BMP, 0, BMP.yres / 2, BMP.yres - 1, BMP.yres / 2, BMP.gray);
        bmp_put_line_fast(&BMP, BMP.yres / 2, 0, BMP.yres / 2, BMP.yres - 1, BMP.gray);
    }
    //------------------------------------------------------------------------
    if(ez_ilda_pointer->show_floating_axis)
    {
        ez_3D_float   x1(EZ_MIN_SHORT, 0, 0)
                    , x2(EZ_MAX_SHORT, 0, 0)
                    , y1(0, EZ_MIN_SHORT, 0)
                    , y2(0, EZ_MAX_SHORT, 0)
                    , z1(0, 0, EZ_MIN_SHORT)
                    , z2(0, 0, EZ_MAX_SHORT);

        if(rotate_view)
        {
            x1 = rotate_point(x1, ez_ilda_pointer->view_angle);
            x2 = rotate_point(x2, ez_ilda_pointer->view_angle);
            y1 = rotate_point(y1, ez_ilda_pointer->view_angle);
            y2 = rotate_point(y2, ez_ilda_pointer->view_angle);
            z1 = rotate_point(z1, ez_ilda_pointer->view_angle);
            z2 = rotate_point(z2, ez_ilda_pointer->view_angle);
        }
        if(offset_view)
        {
            x1 += ez_ilda_pointer->view_offset;
            x2 += ez_ilda_pointer->view_offset;
            y1 += ez_ilda_pointer->view_offset;
            y2 += ez_ilda_pointer->view_offset;
            z1 += ez_ilda_pointer->view_offset;
            z2 += ez_ilda_pointer->view_offset;
        }
        if(scale_view)
        {
            x1 *= ez_ilda_pointer->view_scale;
            x2 *= ez_ilda_pointer->view_scale;
            y1 *= ez_ilda_pointer->view_scale;
            y2 *= ez_ilda_pointer->view_scale;
            z1 *= ez_ilda_pointer->view_scale;
            z2 *= ez_ilda_pointer->view_scale;
        }
        // X axis
        bmp_put_line_fast(&BMP,             (int)(x1.x * scale) + bmp_center
                              , BMP.yres - ((int)(x1.y * scale) + bmp_center)
                              ,             (int)(x2.x * scale) + bmp_center
                              , BMP.yres - ((int)(x2.y * scale) + bmp_center)
                              , BMP.gray
                         );
        // X axis icon
        bmp_put_line_fast(&BMP,             (int)((x2.x - 500) * scale) + bmp_center
                              , BMP.yres - ((int)((x2.y - 500) * scale) + bmp_center)
                              ,             (int)((x2.x + 500) * scale) + bmp_center
                              , BMP.yres - ((int)((x2.y + 500) * scale) + bmp_center)
                              , BMP.make_rgb(255,0,0)
                         );
        bmp_put_line_fast(&BMP,             (int)((x2.x - 500) * scale) + bmp_center
                              , BMP.yres - ((int)((x2.y + 500) * scale) + bmp_center)
                              ,             (int)((x2.x + 500) * scale) + bmp_center
                              , BMP.yres - ((int)((x2.y - 500) * scale) + bmp_center)
                              , BMP.make_rgb(255,0,0)
                         );
        // Y axis
        bmp_put_line_fast(&BMP,             (int)(y1.x * scale) + bmp_center
                              , BMP.yres - ((int)(y1.y * scale) + bmp_center)
                              ,             (int)(y2.x * scale) + bmp_center
                              , BMP.yres - ((int)(y2.y * scale) + bmp_center)
                              , BMP.gray
                         );
        // Y axis icon
        bmp_put_line_fast(&BMP,             (int)((y2.x - 500) * scale) + bmp_center
                              , BMP.yres - ((int)((y2.y + 500) * scale) + bmp_center)
                              ,             (int)((y2.x       ) * scale) + bmp_center
                              , BMP.yres - ((int)((y2.y       ) * scale) + bmp_center)
                              , BMP.make_rgb(0,255,0)
                         );
        bmp_put_line_fast(&BMP,             (int)((y2.x + 500) * scale) + bmp_center
                              , BMP.yres - ((int)((y2.y + 500) * scale) + bmp_center)
                              ,             (int)((y2.x       ) * scale) + bmp_center
                              , BMP.yres - ((int)((y2.y       ) * scale) + bmp_center)
                              , BMP.make_rgb(0,255,0)
                         );
        bmp_put_line_fast(&BMP,             (int)((y2.x       ) * scale) + bmp_center
                              , BMP.yres - ((int)((y2.y       ) * scale) + bmp_center)
                              ,             (int)((y2.x       ) * scale) + bmp_center
                              , BMP.yres - ((int)((y2.y - 500) * scale) + bmp_center)
                              , BMP.make_rgb(0,255,0)
                         );
        // Z axis
        bmp_put_line_fast(&BMP,             (int)(z1.x * scale) + bmp_center
                              , BMP.yres - ((int)(z1.y * scale) + bmp_center)
                              ,             (int)(z2.x * scale) + bmp_center
                              , BMP.yres - ((int)(z2.y * scale) + bmp_center)
                              , BMP.gray
                         );
        // Z axis icon
        bmp_put_line_fast(&BMP,             (int)((z2.x - 500) * scale) + bmp_center
                              , BMP.yres - ((int)((z2.y + 500) * scale) + bmp_center)
                              ,             (int)((z2.x + 500) * scale) + bmp_center
                              , BMP.yres - ((int)((z2.y + 500) * scale) + bmp_center)
                              , BMP.make_rgb(0,0,255)
                         );
        bmp_put_line_fast(&BMP,             (int)((z2.x + 500) * scale) + bmp_center
                              , BMP.yres - ((int)((z2.y + 500) * scale) + bmp_center)
                              ,             (int)((z2.x - 500) * scale) + bmp_center
                              , BMP.yres - ((int)((z2.y - 500) * scale) + bmp_center)
                              , BMP.make_rgb(0,0,255)
                         );
        bmp_put_line_fast(&BMP,             (int)((z2.x - 500) * scale) + bmp_center
                              , BMP.yres - ((int)((z2.y - 500) * scale) + bmp_center)
                              ,             (int)((z2.x + 500) * scale) + bmp_center
                              , BMP.yres - ((int)((z2.y - 500) * scale) + bmp_center)
                              , BMP.make_rgb(0,0,255)
                         );
    }
    //------------------------------------------------------------------------
    if(ez_ilda_pointer->show_floating_bounds)
    {
        ez_3D_float   a(EZ_MIN_SHORT, EZ_MIN_SHORT, EZ_MIN_SHORT)
                    , b(EZ_MAX_SHORT, EZ_MIN_SHORT, EZ_MIN_SHORT)
                    , c(EZ_MAX_SHORT, EZ_MIN_SHORT, EZ_MAX_SHORT)
                    , d(EZ_MIN_SHORT, EZ_MIN_SHORT, EZ_MAX_SHORT)
                    , e(EZ_MIN_SHORT, EZ_MAX_SHORT, EZ_MIN_SHORT)
                    , f(EZ_MAX_SHORT, EZ_MAX_SHORT, EZ_MIN_SHORT)
                    , g(EZ_MAX_SHORT, EZ_MAX_SHORT, EZ_MAX_SHORT)
                    , h(EZ_MIN_SHORT, EZ_MAX_SHORT, EZ_MAX_SHORT);

        if(rotate_view)
        {
            a = rotate_point(a, ez_ilda_pointer->view_angle);
            b = rotate_point(b, ez_ilda_pointer->view_angle);
            c = rotate_point(c, ez_ilda_pointer->view_angle);
            d = rotate_point(d, ez_ilda_pointer->view_angle);
            e = rotate_point(e, ez_ilda_pointer->view_angle);
            f = rotate_point(f, ez_ilda_pointer->view_angle);
            g = rotate_point(g, ez_ilda_pointer->view_angle);
            h = rotate_point(h, ez_ilda_pointer->view_angle);
        }
        if(offset_view)
        {
            a += ez_ilda_pointer->view_offset;
            b += ez_ilda_pointer->view_offset;
            c += ez_ilda_pointer->view_offset;
            d += ez_ilda_pointer->view_offset;
            e += ez_ilda_pointer->view_offset;
            f += ez_ilda_pointer->view_offset;
            g += ez_ilda_pointer->view_offset;
            h += ez_ilda_pointer->view_offset;
        }
        if(scale_view)
        {
            a *= ez_ilda_pointer->view_scale;
            b *= ez_ilda_pointer->view_scale;
            c *= ez_ilda_pointer->view_scale;
            d *= ez_ilda_pointer->view_scale;
            e *= ez_ilda_pointer->view_scale;
            f *= ez_ilda_pointer->view_scale;
            g *= ez_ilda_pointer->view_scale;
            h *= ez_ilda_pointer->view_scale;
        }

        a *= scale;
        b *= scale;
        c *= scale;
        d *= scale;
        e *= scale;
        f *= scale;
        g *= scale;
        h *= scale;

        bmp_put_line_fast(&BMP,             (int)a.x + bmp_center
                              , BMP.yres - ((int)a.y + bmp_center)
                              ,             (int)b.x + bmp_center
                              , BMP.yres - ((int)b.y + bmp_center)
                              , BMP.gray
                         );
        bmp_put_line_fast(&BMP,             (int)b.x + bmp_center
                              , BMP.yres - ((int)b.y + bmp_center)
                              ,             (int)c.x + bmp_center
                              , BMP.yres - ((int)c.y + bmp_center)
                              , BMP.gray
                         );
        bmp_put_line_fast(&BMP,             (int)c.x + bmp_center
                              , BMP.yres - ((int)c.y + bmp_center)
                              ,             (int)d.x + bmp_center
                              , BMP.yres - ((int)d.y + bmp_center)
                              , BMP.gray
                         );
        bmp_put_line_fast(&BMP,             (int)d.x + bmp_center
                              , BMP.yres - ((int)d.y + bmp_center)
                              ,             (int)a.x + bmp_center
                              , BMP.yres - ((int)a.y + bmp_center)
                              , BMP.gray
                         );
        bmp_put_line_fast(&BMP,             (int)a.x + bmp_center
                              , BMP.yres - ((int)a.y + bmp_center)
                              ,             (int)e.x + bmp_center
                              , BMP.yres - ((int)e.y + bmp_center)
                              , BMP.gray
                         );
        bmp_put_line_fast(&BMP,             (int)b.x + bmp_center
                              , BMP.yres - ((int)b.y + bmp_center)
                              ,             (int)f.x + bmp_center
                              , BMP.yres - ((int)f.y + bmp_center)
                              , BMP.gray
                         );
        bmp_put_line_fast(&BMP,             (int)c.x + bmp_center
                              , BMP.yres - ((int)c.y + bmp_center)
                              ,             (int)g.x + bmp_center
                              , BMP.yres - ((int)g.y + bmp_center)
                              , BMP.gray
                         );
        bmp_put_line_fast(&BMP,             (int)d.x + bmp_center
                              , BMP.yres - ((int)d.y + bmp_center)
                              ,             (int)h.x + bmp_center
                              , BMP.yres - ((int)h.y + bmp_center)
                              , BMP.gray
                         );
        bmp_put_line_fast(&BMP,             (int)e.x + bmp_center
                              , BMP.yres - ((int)e.y + bmp_center)
                              ,             (int)f.x + bmp_center
                              , BMP.yres - ((int)f.y + bmp_center)
                              , BMP.gray
                         );
        bmp_put_line_fast(&BMP,             (int)f.x + bmp_center
                              , BMP.yres - ((int)f.y + bmp_center)
                              ,             (int)g.x + bmp_center
                              , BMP.yres - ((int)g.y + bmp_center)
                              , BMP.gray
                         );
        bmp_put_line_fast(&BMP,             (int)g.x + bmp_center
                              , BMP.yres - ((int)g.y + bmp_center)
                              ,             (int)h.x + bmp_center
                              , BMP.yres - ((int)h.y + bmp_center)
                              , BMP.gray
                         );
        bmp_put_line_fast(&BMP,             (int)h.x + bmp_center
                              , BMP.yres - ((int)h.y + bmp_center)
                              ,             (int)e.x + bmp_center
                              , BMP.yres - ((int)e.y + bmp_center)
                              , BMP.gray
                         );
    }
    //------------------------------------------------------------------------
    if(ez_ilda_pointer->show_arbitrary_point)
    {
        ez_3D_float   x1(-5000, 0, 0)
                    , x2( 5000, 0, 0)
                    , y1(0, -5000, 0)
                    , y2(0,  5000, 0)
                    , z1(0, 0, -5000)
                    , z2(0, 0,  5000);

        x1 = x1 += ez_ilda_pointer->arbitrary_point;
        x2 = x2 += ez_ilda_pointer->arbitrary_point;
        y1 = y1 += ez_ilda_pointer->arbitrary_point;
        y2 = y2 += ez_ilda_pointer->arbitrary_point;
        z1 = z1 += ez_ilda_pointer->arbitrary_point;
        z2 = z2 += ez_ilda_pointer->arbitrary_point;

        if(rotate_view)
        {
            x1 = rotate_point(x1, ez_ilda_pointer->view_angle);
            x2 = rotate_point(x2, ez_ilda_pointer->view_angle);
            y1 = rotate_point(y1, ez_ilda_pointer->view_angle);
            y2 = rotate_point(y2, ez_ilda_pointer->view_angle);
            z1 = rotate_point(z1, ez_ilda_pointer->view_angle);
            z2 = rotate_point(z2, ez_ilda_pointer->view_angle);
        }
        if(offset_view)
        {
            x1 += ez_ilda_pointer->view_offset;
            x2 += ez_ilda_pointer->view_offset;
            y1 += ez_ilda_pointer->view_offset;
            y2 += ez_ilda_pointer->view_offset;
            z1 += ez_ilda_pointer->view_offset;
            z2 += ez_ilda_pointer->view_offset;
        }
        if(scale_view)
        {
            x1 *= ez_ilda_pointer->view_scale;
            x2 *= ez_ilda_pointer->view_scale;
            y1 *= ez_ilda_pointer->view_scale;
            y2 *= ez_ilda_pointer->view_scale;
            z1 *= ez_ilda_pointer->view_scale;
            z2 *= ez_ilda_pointer->view_scale;
        }
        bmp_put_line_fast(&BMP,             (int)(x1.x * scale) + bmp_center
                              , BMP.yres - ((int)(x1.y * scale) + bmp_center)
                              ,             (int)(x2.x * scale) + bmp_center
                              , BMP.yres - ((int)(x2.y * scale) + bmp_center)
                              , BMP.make_rgb(255,0,0)
                         );
        bmp_put_line_fast(&BMP,             (int)(y1.x * scale) + bmp_center
                              , BMP.yres - ((int)(y1.y * scale) + bmp_center)
                              ,             (int)(y2.x * scale) + bmp_center
                              , BMP.yres - ((int)(y2.y * scale) + bmp_center)
                              , BMP.make_rgb(0,255,0)
                         );
        bmp_put_line_fast(&BMP,             (int)(z1.x * scale) + bmp_center
                              , BMP.yres - ((int)(z1.y * scale) + bmp_center)
                              ,             (int)(z2.x * scale) + bmp_center
                              , BMP.yres - ((int)(z2.y * scale) + bmp_center)
                              , BMP.make_rgb(0,0,255)
                         );
    }
    //------------------------------------------------------------------------
    if(ez_ilda_pointer->show_palettes)
        ez_ilda_pointer->render_palette_in_ezfb_bitmap();
    //------------------------------------------------------------------------
    if(ez_ilda_pointer->show_stats)
    {
        if(is_selected)
            bmp_printf(   &BMP
                        , x_off
                        , y_off
                        , BMP.make_rgb(255,255,0)
                        , BMP.make_rgb(128,0,0)
                        , 0
                        , ez_ilda_pointer->font_size_factor
                        , "frame    %6d"
                        , ez_ilda_pointer->frame_index
                      );
        else
            bmp_printf(   &BMP
                        , x_off
                        , y_off
                        , BMP.make_rgb(255,255,0)
                        , BMP.black
                        , 0
                        , ez_ilda_pointer->font_size_factor
                        , "frame    %6d"
                        , ez_ilda_pointer->frame_index
                      );
        //------------------------------------------------------------------------
        y_off += y_line;
        if(is_unique)
            bmp_printf(   &BMP
                        , x_off
                        , y_off
                        , BMP.make_rgb(0,255,255)
                        , BMP.black
                        , 0
                        , ez_ilda_pointer->font_size_factor
                        , "vertices %6d"
                        , quantity
                      );
        else
            bmp_printf(   &BMP
                        , x_off
                        , y_off
                        , BMP.make_rgb(0,255,255)
                        , BMP.make_rgb(0,128,0)
                        , 0
                        , ez_ilda_pointer->font_size_factor
                        , "vertices %6d"
                        , quantity
                      );
        //------------------------------------------------------------------------
        y_off += y_line;
        bmp_printf(   &BMP
                    , x_off
                    , y_off
                    , BMP.make_rgb(255,0,255)
                    , BMP.black
                    , 0
                    , ez_ilda_pointer->font_size_factor
                    , "segments %6d"
                    , number_of_segments()
                  );
    }
    //------------------------------------------------------------------------
    if(quantity > 1)
    {
        //--------------------------------------------------------------------
        if(ez_ilda_pointer->show_blanking)
        {
            for(i = 1; i < quantity; i++)
            {
                if(at(i).is_blank())
                {
                    if(rotate_view)
                    {
                        _0 = rotate_point(at(i - 1), ez_ilda_pointer->view_angle);
                        _1 = rotate_point(at(i    ), ez_ilda_pointer->view_angle);
                    }
                    else
                    {
                        _0 = at(i - 1);
                        _1 = at(i    );
                    }
                    if(offset_view)
                    {
                        _0 += ez_ilda_pointer->view_offset;
                        _1 += ez_ilda_pointer->view_offset;
                    }
                    if(scale_view)
                    {
                        _0 *= ez_ilda_pointer->view_scale;
                        _1 *= ez_ilda_pointer->view_scale;
                    }
                    xx = (int)(_1.x * scale) + bmp_center;
                    yy = (int)(_1.y * scale) + bmp_center;

                    bmp_put_line_fast(&BMP,             (int)(_0.x * scale) + bmp_center
                                          , BMP.yres - ((int)(_0.y * scale) + bmp_center)
                                          ,             xx
                                          , BMP.yres - (yy)
                                          , BMP.gray
                                     );
                    if(ez_ilda_pointer->show_points)
                        bmp_put_rectangle(&BMP,             xx - 1
                                              , BMP.yres - (yy - 1)
                                              ,             xx + 1
                                              , BMP.yres - (yy + 1)
                                              , BMP.gray
                                              ,0
                                         );
                    if(ez_ilda_pointer->show_cursors && is_point_selected(i))
                        bmp_put_rectangle(&BMP,             xx - 3
                                              , BMP.yres - (yy - 3)
                                              ,             xx + 3
                                              , BMP.yres - (yy + 3)
                                              , BMP.gray
                                              ,0
                                         );
                }
            }
        } // end if(ez_ilda_pointer->show_blanking)
        //--------------------------------------------------------------------
        for(i = 1; i < quantity; i++)
        {            
            if(at(i).is_lit())
            {
                if(rotate_view)
                {
                    _0 = rotate_point(at(i - 1), ez_ilda_pointer->view_angle);
                    _1 = rotate_point(at(i    ), ez_ilda_pointer->view_angle);
                }
                else
                {
                    _0 = at(i - 1);
                    _1 = at(i    );
                }
                if(offset_view)
                {
                    _0 += ez_ilda_pointer->view_offset;
                    _1 += ez_ilda_pointer->view_offset;
                }
                if(scale_view)
                {
                    _0 *= ez_ilda_pointer->view_scale;
                    _1 *= ez_ilda_pointer->view_scale;
                }
                xx = (int)(_1.x * scale) + bmp_center;
                yy = (int)(_1.y * scale) + bmp_center;
                //------------------------------------------------------------
                if(at(i).color())
                {
                    bmp_put_line_fast(&BMP,             (int)(_0.x * scale) + bmp_center
                                          , BMP.yres - ((int)(_0.y * scale) + bmp_center)
                                          ,             xx
                                          , BMP.yres - (yy)
                                          , BMP.make_rgb(at(i).r, at(i).g, at(i).b)
                                     );
                    if(ez_ilda_pointer->show_points)
                        bmp_put_rectangle(&BMP,             xx - 1
                                              , BMP.yres - (yy - 1)
                                              ,             xx + 1
                                              , BMP.yres - (yy + 1)
                                              , BMP.make_rgb(at(i).r, at(i).g, at(i).b)
                                              ,0
                                         );
                }
                //------------------------------------------------------------
                if(ez_ilda_pointer->show_cursors && is_point_selected(i))
                    bmp_put_rectangle(&BMP,             xx - 3
                                          , BMP.yres - (yy - 3)
                                          ,             xx + 3
                                          , BMP.yres - (yy + 3)
                                          ,   at(i).is_black()
                                            ? BMP.white
                                            : BMP.make_rgb(at(i).r,
                                                           at(i).g,
                                                           at(i).b
                                                          )
                                          ,0
                                     );
            }
        }
        //--------------------------------------------------------------------
        if(ez_ilda_pointer->show_cursors)
        {
            select_start_color =   at(point_select_start).is_blank()
                                 ? BMP.gray
                                 :   at(point_select_start).is_black()
                                   ? BMP.white
                                   : BMP.make_rgb(at(point_select_start).r,
                                                  at(point_select_start).g,
                                                  at(point_select_start).b
                                                 );
            //----------------------------------------------------------------
            select_end_color   =   at(point_select_end).is_blank()
                                 ? BMP.gray
                                 :   at(point_select_end).is_black()
                                   ? BMP.white
                                   : BMP.make_rgb(at(point_select_end).r,
                                                  at(point_select_end).g,
                                                  at(point_select_end).b
                                                 );
            //----------------------------------------------------------------
            if(rotate_view)
            {
                _0 = rotate_point(at(point_select_start), ez_ilda_pointer->view_angle);
                _1 = rotate_point(at(point_select_end  ), ez_ilda_pointer->view_angle);
            }
            else
            {
                _0 = at(point_select_start);
                _1 = at(point_select_end  );
            }
            if(offset_view)
            {
                _0 += ez_ilda_pointer->view_offset;
                _1 += ez_ilda_pointer->view_offset;
            }
            if(scale_view)
            {
                _0 *= ez_ilda_pointer->view_scale;
                _1 *= ez_ilda_pointer->view_scale;
            }
            //----------------------------------------------------------------
            // start point select cursor
            bmp_put_rectangle(&BMP,             (int)((_0.x - 1000) * scale) + bmp_center
                                  , BMP.yres - ((int)((_0.y - 1000) * scale) + bmp_center)
                                  ,             (int)((_0.x + 1000) * scale) + bmp_center
                                  , BMP.yres - ((int)((_0.y + 1000) * scale) + bmp_center)
                                  , select_start_color
                                  , 0
                             );
            //----------------------------------------------------------------
            // end point select cursor
            bmp_put_line_fast     (&BMP,             (int)((_1.x       ) * scale) + bmp_center
                                       , BMP.yres - ((int)((_1.y - 1000) * scale) + bmp_center)
                                       ,             (int)((_1.x       ) * scale) + bmp_center
                                       , BMP.yres - ((int)((_1.y + 1000) * scale) + bmp_center)
                                       , select_end_color
                                  );
            bmp_put_line_fast     (&BMP,             (int)((_1.x + 1000) * scale) + bmp_center
                                       , BMP.yres - ((int)((_1.y       ) * scale) + bmp_center)
                                       ,             (int)((_1.x - 1000) * scale) + bmp_center
                                       , BMP.yres - ((int)((_1.y       ) * scale) + bmp_center)
                                       , select_end_color
                                  );
            bmp_put_line_fast     (&BMP,             (int)((_1.x - 1000) * scale) + bmp_center
                                       , BMP.yres - ((int)((_1.y - 1000) * scale) + bmp_center)
                                       ,             (int)((_1.x + 1000) * scale) + bmp_center
                                       , BMP.yres - ((int)((_1.y + 1000) * scale) + bmp_center)
                                       , select_end_color
                                  );
            bmp_put_line_fast     (&BMP,             (int)((_1.x + 1000) * scale) + bmp_center
                                       , BMP.yres - ((int)((_1.y - 1000) * scale) + bmp_center)
                                       ,             (int)((_1.x - 1000) * scale) + bmp_center
                                       , BMP.yres - ((int)((_1.y + 1000) * scale) + bmp_center)
                                       , select_end_color
                                  );
            //----------------------------------------------------------------
            // icons
            y_off += (2 * y_line);
            bmp_printf(   &BMP
                        , x_off
                        , y_off
                        , BMP.make_rgb(255,0,255)
                        , BMP.black
                        , 0
                        , ez_ilda_pointer->font_size_factor
                        , "           %4d"
                        , segment_index_of_point(point_select_start)
                      );
            bmp_printf(   &BMP
                        , x_off
                        , y_off
                        , BMP.make_rgb(0,255,255)
                        , BMP.black
                        , 0
                        , ez_ilda_pointer->font_size_factor
                        , "      %4d"
                        , point_select_start
                      );
            bmp_printf(   &BMP
                        , x_off
                        , y_off
                        , BMP.white
                        , BMP.black
                        , 0
                        , ez_ilda_pointer->font_size_factor
                        , "  %3d"
                        , at(point_select_start).c
                      );
            //----------------------------------------------------------------
            bmp_put_rectangle(&BMP, x_off
                                  , y_off
                                  , x_off + (int)(2000 * scale)
                                  , y_off + (int)(2000 * scale)
                                  , select_start_color
                                  , 0
                             ); // start cursor icon
            //-----------------------------------------------------------------
            y_off += (3 * y_line);
            bmp_printf(   &BMP
                        , x_off
                        , y_off
                        , BMP.make_rgb(255,0,255)
                        , BMP.black
                        , 0
                        , ez_ilda_pointer->font_size_factor
                        , "           %4d"
                        , segment_index_of_point(point_select_end)
                      );
            bmp_printf(   &BMP
                        , x_off
                        , y_off
                        , BMP.make_rgb(0,255,255)
                        , BMP.black
                        , 0
                        , ez_ilda_pointer->font_size_factor
                        , "      %4d"
                        , point_select_end
                      );
            bmp_printf(   &BMP
                        , x_off
                        , y_off
                        , BMP.white
                        , BMP.black
                        , 0
                        , ez_ilda_pointer->font_size_factor
                        , "  %3d"
                        , at(point_select_end).c
                      );
            //----------------------------------------------------------------
            bmp_put_line_fast     (&BMP, x_off + (int)(1000 * scale)
                                       , y_off
                                       , x_off + (int)(1000 * scale)
                                       , y_off + (int)(2000 * scale)
                                       , select_end_color
                                  );
            bmp_put_line_fast     (&BMP, x_off + (int)(2000 * scale)
                                       , y_off + (int)(1000 * scale)
                                       , x_off
                                       , y_off + (int)(1000 * scale)
                                       , select_end_color
                                  );
            bmp_put_line_fast     (&BMP, x_off
                                       , y_off
                                       , x_off + (int)(2000 * scale)
                                       , y_off + (int)(2000 * scale)
                                       , select_end_color
                                  );
            bmp_put_line_fast     (&BMP, x_off + (int)(2000 * scale)
                                       , y_off
                                       , x_off
                                       , y_off + (int)(2000 * scale)
                                       , select_end_color
                                  ); // end cursor icon
            //----------------------------------------------------------------
        } // end if(ez_ilda_pointer->show_cursors)
        //--------------------------------------------------------------------
    } // end if(quantity > 1)
    //------------------------------------------------------------------------
    if(    ez_ilda_pointer->show_ui_message
        && ui_message.size()
      )
    {
        y_off = BMP.yres - y_line;
        bmp_printf(   &BMP
                    , x_off
                    , y_off
                    , BMP.make_rgb(255,255,0)
                    , BMP.make_rgb(63,0,63)
                    , 0
                    , ez_ilda_pointer->font_size_factor
                    , "%s"
                    , ui_message.c_str()
                  );
    }
    return;
}

//############################################################################
void ez_ilda_frame::save_as_wave(string file) const
{
    if(    file.size() < 4
        || file.substr(file.size() - 4) != ".wav"
      )
        file += ".wav";

    ofstream out((EZ_ILDA_WAVE_SHARE + file).c_str(), ios::out | ios::binary);
    if(out.is_open())
    {
        if(quantity > 1)
        {
            ez_ilda_point    point;

            ez_ilda_segment  intro,
                             outro;

            ez_ilda_frame    copy(*this);

            copy.rotate_around_origin();

            if(ez_ilda_pointer->auto_flatten_z)
                copy.flatten_z();

            if(ez_ilda_pointer->auto_optimize)
                copy.optimize_points();

            int i,
                frame_scan,
                blank_dwell = 0,
                total_frame_scans = (int)ceil((   ez_ilda_pointer->wave_sample_rate
                                                * ez_ilda_pointer->still_frame_duration
                                              ) / copy.quantity
                                             ),
                total_wave_samples = total_frame_scans * copy.quantity;
            //--------------------------------------------------------------------
            if(ez_ilda_pointer->auto_optimize)
            {
                blank_dwell = ez_ilda_pointer->wave_sample_rate / 220;
                //----------------------------------------------------------------
                for(i = 0; i < blank_dwell; i++)
                    intro.push_back(ez_ilda_point()); // the origin
                //----------------------------------------------------------------
                point = copy.first_lit_point();
                point.blank();
                intro += ez_ilda_segment(intro[0], point); // line from origin to first point
                //----------------------------------------------------------------
                outro = ez_ilda_segment(point, intro[0]); // line from last point to origin
                for(i = 0; i < blank_dwell; i++)
                    outro.push_back(ez_ilda_point()); // the origin
                //----------------------------------------------------------------
                total_wave_samples =   intro.number_of_points()
                                     +   total_frame_scans
                                       * copy.quantity
                                     + outro.number_of_points();
            }
            //--------------------------------------------------------------------
            ez_wave_header  header(total_wave_samples, ez_ilda_pointer->wave_sample_rate);
            header.to_ofstream(out);
            //--------------------------------------------------------------------            
            if(ez_ilda_pointer->auto_optimize)
                intro.to_ofstream_wave(out, 0x00, false);
            //--------------------------------------------------------------------            
            ((ez_ilda_segment)copy).to_ofstream_wave(out, 0x01, is_unique); // first frame unique
            //--------------------------------------------------------------------            
            for(frame_scan = 1; frame_scan < total_frame_scans; frame_scan++)
                ((ez_ilda_segment)copy).to_ofstream_wave(out, 0x01, false); // the rest are not
            //--------------------------------------------------------------------            
            if(ez_ilda_pointer->auto_optimize)
                outro.to_ofstream_wave(out, 0x00, false);
            //--------------------------------------------------------------------
        }
        out.close();
    }
    return;
}

//############################################################################
void ez_ilda_frame::to_ofstream_wave(ofstream& out) const
{
    if(quantity > 1)
    {
        int frame_scan,
            total_frame_scans =   (ez_ilda_pointer->frames_per_second > 0)
                                ? (int)ceil(   (    ez_ilda_pointer->wave_sample_rate 
                                                  / quantity
                                                )
                                              / ez_ilda_pointer->frames_per_second
                                           )
                                : 1;
        //--------------------------------------------------------------------
        ez_ilda_segment::to_ofstream_wave(out, 0x01, is_unique);

        for(frame_scan = 1; frame_scan < total_frame_scans; frame_scan++)
            ez_ilda_segment::to_ofstream_wave(out, 0x01, false);
        //--------------------------------------------------------------------
        transition_to_next_frame.to_ofstream_wave(out, false, false);
    }
    //------------------------------------------------------------------------
    return;
}

//############################################################################
void ez_ilda_frame::to_ofstream_code(ofstream& out, const char* indent) const
{
    out <<  indent
        << "frame.format = "
        <<  (int)format
        <<  ";\n"
        <<  indent
        << "frame.palette_index = "
        <<  palette_index
        <<  ";\n";

    for(int i = 0; i < quantity; i++)
        out << indent
            <<  "frame.push_back(ez_ilda_point("
            << setw(6)
            <<  at(i).x
            <<  ", "
            << setw(6)
            <<  at(i).y
            <<  ", "
            << setw(6)
            <<  at(i).z
            <<  ", "
            << setw(3)
            <<  (int)at(i).r
            <<  ", "
            << setw(3)
            <<  (int)at(i).g
            <<  ", "
            << setw(3)
            <<  (int)at(i).b
            <<  ", "
            << setw(3)
            <<  (int)at(i).k
            <<  ", "
            << setw(3)
            <<  (int)at(i).c
            <<  "));\n";

    out << indent << "frame.quantity = (u_short)frame.quantity;\n";
    return;
}

//############################################################################
void ez_ilda_frame::save_as_code(string object_name) const
{
    ofstream out((EZ_ILDA_CPP_SHARE + object_name + ".cpp").c_str(), ios::out);
    if(out.is_open())
    {
        out << "//############################################################################\n"
               "// ~  This file is the product of LaserBoy!  ~\n"
               "// ~ International Laser Display Association ~\n"
               "// ~    ez_ilda file format driver in C++    ~\n"
               "//\n"
               "// by James Lehman\n"
               "// james@akrobiz.com\n"
               "//\n"
               "//############################################################################\n\n\n\n"
               "//############################################################################\n"

               "ez_ilda_frame "
            << object_name
            << "()\n"
               "{\n"
               "    ez_ilda_frame frame;\n";

        to_ofstream_code(out, "    ");

        out << "    return frame;\n"
               "}\n\n"
               "//############################################################################\n"
               "//////////////////////////////////////////////////////////////////////////////\n"
               "//############################################################################\n\n\n";

        out.close();
    }
    return;
}

//############################################################################
void ez_ilda_frame::save_as_dxf(string file) const
{
    if(    file.size() < 4
        || file.substr(file.size() - 4) != ".dxf"
      )
        file += ".dxf";

    ofstream out((EZ_ILDA_DXF_SHARE + file).c_str(), ios::out);
    if(out.is_open())
    {
        int                segment_vector_index,
                           point_index;
        ez_ilda_frame_set  segments = find_segments();

        out <<  " 0\n"
                "SECTION\n"
                " 2\n"
                "ENTITIES\n";

        for(segment_vector_index = 0; segment_vector_index < number_of_segments(); segment_vector_index++)
        {
            if(segments[segment_vector_index].number_of_lit_points() == 1)
            {
                out <<  " 0\n"
                        "LINE\n"
                        " 62\n  ";

                out <<  ez_ilda_pointer->
                        palette_picker(EZ_DXF)
                        .best_match(segments[segment_vector_index].first_lit_point());

                out <<  "\n 10\n";
                out <<  segments[segment_vector_index].first_lit_anchor().x;
                out <<  "\n 20\n";
                out <<  segments[segment_vector_index].first_lit_anchor().y;
                out <<  "\n 30\n";
                out <<  segments[segment_vector_index].first_lit_anchor().z;
                out <<  "\n 11\n";
                out <<  segments[segment_vector_index].first_lit_point().x;
                out <<  "\n 21\n";
                out <<  segments[segment_vector_index].first_lit_point().y;
                out <<  "\n 31\n";
                out <<  segments[segment_vector_index].first_lit_point().z;
                out <<  "\n";
            }
            else
            {
                out <<  " 0\n"
                        "POLYLINE\n"
                        " 70\n";
                out <<  segments[segment_vector_index].is_closed_polygon();
                out <<  "\n 10\n"
                        "0.0\n"
                        " 20\n"
                        "0.0\n"
                        " 30\n"
                        "0.0\n"
                        //---------
                        " 0\n"
                        "VERTEX\n"
                        " 62\n";
                out <<  ez_ilda_pointer->
                        palette_picker(EZ_DXF)
                        .best_match(segments[segment_vector_index].at(1));
                out <<  "\n 10\n";
                out <<  segments[segment_vector_index].first_lit_anchor().x;
                out <<  "\n 20\n";
                out <<  segments[segment_vector_index].first_lit_anchor().y;
                out <<  "\n 30\n";
                out <<  segments[segment_vector_index].first_lit_anchor().z;
                out <<  "\n";

                for(point_index = 1; point_index < segments[segment_vector_index].number_of_points() - 1; point_index++)
                {
                    out <<  " 0\n"
                        <<  "VERTEX\n"
                        <<  " 62\n";
                    out <<  ez_ilda_pointer->
                            palette_picker(EZ_DXF)
                            .best_match(segments[segment_vector_index].at(point_index + 1));
                    out <<  "\n 10\n";
                    out <<  segments[segment_vector_index].at(point_index).x;
                    out <<  "\n 20\n";
                    out <<  segments[segment_vector_index].at(point_index).y;
                    out <<  "\n 30\n";
                    out <<  segments[segment_vector_index].at(point_index).z;
                    out <<  "\n";
                }
                out <<  " 0\n"
                        "VERTEX\n"
                        " 62\n"
                        "0\n"
                        " 10\n";
                out <<  segments[segment_vector_index].at(point_index).x;
                out <<  "\n 20\n";
                out <<  segments[segment_vector_index].at(point_index).y;
                out <<  "\n 30\n";
                out <<  segments[segment_vector_index].at(point_index).z;
                out <<  "\n";
            }
        }
        out <<  " 0\n"
                "ENDSEC\n"
                " 0\n"
                "EOF\n";

        out.close();
    }
    return;
}

//############################################################################
void ez_ilda_frame::save_as_bmp(string file) const
{
    if(    file.size() < 4
        || file.substr(file.size() - 4) != ".bmp"
      )
        file += ".bmp";
    //------------------------------------------------------------------------
    if(quantity > 1)
    {
        char   file_name[80];
        struct ezfb_bitmap bmp = {0};

        if(palette_index == EZ_TRUE_COLOR)
            bmp_init(&bmp, ez_ilda_pointer->output_bmp_size, ez_ilda_pointer->output_bmp_size, 24);
        else
        {
            bmp_init(&bmp, ez_ilda_pointer->output_bmp_size, ez_ilda_pointer->output_bmp_size,  8);
            ez_ilda_pointer->palette_picker(palette_index).to_ezfb_bmp_palette(&bmp);
        }
        bmp_clear(&bmp, bmp.black);

        u_short         i;
        ez_3D_float    _0,
                       _1;
        int            bmp_center  = bmp.yres / 2;
        float          scale       = bmp.yres / 65536.0;
        //--------------------------------------------------------------------
        for(i = 1; i < quantity; i++)
        {
            if(at(i).color())
            {
                _0 = at(i - 1);
                _1 = at(i    );
                //------------------------------------------------------------

                if(palette_index == EZ_TRUE_COLOR)
                    bmp_put_line_fast(&bmp,  (int)(_0.x * scale) + bmp_center
                                          , ((int)(_0.y * scale) + bmp_center)
                                          ,  (int)(_1.x * scale) + bmp_center
                                          , ((int)(_1.y * scale) + bmp_center)
                                          , BMP.make_rgb(at(i).r, at(i).g, at(i).b)
                                     );
                else
                    bmp_put_line_fast(&bmp,  (int)(_0.x * scale) + bmp_center
                                          , ((int)(_0.y * scale) + bmp_center)
                                          ,  (int)(_1.x * scale) + bmp_center
                                          , ((int)(_1.y * scale) + bmp_center)
                                          , at(i).c
                                     );
            }
        }
        strcpy(file_name, (EZ_ILDA_BMP_SHARE + file).c_str());
        bmp_save_as(&bmp, file_name);
        bmp_free(&bmp);
    }
    return;
}

//############################################################################
void ez_ilda_frame::display_stats() const
{
    int points = quantity      ,
        lit    = number_of_lit_points()  ,
        blank  = number_of_blank_points(),
        y_line = 8 * ez_ilda_pointer->font_size_factor,
        x_off  = EZ_STATS_DISPLAY_OFFSET * ez_ilda_pointer->font_size_factor,
        y_off  = BMP.yres - 19 * y_line; // 19 lines from the bottom

    bmp_printf(   &BMP
                , x_off
                , y_off
                , BMP.white
                , BMP.black
                , 0
                , ez_ilda_pointer->font_size_factor
                , "Frame stats"
              );

    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , ez_ilda_pointer->font_size_factor
                , "frame is %s"
                , ((is_selected)?("selected"):("not selected"))
              );

    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , ez_ilda_pointer->font_size_factor
                , "points ......... %d"
                , points
              );

    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , ez_ilda_pointer->font_size_factor
                , "lit points ..... %d"
                , lit
              );

    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , ez_ilda_pointer->font_size_factor
                , "blank points ... %d"
                , blank
              );

    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , ez_ilda_pointer->font_size_factor
                , "segments ....... %d"
                , number_of_segments()
              );

    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , ez_ilda_pointer->font_size_factor
                , "................ %f%% lit"
                , (lit/(float)points) * 100.0
              );

    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , ez_ilda_pointer->font_size_factor
                , "................ %f%% blank"
                , (blank/(float)points) * 100.0
              );

    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , ez_ilda_pointer->font_size_factor
                , "longest vector . %f points"
                , max_distance()
              );

    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , ez_ilda_pointer->font_size_factor
                , "total length ... %f points"
                , total_distance()
              );

    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , ez_ilda_pointer->font_size_factor
                , "sharpest angle . %f degrees"
                , max_angle() * one_radian
              );

    y_off += y_line;
    bmp_printf(   &BMP
                , x_off
                , y_off
                , next_hue()
                , BMP.black
                , 0
                , ez_ilda_pointer->font_size_factor
                , "total angle .... %f degrees"
                , total_angle() * one_radian
              );

    if(ez_ilda_pointer->show_cursors)
    {
        y_off += y_line;
        bmp_printf(   &BMP
                    , x_off
                    , y_off
                    , next_hue()
                    , BMP.black
                    , 0
                    , ez_ilda_pointer->font_size_factor
                    , "start select . % 5d (X %6d) (Y %6d) (Z %6d)"
                    , point_select_start
                    , at(point_select_start).x
                    , at(point_select_start).y
                    , at(point_select_start).z
                  );
        y_off += y_line;
        bmp_printf(   &BMP
                    , x_off
                    , y_off
                    , next_hue()
                    , BMP.black
                    , 0
                    , ez_ilda_pointer->font_size_factor
                    , "end select ... %5d (X %6d) (Y %6d) (Z %6d)"
                    , point_select_end
                    , at(point_select_end).x
                    , at(point_select_end).y
                    , at(point_select_end).z
                  );
        y_off += y_line;
        bmp_printf(   &BMP
                    , x_off
                    , y_off
                    , next_hue()
                    , BMP.black
                    , 0
                    , ez_ilda_pointer->font_size_factor
                    , "select center ...... (X %6.0f) (Y %6.0f) (Z %6.0f)"
                    , selected_segment().find_center().x
                    , selected_segment().find_center().y
                    , selected_segment().find_center().z
                  );
    }
    if(ez_ilda_pointer->show_arbitrary_point)
    {
        y_off += y_line;
        bmp_printf(   &BMP
                    , x_off
                    , y_off
                    , next_hue()
                    , BMP.black
                    , 0
                    , ez_ilda_pointer->font_size_factor
                    , "arbitrary point .... (X %6.0f) (Y %6.0f) (Z %6.0f)"
                    , ez_ilda_pointer->arbitrary_point.x
                    , ez_ilda_pointer->arbitrary_point.y
                    , ez_ilda_pointer->arbitrary_point.z
                  );
    }
    bmp_to_SDL_Surface(&BMP, screen);
    return;
}

//############################################################################








//############################################################################
ez_ilda_frame blank_frame()
{
    ez_ilda_frame frame(blank_segment(), ez_ilda_header());
    frame.quantity = (u_short)frame.quantity;
    frame.is_2D();
    return frame;
}

//############################################################################
ez_ilda_frame NULL_frame()
{
    ez_ilda_frame frame(NULL_segment(), ez_ilda_header());
    frame.quantity = (u_short)frame.quantity;
    frame.is_2D();
    return frame;
}

//############################################################################
//////////////////////////////////////////////////////////////////////////////
//############################################################################



