//###################################################################################
// ~ International Laser Display Association ~
// ~     ILDA File Format Driver in C++      ~
// ~             Built for EZFB              ~
//
// by James Lehman
// Extra Stimulus Inc.
// james@akrobiz.com
//
// began: October 2003
//
//###################################################################################
#ifndef __EZ_ILDA_SEGMENT_SET_DEFINITIONS__
#define __EZ_ILDA_SEGMENT_SET_DEFINITIONS__

//###################################################################################
#include "ez_ilda_segment.hpp"

//###################################################################################
class ez_ilda_segment_set : public vector<ez_ilda_segment>
{
public:
    ez_ilda_segment_set()    {};
    //-------------------------------------------------------------------------------
   ~ez_ilda_segment_set()    {    clear();    };
    //-------------------------------------------------------------------------------
    inline ez_ilda_segment_set&  operator += (const ez_ilda_segment_set& segment_set)
                                {
                                    insert(end(), segment_set.begin(), segment_set.end());
                                    return *this;
                                };
    //-------------------------------------------------------------------------------
    inline ez_ilda_segment_set&  operator += (const ez_ilda_segment& segment)
                                {
                                    push_back(segment);
                                    return *this;
                                };
    //-------------------------------------------------------------------------------
    inline ez_ilda_segment_set&  operator =  (const ez_ilda_segment_set& segment_set)
                                {
                                    clear();
                                    insert(begin(), segment_set.begin(), segment_set.end());
                                    return *this;
                                };
    //-------------------------------------------------------------------------------
    inline bool is_2D() const
                        {
                            if(size() > 0)
                            {
                                for(int i = 0; i < size(); i++)
                                    if(at(i).is_2D() == false)
                                        return false;
                            }
                            return true;
                        };
    //-------------------------------------------------------------------------------
    inline ez_ilda_point  first_lit_point() const
                        {
                            for(int i = 1; i < number_of_points(); i++)
                                if(at(i).is_lit())
                                    return at(i);
                            return ez_ilda_point(0);
                        };
    //-------------------------------------------------------------------------------
    inline int number_of_points() const
                        {
                            int total = 0;
                            for(int i = 0; i < size(); i++)                                
                                total += at(i).size();
                            return total;
                        };
    //-------------------------------------------------------------------------------
    inline int first_lit_point_index() const
                        {
                            if(size())
                                for(int i = 1; i < at(0).number_of_points(); i++)
                                    if(at(0).at(i).is_lit())
                                        return i;
                            return -1;
                        };
    //-------------------------------------------------------------------------------
    inline ez_ilda_point first_lit_anchor() const
                        {
                            for(int i = 1; i < number_of_points(); i++)
                                if(at(i).is_lit())
                                    return at(i - 1);
                            return ez_ilda_point(0);
                        };
    //-------------------------------------------------------------------------------
    inline int first_lit_anchor_index() const
                        {
                            for(int i = 1; i < number_of_points(); i++)
                                if(at(i).is_lit())
                                    return (i - 1);
                            return -1;
                        };
    //-------------------------------------------------------------------------------
    inline ez_ilda_point last_lit_anchor() const
                        {
                            for(int i = number_of_points() - 1; i > 0; i--)
                                if(at(i).is_lit())
                                    return at(i - 1);
                            return ez_ilda_point(0);
                        };
    //-------------------------------------------------------------------------------
    inline ez_ilda_point last_lit_point() const
                        {
                            for(int i = number_of_points() - 1; i > 0; i--)
                                if(at(i).is_lit())
                                    return at(i);
                            return ez_ilda_point(0);
                        };
    //-------------------------------------------------------------------------------
    inline int last_lit_point_index() const
                        {
                            for(int i = number_of_points() - 1; i > 0; i--)
                                if(at(i).is_lit())
                                    return i;
                            return -1;
                        };
    //-------------------------------------------------------------------------------
    inline int number_of_lit_points() const
                        {
                            int count = 0;
                            for(int i = 1; i < number_of_points(); i++)
                                if(at(i).is_lit())
                                    count++;
                            return count;
                        };
    //-------------------------------------------------------------------------------
    inline int number_of_blank_points() const
                        {
                            int count = 0;
                            for(int i = 1; i < number_of_points(); i++)
                                if(at(i).is_blank())
                                    count++;
                            return count;
                        };
    //-------------------------------------------------------------------------------
    inline int number_of_unique_colors() const
        {
            if(number_of_points() > 1)
            {
                int           i, j;
                ez_palette  palette;
                palette.push_back((ez_color)first_lit_point());
                for(i = first_lit_point_index() + 1; i < number_of_points(); i++)
                {
                    for(j = 0; j < palette.number_of_colors(); j++)
                    {
                        if(palette[j] == (ez_color)at(i))
                            break;
                    }
                    if(j == palette.number_of_colors())
                        palette.push_back(at(i));
                }
                return palette.number_of_colors();
            }
        };
    //-------------------------------------------------------------------------------
    inline ez_palette as_ez_palette() const
                {
                    ez_palette palette(number_of_points());
                    for(int i = 0; i < number_of_points(); i++)
                        palette[i] = (ez_color)(at(i));
                    return palette;
                };
    //-------------------------------------------------------------------------------
};

//###################################################################################
#endif

//###################################################################################
/////////////////////////////////////////////////////////////////////////////////////
//###################################################################################



