//############################################################################
//
// by James Lehman
// james@akrobiz.com
//
//############################################################################
#include "ez_ilda.hpp"

#define CHARS_IN_LINE 80
    
//############################################################################
void clear_string(char formatted_string[80])
{
    int i;
    for(i = 0; i < CHARS_IN_LINE; i++)
        formatted_string[i] = 0x00;
    return;
}
//############################################################################
int main(int argc, char *argv[])
{
    bool     running              = true ,
             show_help            = true ,
             in_sub_menu          = false,
             ui_visuals_flip_flop = false,
             show_frame_stats     = false,
             show_frame_set_stats = false,
             teminate_requested   = false;

    u_int    xres                 = 640,
             yres                 = 480,
             time_zero;

    string   file_name,
             ui_message;                         

    char     formatted_string[80];

    if(argc >= 2)
    {
        xres = atoi(argv[1]);
        yres = atoi(argv[2]);
    }

    ez_ilda  ild(xres, yres); // this is it!

    ild.from_file("in.ild");
    if(ild.error)
        ui_message = "NO DATA!";
    else
        ui_message = "in.ild";

    clear_string(formatted_string);
    ild.display(ui_message);
    ild.display_program_menu();

    while(running)
    {                         
        do
        {                                      
               SDL_WaitEvent(&event);
        } while(    (event.type != SDL_QUIT)
                 && (event.type != SDL_KEYDOWN)
               );
               
        if(event.type == SDL_QUIT)
            running = false;
            
        switch(event.key.keysym.sym)
        {
            //----------------------------------------------------------------
            case SDLK_ESCAPE:
                teminate_requested = true;
                break;
            //----------------------------------------------------------------






            //----------------------------------------------------------------
            case SDLK_UP: // display next palette
                ild.target_palette_index++;
                if(ild.target_palette_index >= ild.number_of_palettes())
                    ild.target_palette_index = 0;
                ui_message = "";
                break;
            //----------------------------------------------------------------
            case SDLK_DOWN: // display previous palette
                ild.target_palette_index--;
                if(ild.target_palette_index < 0)
                    ild.target_palette_index = ild.number_of_palettes() - 1;
                ui_message = "";
                break;                    
            //----------------------------------------------------------------
            case SDLK_RIGHT: // display previous frame
                if(    ild.loop_animation
                    || (ild.frame_index < ild.number_of_frames() - 1)
                  )
                    ++ild;
                ui_message = "";
                break;
            //----------------------------------------------------------------
            case SDLK_LEFT: // display next frame
                if(    ild.loop_animation
                    || (ild.frame_index > 0)
                  )
                    --ild;
                ui_message = "";
                break;
            //----------------------------------------------------------------






            //----------------------------------------------------------------
            case SDLK_DELETE:
                ild.delete_current_frame();
                ui_message = "deleted";
                break;
            //---------------------------------------------------------------- 







            //----------------------------------------------------------------
            case '`': // start animation forward
            //case '~': // start animation reverse
                SDL_Delay(250);
                if(event.key.keysym.mod & KMOD_SHIFT)
                {
                    SDL_PollEvent(&event);
                    while(    (event.type != SDL_KEYDOWN)
                           && (    ild.loop_animation
                                || (ild.frame_index > 0)
                              )
                         )
                    {
                        time_zero = SDL_GetTicks();
                        --ild;
                        ild.display("- show");
                        if(ild.approximate_frame_rate)
                        {
                            do{;}
                            while(   SDL_GetTicks() - time_zero
                                   < (u_int)((1.0 / ild.frames_per_second) * 1000.0)
                                 );
                        }
                        SDL_PollEvent(&event);
                    }
                }
                else
                {
                    SDL_PollEvent(&event);
                    while(    (event.type != SDL_KEYDOWN)
                           && (    ild.loop_animation
                                || (ild.frame_index < ild.number_of_frames() - 1)
                              )
                         )
                    {
                        time_zero = SDL_GetTicks();
                        ++ild;
                        ild.display("+ show");
                        if(ild.approximate_frame_rate)
                        {
                            do{;}
                            while(   SDL_GetTicks() - time_zero
                                   < (u_int)((1.0 / ild.frames_per_second) * 1000.0)
                                 );
                        }
                        SDL_PollEvent(&event);
                    }
                }
                ui_message = "";
                break;
            //----------------------------------------------------------------







            //----------------------------------------------------------------
            case '-': // start rotation formard
            //case '_': // start rotation reverse
                SDL_Delay(250);
                if(event.key.keysym.mod & KMOD_SHIFT)
                {
                    SDL_PollEvent(&event);
                    while(    (event.type != SDL_KEYDOWN)
                           && (    ild.loop_animation
                                || (ild.frame_index > 0)
                              )
                         )
                    {
                        time_zero = SDL_GetTicks();
                        ild.view_angle -= one_degree;
                        ild.display("- spin");
                        if(ild.approximate_frame_rate)
                        {
                            do{;}
                            while(   SDL_GetTicks() - time_zero
                                   < (u_int)((1.0 / ild.frames_per_second) * 1000.0)
                                 );
                        }
                        SDL_PollEvent(&event);
                    }
                }
                else
                {
                    SDL_PollEvent(&event);
                    while(    (event.type != SDL_KEYDOWN)
                           && (    ild.loop_animation
                                || (ild.frame_index < ild.number_of_frames() - 1)
                              )
                         )
                    {
                        time_zero = SDL_GetTicks();
                        ild.view_angle += one_degree;
                        ild.display("+ spin");
                        if(ild.approximate_frame_rate)
                        {
                            do{;}
                            while(   SDL_GetTicks() - time_zero
                                   < (u_int)((1.0 / ild.frames_per_second) * 1000.0)
                                 );
                        }
                        SDL_PollEvent(&event);
                    }
                }
                ui_message = "";
                break;
            //----------------------------------------------------------------








            //----------------------------------------------------------------
            case '=': // start rotation formard
            //case '+': // start rotation reverse
                SDL_Delay(250);
                if(event.key.keysym.mod & KMOD_SHIFT)
                {
                    SDL_PollEvent(&event);
                    while(    (event.type != SDL_KEYDOWN)
                           && (    ild.loop_animation
                                || (ild.frame_index > 0)
                              )
                         )
                    {
                        time_zero = SDL_GetTicks();
                        --ild;
                        ild.view_angle -= one_degree;
                        ild.display("- show & spin");
                        if(ild.approximate_frame_rate)
                        {
                            do{;}
                            while(   SDL_GetTicks() - time_zero
                                   < (u_int)((1.0 / ild.frames_per_second) * 1000.0)
                                 );
                        }
                        SDL_PollEvent(&event);
                    }
                }
                else
                {
                    SDL_PollEvent(&event);
                    while(    (event.type != SDL_KEYDOWN)
                           && (    ild.loop_animation
                                || (ild.frame_index < ild.number_of_frames() - 1)
                              )
                         )
                    {
                        time_zero = SDL_GetTicks();
                        ++ild;
                        ild.view_angle += one_degree;
                        ild.display("+ show & spin");
                        if(ild.approximate_frame_rate)
                        {
                            do{;}
                            while(   SDL_GetTicks() - time_zero
                                   < (u_int)((1.0 / ild.frames_per_second) * 1000.0)
                                 );
                        }
                        SDL_PollEvent(&event);
                    }
                }
                ui_message = "";
                break;
            //----------------------------------------------------------------





            //----------------------------------------------------
            case '/': // show frame set stats
            //case '?': // show frame stats
                if(event.key.keysym.mod & KMOD_SHIFT)
                    show_frame_stats = !show_frame_stats;
                else
                    show_frame_set_stats = !show_frame_set_stats;
                ui_message = "";
                break;                    
            //----------------------------------------------------






            //----------------------------------------------------------------
            case '1':
            //case '!':
                if(event.key.keysym.mod & KMOD_SHIFT)
                {
                    ild.view_angle.x -= one_degree;
                    ui_message = "- spin around X";
                }
                else
                {
                    ild.view_angle.x += one_degree;
                    ui_message = "+ spin around X";
                }
                break;                    
            //----------------------------------------------------------------
            case '2':
            //case '@':
                if(event.key.keysym.mod & KMOD_SHIFT)
                {
                    ild.view_angle.y -= one_degree;
                    ui_message = "- spin around Y";
                }
                else
                {
                    ild.view_angle.y += one_degree;
                    ui_message = "+ spin around Y";
                }
                break;                    
            //----------------------------------------------------------------
            case '3':
            //case '#':
                if(event.key.keysym.mod & KMOD_SHIFT)
                {
                    ild.view_angle.z -= one_degree;
                    ui_message = "- spin around Z";
                }
                else
                {
                    ild.view_angle.z += one_degree;
                    ui_message = "+ spin around Z";
                }
                break;                    
            //----------------------------------------------------------------
            case '4':
            //case '$':
                if(event.key.keysym.mod & KMOD_SHIFT)
                {
                    ild.view_angle -= one_degree;
                    ui_message = "- spin around XYZ";
                }
                else
                {
                    ild.view_angle += one_degree;
                    ui_message = "+ spin around XYZ";
                }
                break;                    
                //----------------------------------------------------------------
            case '5':
            //case '%':
                if(event.key.keysym.mod & KMOD_SHIFT)
                {
                    ild.view_angle.x = 0;
                    ild.view_angle.y = pi;
                    ild.view_angle.z = 0;
                    ui_message = "back";
                }
                else
                {
                    ild.view_angle   = 0;
                    ui_message = "front";
                }
                break;                    
            //----------------------------------------------------------------
            case '6':
            //case '^':
                if(event.key.keysym.mod & KMOD_SHIFT)
                {
                    ild.view_angle.x = 0;
                    ild.view_angle.y = -half_pi;
                    ild.view_angle.z = 0;
                    ui_message = "right side";
                }
                else
                {
                    ild.view_angle.x = 0;
                    ild.view_angle.y = half_pi;
                    ild.view_angle.z = 0;
                    ui_message = "left side";
                }
                break;                    
            //----------------------------------------------------------------
            case '7':
            //case '&':
                if(event.key.keysym.mod & KMOD_SHIFT)
                {
                    ild.view_angle.x = -half_pi;
                    ild.view_angle.y = 0;
                    ild.view_angle.z = 0;
                    ui_message = "bottom";
                }
                else
                {
                    ild.view_angle.x = half_pi;
                    ild.view_angle.y = 0;
                    ild.view_angle.z = 0;
                    ui_message = "top";
                }
                break;                    
            //----------------------------------------------------------------
            case '8':
            //case '*':
                if(event.key.keysym.mod & KMOD_SHIFT)
                {
                    ild.view_angle -= one_degree;
                    if(    ild.loop_animation
                        || (ild.frame_index > 0)
                      )
                        --ild;
                    ui_message = "- show & spin";
                }
                else
                {
                    ild.view_angle += one_degree;
                    if(    ild.loop_animation
                        || (ild.frame_index < ild.number_of_frames() - 1)
                      )
                        ++ild;
                    ui_message = "+ show & spin";
                }
                break;
            //----------------------------------------------------------------
            
            
            
                
            
            
                
            //----------------------------------------------------------------
            case '9': // go to first frame
            //case '(': // set, unset start frame select
                if(event.key.keysym.mod & KMOD_SHIFT)
                {
                    ild.toggle_frame_select_start();
                    if(ild.frame_select_start != -1)
                    {
                        sprintf(formatted_string, "start = %d", ild.frame_select_start);
                        ui_message = formatted_string;
                        clear_string(formatted_string);
                    }
                    else
                        ui_message = "start unset";
                }
                else
                {
                    ild.frame(0);
                    ui_message = "first frame";                            
                }
                break;
            //----------------------------------------------------------------
            case '0': // go to last frame
            //case ')': // set, unset end frame select
                if(event.key.keysym.mod & KMOD_SHIFT)
                {
                    ild.toggle_frame_select_end();
                    if(ild.frame_select_end != -1)
                    {
                        sprintf(formatted_string, "end = %d", ild.frame_select_end);
                        ui_message = formatted_string;
                        clear_string(formatted_string);                        
                    }
                    else
                        ui_message = "end unset";
                }
                else
                {
                    ild.frame(ild.number_of_frames() - 1);
                    ui_message = "last frame";
                }
                break;
            //----------------------------------------------------------------
            case ' ': // [Space] select, deselect current frame
                ild.current_frame().is_selected = !(ild.current_frame().is_selected);
                if(ild.current_frame().is_selected)
                    ui_message = "select frame";
                else
                    ui_message = "deselect frame";
                break;
            //----------------------------------------------------------------
            
            
            
            
            
            
            
                            
            
            
            
            
            //----------------------------------------------------------------                    
            case 'q':
                if(event.key.keysym.mod & KMOD_SHIFT)
                {
                    ild.view_offset.x -= 200;
                    ui_message = "- pan X";
                }
                else
                {
                    ild.view_offset.x += 200;
                    ui_message = "+ pan X";
                }
                break;                    
            //----------------------------------------------------------------
            case 'w':
                if(event.key.keysym.mod & KMOD_SHIFT)
                {
                    ild.view_offset.y -= 200;
                    ui_message = "- pan Y";
                }
                else
                {
                    ild.view_offset.y += 200;
                    ui_message = "+ pan Y";
                }
                break;                    
            //----------------------------------------------------------------
            case 'e':
                if(event.key.keysym.mod & KMOD_SHIFT)
                {
                    ild.view_offset.z -= 200;
                    ui_message = "- pan Z";
                }
                else
                {
                    ild.view_offset.z += 200;
                    ui_message = "+ pan Z";
                }
                break;                    
            //----------------------------------------------------------------
            case 'r':
                if(event.key.keysym.mod & KMOD_SHIFT)
                {
                    ild.view_offset -= 200;
                    ui_message = "- pan XYZ";
                }
                else
                {
                    ild.view_offset += 200;
                    ui_message = "+ pan XYZ";
                }
                break;                    
            //----------------------------------------------------------------
            case 't':
                ild.view_offset = 0.0;
                ui_message = "on center";
                break;
            //----------------------------------------------------------------
            case 'a':
                if(event.key.keysym.mod & KMOD_SHIFT)
                {
                    ild.view_scale.x *= 0.99;
                    ui_message = "- size X";
                }
                else
                {
                    ild.view_scale.x *= 1.01;
                    ui_message = "+ size X";
                }
                break;                    
            //----------------------------------------------------------------
            case 's':
                if(event.key.keysym.mod & KMOD_SHIFT)
                {
                    ild.view_scale.y *= 0.99;
                    ui_message = "- size Y";
                }
                else
                {
                    ild.view_scale.y *= 1.01;
                    ui_message = "+ size Y";
                }
                break;                    
            //----------------------------------------------------------------
            case 'd':
                if(event.key.keysym.mod & KMOD_SHIFT)
                {
                    ild.view_scale.z *= 0.99;
                    ui_message = "- size Z";
                }
                else
                {
                    ild.view_scale.z *= 1.01;
                    ui_message = "+ size Z";
                }
                break;                    
            //----------------------------------------------------------------
            case 'f':
                if(event.key.keysym.mod & KMOD_SHIFT)
                {
                    ild.view_scale *= 0.99;
                    ui_message = "- size XYZ";
                }
                else
                {
                    ild.view_scale *= 1.01;
                    ui_message = "+ size XYZ";
                }
                break;                    
            //----------------------------------------------------------------
            case 'g':                         
                ild.view_scale = 1.0;
                ui_message = "actual size";
                break;
            //----------------------------------------------------------------                                        















            //----------------------------------------------------------------
            case SDLK_TAB:
                ild.display("program settings");
                  ild.display_settings_menu();
                in_sub_menu = true;
                while(in_sub_menu)
                {
                    SDL_WaitEvent(&event);
                    if(event.type == SDL_QUIT)
                    {
                        in_sub_menu = false;
                        running     = false;
                    }
                    else if(event.type == SDL_KEYDOWN)
                    {
                        switch(event.key.keysym.sym)
                        {
                            //----------------------------------------------------------------
                            case SDLK_ESCAPE:
                                in_sub_menu = false;
                                ui_message = "action canceled";
                                break;                                                    
                            //--------------------------------------------------------
                            case '1':
                                ild.zero_angle_tolerance = (ild.display_prompt_and_echo_float("enter new zero angle tolerance in degrees") * one_degree);
                                in_sub_menu = false;
                                break;
                            //--------------------------------------------------------
                            case '2':
                                ild.distance_delta_max = ild.display_prompt_and_echo_float("enter new distance delta max in points");
                                in_sub_menu = false;
                                break;
                            //--------------------------------------------------------
                            case '3':
                                ild.longest_dwell_microsec = ild.display_prompt_and_echo_float("enter dwell in microseconds for 180 angle");
                                in_sub_menu = false;
                                break;
                            //--------------------------------------------------------
                            case '4':
                                ild.insignificant_distance = ild.display_prompt_and_echo_float("enter insignificant distance in points");
                                in_sub_menu = false;
                                break;
                            //--------------------------------------------------------
                            case '5':
                                ild.rendered_arc_angle = ild.display_prompt_and_echo_float("rendered arc segment angle in degrees");
                                in_sub_menu = false;
                                break;
                            //--------------------------------------------------------
                            case '6':
                                ild.move_points_per_key = ild.display_prompt_and_echo_int("enter new move points per key");
                                in_sub_menu = false;
                                break;
                            //--------------------------------------------------------
                            case '7':
                                ild.recolor_span_factor = ild.display_prompt_and_echo_float("enter new recolor palette span factor");
                                in_sub_menu = false;
                                break;
                            //--------------------------------------------------------
                            case '8':
                                ild.frames_per_second = ild.display_prompt_and_echo_float("enter new frames per second");
                                in_sub_menu = false;
                                break;
                            //--------------------------------------------------------
                            case '9':
                                ild.still_frame_duration = ild.display_prompt_and_echo_float("enter new durration for single frame wave output");
                                in_sub_menu = false;
                                break;
                            //--------------------------------------------------------
                            case '0':
                                ild.wave_sample_rate = ild.display_prompt_and_echo_int("enter new output wave sample rate");
                                in_sub_menu = false;
                                break;
                            //--------------------------------------------------------
                            case 'a':
                                ild.font_size_factor = ild.display_prompt_and_echo_int("display fonst size factor");
                                in_sub_menu = false;
                                break;
                            //--------------------------------------------------------
                            case 'b':
                                ild.output_bmp_size = ild.display_prompt_and_echo_int("output bitmap size in pixels (square): ");
                                in_sub_menu = false;
                                break;
                            //--------------------------------------------------------
                        }
                    }
                }
                break;
            //----------------------------------------------------------------                                        







            //----------------------------------------------------------------
            case 'u': // UI visuals
                ild.display("UI visuals");
                ild.display_visuals_menu();
                in_sub_menu = true;
                while(in_sub_menu)
                {
                    SDL_WaitEvent(&event);
                    if(event.type == SDL_QUIT)
                    {
                        in_sub_menu = false;
                        running     = false;
                    }
                    else if(event.type == SDL_KEYDOWN)
                    {
                        switch(event.key.keysym.sym)
                        {
                            //----------------------------------------------------------------
                            case SDLK_ESCAPE:
                                in_sub_menu = false;
                                ui_message = "action canceled";
                                break;
                            //--------------------------------------------------------
                            case '1':
                                ild.show_points          = !ild.show_points;
                                in_sub_menu = false;
                                if(ild.show_points)
                                    ui_message = "show points";
                                else
                                    ui_message = "hide points";
                                break;
                            //--------------------------------------------------------
                            case '2':
                                ild.show_blanking        = !ild.show_blanking;
                                in_sub_menu = false;
                                if(ild.show_blanking)
                                    ui_message = "show blanking";
                                else
                                    ui_message = "hide blanking";
                                break;
                            //--------------------------------------------------------
                            case '3':
                                ild.show_palettes        = !ild.show_palettes;
                                in_sub_menu = false;
                                if(ild.show_palettes)
                                    ui_message = "show palettes";
                                else
                                    ui_message = "hide palettes";
                                break;
                            //--------------------------------------------------------
                            case '4':
                                ild.show_stats           = !ild.show_stats;
                                in_sub_menu = false;
                                if(ild.show_stats)
                                    ui_message = "show stats";
                                else
                                    ui_message = "hide stats";
                                break;
                            //--------------------------------------------------------
                            case '5':
                                ild.show_fixed_bounds    = !ild.show_fixed_bounds;
                                in_sub_menu = false;
                                if(ild.show_fixed_bounds)
                                    ui_message = "show fixed bounds";
                                else
                                    ui_message = "hide fixed bounds";
                                break;
                            //--------------------------------------------------------
                            case '6':
                                ild.show_fixed_origin    = !ild.show_fixed_origin;
                                in_sub_menu = false;
                                if(ild.show_fixed_origin)
                                    ui_message = "show fixed origin";
                                else
                                    ui_message = "hide fixed origin";
                                break;
                            //--------------------------------------------------------
                            case '7':
                                ild.show_floating_axis   = !ild.show_floating_axis;
                                in_sub_menu = false;
                                if(ild.show_floating_axis)
                                    ui_message = "show floating axis";
                                else
                                    ui_message = "hide floating axis";
                                break;
                            //--------------------------------------------------------
                            case '8':
                                ild.show_floating_bounds = !ild.show_floating_bounds;
                                in_sub_menu = false;
                                if(ild.show_floating_bounds)
                                    ui_message = "show floating bounds";
                                else
                                    ui_message = "hide floating bounds";
                                break;
                            //--------------------------------------------------------
                            case '9':
                                ild.show_ui_message = !ild.show_ui_message;
                                in_sub_menu = false;
                                if(ild.show_floating_bounds)
                                    ui_message = "show ui message";
                                break;
                            //--------------------------------------------------------
                            case '0':
                                ui_visuals_flip_flop     = !ui_visuals_flip_flop;
                                ild.show_points          =  ui_visuals_flip_flop;
                                ild.show_blanking        =  ui_visuals_flip_flop;
                                ild.show_palettes        =  ui_visuals_flip_flop;
                                ild.show_stats           =  ui_visuals_flip_flop;
                                ild.show_fixed_bounds    =  ui_visuals_flip_flop;
                                ild.show_fixed_origin    =  ui_visuals_flip_flop;
                                ild.show_floating_axis   =  ui_visuals_flip_flop;
                                ild.show_floating_bounds =  ui_visuals_flip_flop;
                                ild.show_ui_message      =  ui_visuals_flip_flop;
                                in_sub_menu              =  false;
                                if(ui_visuals_flip_flop)
                                    ui_message = "show visuals";
                                else
                                    ui_message = "hide visuals";
                                break;
                            //--------------------------------------------------------
                            case 'a':
                                ild.show_effects_generation = !ild.show_effects_generation;
                                in_sub_menu = false;
                                if(ild.show_effects_generation)
                                    ui_message = "show live rendering";
                                else
                                    ui_message = "hide live rendering";
                                break;
                            //--------------------------------------------------------
                            case 'b':
                                ild.auto_flatten_z = !ild.auto_flatten_z;
                                in_sub_menu = false;
                                if(ild.auto_flatten_z)
                                    ui_message = "flatten Z for wave";
                                else
                                    ui_message = "keep Z for wave";
                                break;
                            //--------------------------------------------------------
                            case 'c':
                                ild.auto_optimize = !ild.auto_optimize;
                                in_sub_menu = false;
                                if(ild.auto_optimize)
                                    ui_message = "optimize wave output";
                                else
                                    ui_message = "raw wave output";
                                break;
                            //--------------------------------------------------------
                            case 'd':
                                ild.loop_animation = !ild.loop_animation;
                                in_sub_menu = false;
                                if(ild.loop_animation)
                                    ui_message = "loop animation";
                                else
                                    ui_message = "stop animatiion at ends";
                                break;
                            //--------------------------------------------------------
                            case 'e':
                                ild.approximate_frame_rate = !ild.approximate_frame_rate;
                                in_sub_menu = false;
                                if(ild.approximate_frame_rate)
                                    ui_message = "aproximate frame rate";
                                else
                                    ui_message = "as fast as possible";
                                break;
                            //--------------------------------------------------------
                            case 'f':
                                ild.auto_scale_dxf = !ild.auto_scale_dxf;
                                in_sub_menu = false;
                                if(ild.auto_scale_dxf)
                                {
                                    ild.maintain_dxf_origin = false;
                                    ui_message = "scale dxf import";
                                }
                                else                                
                                    ui_message = "no scale dxf import";
                                break;
                            //--------------------------------------------------------
                            case 'g':
                                ild.maintain_dxf_origin = !ild.maintain_dxf_origin;
                                in_sub_menu = false;
                                if(ild.maintain_dxf_origin)
                                {
                                    ild.auto_scale_dxf = false;
                                    ui_message = "maintain dxf origin";
                                }
                                else
                                    ui_message = "ignore dxf origin";
                                break;
                            //--------------------------------------------------------
                            case 'h':
                                ild.black_dwell_points = !ild.black_dwell_points;
                                in_sub_menu = false;
                                if(ild.black_dwell_points)
                                    ui_message = "black dwell";
                                else
                                    ui_message = "lit dwell";
                                break;
                            //--------------------------------------------------------
                        }
                    }
                }
                break;
            //----------------------------------------------------------------                                        








            //----------------------------------------------------------------
            case 'v': // apply view to frame(s) data
            {
                ild.simplify_view_angle();
                ild.display("view");
                ild.display_veiw_menu();
                in_sub_menu = true;
                while(in_sub_menu)
                {
                    SDL_WaitEvent(&event);
                    if(event.type == SDL_QUIT)
                    {
                        in_sub_menu = false;                                          
                        running     = false;
                    }
                    else if(event.type == SDL_KEYDOWN)
                    {
                        switch(event.key.keysym.sym)
                        {
                            //----------------------------------------------------------------
                            case SDLK_ESCAPE:
                                in_sub_menu = false;
                                ui_message = "action canceled";
                                break;
                            //--------------------------------------------------------
                            case '1':
                            //case '!':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.display("spinning all");
                                    ild.display_please_wait();
                                    ild.rotate_around_origin();
                                    ild.view_angle *= one_degree;
                                    ild.view_angle = 0.0;
                                    in_sub_menu = false;
                                    ui_message = "set spin all";
                                }
                                else
                                {
                                    ild.current_frame().rotate_around_origin();
                                    ild.view_angle *= one_degree;
                                    ild.view_angle = 0.0;
                                    in_sub_menu = false;
                                    ui_message = "set spin";
                                }
                                break;                    
                            //--------------------------------------------------------
                            case '2':
                            //case '@':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.display("moving all");
                                    ild.display_please_wait();
                                    ild.move();
                                    ild.view_offset = 0.0;                
                                    in_sub_menu = false;
                                    ui_message = "set move all";
                                }
                                else
                                {
                                    ild.current_frame().move();
                                    ild.view_offset = 0.0;                
                                    in_sub_menu = false;
                                    ui_message = "set move";
                                }
                                break;                    
                            //--------------------------------------------------------
                            case '3':
                            //case '#':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.display("sizing all");
                                    ild.display_please_wait();
                                    ild.scale_around_origin();
                                    ild.view_scale = 1.0;
                                    in_sub_menu = false;
                                    ui_message = "set size all";
                                }
                                else
                                {
                                    ild.current_frame().scale_around_origin();
                                    ild.view_scale = 1.0;
                                    in_sub_menu = false;
                                    ui_message = "set size";
                                }
                                break;                    
                            //--------------------------------------------------------
                            case '4':
                            //case '$':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.display("setting view all");
                                    ild.display_please_wait();
                                    ild.scale_around_origin();
                                    ild.rotate_around_origin();
                                    ild.move();        
                                    ild.view_angle  = 0.0;
                                    ild.view_scale  = 1.0;
                                    ild.view_offset = 0.0;
                                    in_sub_menu = false;
                                    ui_message = "set view all";
                                }
                                else
                                {
                                    ild.current_frame().scale_around_origin();
                                    ild.current_frame().rotate_around_origin();
                                    ild.current_frame().move();        
                                    ild.view_angle  = 0.0;
                                    ild.view_scale  = 1.0;
                                    ild.view_offset = 0.0;
                                    in_sub_menu = false;
                                    ui_message = "set view";
                                }
                                break;                    
                            //--------------------------------------------------------
                        }
                    }
                }
                break;
            }
            //----------------------------------------------------------------                                        







            //----------------------------------------------------------------
            case 'p': // Palette transformations
            {
                ild.display("palette transforms");
                ild.display_palette_tranz_menu();

                in_sub_menu = true;
                while(in_sub_menu)
                {
                    SDL_WaitEvent(&event);
                    if(event.type == SDL_QUIT)
                    {
                        in_sub_menu            = false;                                          
                        running = false;
                    }
                    else if(event.type == SDL_KEYDOWN)
                    {
                        switch(event.key.keysym.sym)
                        {
                            //----------------------------------------------------------------
                            case SDLK_ESCAPE:
                                in_sub_menu = false;
                                ui_message = "action canceled";
                                break;
                            //----------------------------------------------------------------
                            case SDLK_UP: // display next palette
                                ild.target_palette_index++;
                                if(ild.target_palette_index >= ild.number_of_palettes())
                                    ild.target_palette_index = 0;
                                ild.display();
                                ild.display_palette_tranz_menu();
                                SDL_Delay(200);
                                break;
                            //----------------------------------------------------------------
                            case SDLK_DOWN: // display previous palette
                                ild.target_palette_index--;
                                if(ild.target_palette_index < 0)
                                    ild.target_palette_index = ild.number_of_palettes() - 1;
                                ild.display();
                                ild.display_palette_tranz_menu();
                                SDL_Delay(200);
                                break;                    
                            //----------------------------------------------------------------
                            case SDLK_RIGHT: // display previous frame
                                if(    ild.loop_animation
                                    || (ild.frame_index < ild.number_of_frames() - 1)
                                  )
                                    ++ild;
                                ild.display();
                                ild.display_palette_tranz_menu();
                                SDL_Delay(200);
                                break;
                            //----------------------------------------------------------------
                            case SDLK_LEFT: // display next frame
                                if(    ild.loop_animation
                                    || (ild.frame_index > 0)
                                  )
                                    --ild;
                                ild.display();
                                ild.display_palette_tranz_menu();
                                SDL_Delay(200);
                                break;
                            //----------------------------------------------------------------
                                
                                                                        
                                
                                
                            //----------------------------------------------------------------                                                                        
                            case '1':
                            //case '!':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.display("spanning X");
                                    ild.display_please_wait();
                                    ild.rainbow_recolor(0, ild.recolor_span_factor);
                                    ui_message = "span X with palette";
                                }
                                else
                                {
                                    ild.current_frame().rainbow_recolor(0, ild.recolor_span_factor);
                                    ui_message = "span X with palette";
                                }
                                in_sub_menu = false;
                                break;                    
                            //--------------------------------------------------------
                            case '2':
                            //case '@':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.display("spanning Y");
                                    ild.display_please_wait();
                                    ild.rainbow_recolor(1, ild.recolor_span_factor);
                                    ui_message = "span Y with palette";
                                }
                                else
                                {
                                    ild.current_frame().rainbow_recolor(1, ild.recolor_span_factor);
                                    ui_message = "span Y with palette";
                                }
                                in_sub_menu = false;
                                break;                    
                            //--------------------------------------------------------
                            case '3':
                            //case '#':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.display("spanning Z");
                                    ild.display_please_wait();
                                    ild.rainbow_recolor(2, ild.recolor_span_factor);
                                    ui_message = "span Z with palette";
                                }
                                else
                                {
                                    ild.current_frame().rainbow_recolor(2, ild.recolor_span_factor);
                                    ui_message = "span Z with palette";
                                }
                                in_sub_menu = false;
                                break;                    
                            //--------------------------------------------------------
                            case '4':
                            //case '$':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.display("spanning points");
                                    ild.display_please_wait();
                                    ild.rainbow_recolor(3, ild.recolor_span_factor);
                                    ui_message = "span points with palette";
                                }
                                else
                                {
                                    ild.current_frame().rainbow_recolor(3, ild.recolor_span_factor);
                                    ui_message = "span points with palette";
                                }
                                in_sub_menu = false;
                                break;                    
                            //--------------------------------------------------------
                            case '5':
                            //case '%':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.display("indexing points");
                                    ild.display_please_wait();
                                    ild.rainbow_recolor(4, ild.recolor_span_factor);
                                    ui_message = "index points with palette";
                                }
                                else
                                {
                                     ild.current_frame().rainbow_recolor(4, ild.recolor_span_factor);
                                     ui_message = "index points with palette";
                                }
                                in_sub_menu = false;
                                break;                    
                            //--------------------------------------------------------
                            case '6':
                            //case '^':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.display("spanning lit segments");
                                    ild.display_please_wait();
                                    ild.rainbow_recolor(5, ild.recolor_span_factor);
                                    ui_message = "span lit segments with palette";
                                }
                                else
                                {
                                    ild.current_frame().rainbow_recolor(5, ild.recolor_span_factor);
                                    ui_message = "span lit segments with palette";
                                }
                                in_sub_menu = false;
                                break;                    
                            //--------------------------------------------------------
                            case '7':
                            //case '&':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.display("indexing lit segments");
                                    ild.display_please_wait();
                                    ild.rainbow_recolor(6, ild.recolor_span_factor);
                                    ui_message = "index lit segments with palette";
                                }
                                else
                                {
                                    ild.current_frame().rainbow_recolor(6, ild.recolor_span_factor);
                                    ui_message = "index lit segments with palette";
                                }
                                in_sub_menu = false;
                                break;                    
                            //--------------------------------------------------------
                            case '8':
                            //case '*':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.display("random coloring points");
                                    ild.display_please_wait();
                                    ild.rainbow_recolor(7, ild.recolor_span_factor);
                                    ui_message = "random color points with palette";
                                }
                                else
                                {
                                    ild.current_frame().rainbow_recolor(7, ild.recolor_span_factor);
                                    ui_message = "random color points with palette";
                                }
                                in_sub_menu = false;
                                break;                    
                            //--------------------------------------------------------
                            case '9':
                            //case '(':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.display("random coloring lit segments");
                                    ild.display_please_wait();
                                    ild.rainbow_recolor(8, ild.recolor_span_factor);
                                    ui_message = "random color lit segments with palette";
                                }
                                else
                                {
                                    ild.current_frame().rainbow_recolor(8, ild.recolor_span_factor);
                                    ui_message = "random color lit segments with palette";
                                }
                                in_sub_menu = false;
                                break;                    
                            //--------------------------------------------------------
                            case '0':
                            //case ')':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.display("matching target palette by index");
                                    ild.display_please_wait();
                                    ild.accept_target_palette();
                                    ui_message = "target by index";
                                }
                                else
                                {
                                    ild.current_frame().accept_target_palette();
                                    ui_message = "to target by index";
                                }
                                in_sub_menu = false;
                                break;                    
                            //--------------------------------------------------------
                            case 'a':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.display("best color matching target palette");
                                    ild.display_please_wait();
                                    ild.best_match_target_palette();
                                    ui_message = "best match to target palette";
                                }
                                else
                                {
                                    ild.current_frame().best_match_target_palette();
                                    ui_message = "best match to target palette";
                                }
                                in_sub_menu = false;
                                break;
                            //--------------------------------------------------------
                            case 'b':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.display("converting true color to palette");
                                    ild.display_please_wait();
                                    ild.sync_rgb_and_palette();
                                    ui_message = "convert to palette";
                                }
                                else
                                {
                                    ild.current_frame().sync_rgb_and_palette();
                                    ui_message = "convert to palette";
                                }
                                in_sub_menu = false;
                                break;                    
                            //--------------------------------------------------------
                            case 'c':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.display("bit reducing true color to palette");
                                    ild.display_please_wait();
                                    ild.bit_reduce_to_palette();
                                    ui_message = "bit reduce true color to palette";
                                }
                                else
                                {
                                    ild.current_frame().bit_reduce_to_palette();
                                    ui_message = "bit reduce true color to palette";
                                }
                                in_sub_menu = false;
                                break;                    
                            //--------------------------------------------------------
                            case 'd':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.display("promoting palette to true color");
                                    ild.display_please_wait();
                                    ild.promote_to_true_color();
                                    ui_message = "promote to true color";
                                }
                                else
                                {
                                    ild.current_frame().promote_to_true_color();
                                    ui_message = "promote to true color";
                                }
                                in_sub_menu = false;
                                break;                    
                            //--------------------------------------------------------                                    
                        }
                    }
                }
                break;
            }
            //----------------------------------------------------------------                                        






            //----------------------------------------------------------------
            case 'h': // frame transforms
                ild.display();
                ild.display_frame_tranz_menu();
                in_sub_menu = true;
                while(in_sub_menu)
                {
                    SDL_WaitEvent(&event);
                    if(event.type == SDL_QUIT)
                    {
                        in_sub_menu            = false;                                          
                        running = false;
                    }
                    else if(event.type == SDL_KEYDOWN)
                    {
                        switch(event.key.keysym.sym)
                        {
                            //----------------------------------------------------------------
                            case SDLK_ESCAPE:
                                in_sub_menu = false;
                                ui_message = "action canceled";
                                break;
                            //----------------------------------------------------------------
                            case SDLK_RIGHT: // display previous frame
                                if(    ild.loop_animation
                                    || (ild.frame_index < ild.number_of_frames() - 1)
                                  )
                                    ++ild;
                                ild.display();
                                ild.display_frame_tranz_menu();
                                SDL_Delay(200);
                                break;
                            //----------------------------------------------------------------
                            case SDLK_LEFT: // display next frame
                                if(    ild.loop_animation
                                    || (ild.frame_index > 0)
                                  )
                                    --ild;
                                ild.display();
                                ild.display_frame_tranz_menu();
                                SDL_Delay(200);
                                break;
                            //----------------------------------------------------------------
                            case '1':
                            //case '!':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.display("flattening Z");
                                    ild.display_please_wait();                                                        
                                    ild.flatten_z();
                                    ui_message = "flatten z all";
                                }
                                else
                                {
                                    ild.current_frame().flatten_z();
                                    ui_message = "flatten z";
                                }
                                in_sub_menu = false;
                                break;                    
                            //--------------------------------------------------------
                            case '2':
                            //case '@':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.display("Z ordering points");
                                    ild.display_please_wait();
                                    ild.z_order_points(64000);
                                    ui_message = "Z order points all";
                                }
                                else
                                {
                                    ild.current_frame().z_order_points(64000);
                                    ui_message = "z order points";
                                }
                                in_sub_menu = false;
                                break;                    
                            //--------------------------------------------------------
                            case '3':
                            //case '#':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.display("reversing vector order");
                                    ild.display_please_wait();
                                    ild.frame_reverse();
                                    ui_message = "reverse vector order all";
                                }
                                else
                                {
                                    ild.current_frame().reverse();
                                    ui_message = "reverse vector order";
                                }
                                in_sub_menu = false;
                                break;                    
                            //--------------------------------------------------------
                            case '4':
                            //case '$':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.display("mirroring");
                                    ild.display_please_wait();
                                    ild.flip(0);
                                    ui_message = "mirror all";
                                }
                                else
                                {
                                    ild.current_frame().flip(0);
                                    ui_message = "mirror";
                                }
                                in_sub_menu = false;
                                break;                    
                            //--------------------------------------------------------
                            case '5':
                            //case '%':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.display("flipping");
                                    ild.display_please_wait();
                                    ild.flip(1);
                                    ui_message = "flip all";
                                }
                                else
                                {
                                    ild.current_frame().flip(1);
                                    ui_message = "flip";
                                }
                                in_sub_menu = false;
                                break;                    
                            //--------------------------------------------------------
                            case '6':
                            //case '^':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.display("quarter turning CW");
                                    ild.display_please_wait();
                                    ild.quarter_turn(0, 1);
                                    ui_message = "quarter turn CW all";
                                }
                                else
                                {
                                    ild.current_frame().quarter_turn(0, 1);
                                    ui_message = "quarter turn CW";
                                }
                                in_sub_menu = false;
                                break;                    
                            //--------------------------------------------------------
                            case '7':
                            //se '&':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.display("quarter turning CCW");
                                    ild.display_please_wait();
                                    ild.quarter_turn(0, 3);
                                    ui_message = "quarter turn CCW all";
                                }
                                else
                                {
                                    ild.current_frame().quarter_turn(0, 3);
                                    ui_message = "quarter turn CCW";
                                }
                                in_sub_menu = false;
                                break;                    
                            //--------------------------------------------------------
                            case '8':
                            //case '*':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.display("minimizing blank points");
                                    ild.display_please_wait();
                                    ild.minimize_blank_points();
                                    ui_message = "remove blank span all";
                                }
                                else
                                {
                                    ild.current_frame().minimize_blank_points();
                                    ui_message = "remove blank span";
                                }
                                in_sub_menu = false;
                                break;                    
                            //--------------------------------------------------------
                            case '9':
                            //case '(':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.display("removing dwell points");
                                    ild.display_please_wait();
                                    ild.remove_dwell_points();
                                    ui_message = "remove angle dwell all";
                                }
                                else
                                {
                                    ild.current_frame().remove_dwell_points();
                                    ui_message = "remove angle dwell";
                                }
                                in_sub_menu = false;
                                break;                    
                            //--------------------------------------------------------
                            case '0':
                            //case ')':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.display("reducing vectors");
                                    ild.display_please_wait();
                                    ild.minimize_vectors();
                                    ui_message = "reduce vectors all";
                                }
                                else
                                {
                                    ild.current_frame().minimize_vectors();
                                    ui_message = "reduce vectors";
                                }
                                in_sub_menu = false;
                                break;                    
                            //--------------------------------------------------------                                
                            case 'a':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.display("bonding segments");
                                    ild.display_please_wait();
                                    ild.bond_segments();
                                    ui_message = "bond segments all";
                                }
                                else
                                {
                                    ild.current_frame().bond_segments();
                                    ui_message = "bond segments";
                                }
                                in_sub_menu = false;
                                break;                    
                            //--------------------------------------------------------
                            case 'b':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.display("minimizing points");
                                    ild.display_please_wait();
                                    ild.minimize_points();
                                    ui_message = "minimize points all";
                                }
                                else
                                {
                                    ild.current_frame().minimize_points();
                                    ui_message = "minimize points";
                                }
                                in_sub_menu = false;
                                break;                    
                            //--------------------------------------------------------                                
                            case 'c':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.display("fracturing segments");
                                    ild.display_please_wait();
                                    ild.fracture_segments();
                                    ui_message = "fracture segments all";
                                }
                                else
                                {
                                    ild.current_frame().fracture_segments();
                                    ui_message = "fracture segments";
                                }
                                in_sub_menu = false;
                                break;                    
                            //--------------------------------------------------------
                            case 'd':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.display("reordering segments");
                                    ild.display_please_wait();
                                    ild.reorder_segments();
                                    ui_message = "reorder segments all";
                                }
                                else
                                {
                                    ild.current_frame().reorder_segments();
                                    ui_message = "reorder segments";
                                }
                                in_sub_menu = false;
                                break;                    
                            //--------------------------------------------------------
                            case 'e':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.display("adding angle dwell");
                                    ild.display_please_wait();
                                    ild.add_dwell_points();
                                    ui_message = "add angle dwell all";
                                }
                                else
                                {
                                    ild.current_frame().add_dwell_points();
                                    ui_message = "add angle dwell";
                                }
                                in_sub_menu = false;
                                break;                    
                            //--------------------------------------------------------
                            case 'f':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.display("adding distance span");
                                    ild.display_please_wait();
                                    ild.add_span_points();
                                    ui_message = "add distance span all";
                                }
                                else
                                {
                                    ild.current_frame().add_span_points();
                                    ui_message = "add distance span";
                                }
                                in_sub_menu = false;
                                break;                    
                            //--------------------------------------------------------
                            case 'g':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.display("optimizing");
                                    ild.display_please_wait();
                                    ild.optimize_points();
                                    ui_message = "optimize all";
                                }
                                else
                                {
                                    ild.current_frame().optimize_points();
                                    ui_message = "optimize";
                                }
                                in_sub_menu = false;
                                break;                    
                            //--------------------------------------------------------
                        }
                    }
                }
                break;
            //----------------------------------------------------------------







            //----------------------------------------------------------------
            case 'j': // frame set transforms
            case 'J':
                ild.display();
                ild.display_frame_set_tranz_menu();
                in_sub_menu = true;
                while(in_sub_menu)
                {
                    SDL_WaitEvent(&event);
                    if(event.type == SDL_QUIT)
                    {
                        in_sub_menu            = false;                                          
                        running = false;
                    }
                    else if(event.type == SDL_KEYDOWN)
                    {
                        switch(event.key.keysym.sym)
                        {
                            //----------------------------------------------------------------
                            case SDLK_ESCAPE:
                                in_sub_menu = false;
                                ui_message = "action canceled";
                                break;
                            //----------------------------------------------------------------
                            case SDLK_RIGHT: // display previous frame
                                if(    ild.loop_animation
                                    || (ild.frame_index < ild.number_of_frames() - 1)
                                  )
                                    ++ild;
                                ild.display();
                                ild.display_frame_tranz_menu();
                                SDL_Delay(200);
                                break;
                            //----------------------------------------------------------------
                            case SDLK_LEFT: // display next frame
                                if(    ild.loop_animation
                                    || (ild.frame_index > 0)
                                  )
                                    --ild;
                                ild.display();
                                ild.display_frame_tranz_menu();
                                SDL_Delay(200);
                                break;
                            //--------------------------------------------------------
                            case '1':
                                ild.display("deleting selected frames");
                                ild.display_please_wait();
                                ild.delete_selected_frames();
                                ui_message = "deleted selected frames";
                                in_sub_menu = false;
                                break;
                            //--------------------------------------------------------
                            case '2':   
                                ild.display("trimming to selected frames");
                                ild.display_please_wait();
                                ild.trim_to_selected_frames();
                                ui_message = "trimmed to selected frames";
                                in_sub_menu = false;
                                break;
                            //--------------------------------------------------------
                            case '3':
                                ild.display("copying selected frames");
                                ild.display_please_wait();
                                ild = (ild.selected_frames() + ild);
                                ui_message = "copied frames to begining";
                                in_sub_menu = false;
                                break;
                            //--------------------------------------------------------
                            case '4':
                                ild.display("moving selected frames");
                                ild.display_please_wait();
                                {   // Ahhhhhh!  minty fresh scope
                                    ez_ilda_frame_set temp(ild.selected_frames());
                                    ild.delete_selected_frames();
                                    temp += ild;
                                    ild = temp;
                                }
                                ui_message = "moved frames to begining";
                                in_sub_menu = false;
                                break;
                            //--------------------------------------------------------
                            case '5':
                                ild.display("copying selected frames");
                                ild.display_please_wait();
                                ild += ild.selected_frames();
                                ui_message = "copied frames to end";
                                in_sub_menu = false;
                                break;
                            //--------------------------------------------------------
                            case '6':
                                ild.display("moving selected frames");
                                ild.display_please_wait();
                                {
                                    ez_ilda_frame_set temp(ild.selected_frames());
                                    ild.delete_selected_frames();
                                    ild += temp;
                                }
                                ui_message = "moved frames to end";
                                in_sub_menu = false;
                                break;
                            //--------------------------------------------------------
                            case '7':
                                ild.display("deleting redundant frames");
                                ild.display_please_wait();
                                ild.delete_redundant_frames();
                                ui_message = "redundant frames deleted";
                                in_sub_menu = false;
                                break;
                            //--------------------------------------------------------
                            case '8':
                                {
                                    ez_ilda_frame_set temp;
                                    temp += ez_ilda_frame(ild.target_palette_index, 0); // not quite an empty frame!
                                    ild = temp + ild;
                                }
                                ui_message = "add frame to begining";
                                in_sub_menu = false;
                                break;
                            //--------------------------------------------------------
                            case '9':
                                {
                                    int               i;
                                    ez_ilda_frame_set temp;
                                    for(i = 0; i <= ild.frame_index; i++)
                                        temp += ild[i];
                                    temp += ez_ilda_frame(ild.target_palette_index, 0);
                                    for(i = ild.frame_index + 1; i < ild.number_of_frames(); i++)
                                        temp += ild[i];
                                    ild = temp;
                                }
                                ui_message = "add after current frame";
                                in_sub_menu = false;
                                break;
                            //--------------------------------------------------------
                            case '0':
                                ild += ez_ilda_frame(ild.target_palette_index, 0);
                                ui_message = "add frame to end";
                                in_sub_menu = false;
                                break;
                            //--------------------------------------------------------
                            case 'a':
                                ild.display("reversing order of frames");
                                ild.display_please_wait();
                                ild.reverse();
                                ui_message = "reverse order";
                                in_sub_menu = false;
                                break;
                            //--------------------------------------------------------
                            case '>': // explode current frame
                                ild.explode_current_frame();
                                ui_message = "explode segments";
                                in_sub_menu = false;
                                break;
                            //--------------------------------------------------------
                            case '<': // collapse selected frames
                                ild.collapse_selected_frames();
                                ui_message = "collapse segments";
                                in_sub_menu = false;
                                break;
                        }
                    }
                }
                break;
            //----------------------------------------------------------------







            //----------------------------------------------------------------
            case 'k': // draw, color, blank
            {   // local scope
                int         number_of_sides;
                float       temp_float;
                ez_3D_float temp_3D_float;                        
            
                ild.show_cursors = true;
                in_sub_menu      = true;
                ild.display("draw color blank");
                if(show_frame_stats)
                    ild.current_frame().display_stats();
                if(show_frame_set_stats)
                    ild.display_stats();
                while(in_sub_menu)
                {
                    SDL_WaitEvent(&event);
                    if(event.type == SDL_QUIT)
                    {
                        in_sub_menu            = false;                                          
                        running = false;
                    }
                    else if(event.type == SDL_KEYDOWN)
                    {
                        switch(event.key.keysym.sym)
                        {
                            //----------------------------------------------------------------
                            case SDLK_ESCAPE:
                                in_sub_menu      = false;
                                   ild.show_cursors = false;
                                ui_message = "exit draw color blank";
                                break;
                            //----------------------------------------------------------------
                            case SDLK_UP: // display next palette
                                ild.target_palette_index++;
                                if(ild.target_palette_index >= ild.number_of_palettes())
                                    ild.target_palette_index = 0;
                                ui_message = "";
                                break;
                            //----------------------------------------------------------------
                            case SDLK_DOWN: // display previous palette
                                ild.target_palette_index--;
                                if(ild.target_palette_index < 0)
                                    ild.target_palette_index = ild.number_of_palettes() - 1;
                                ui_message = "";
                                break;                    
                            //----------------------------------------------------------------
                            case SDLK_RIGHT: // display previous frame
                                if(    ild.loop_animation
                                    || (ild.frame_index < ild.number_of_frames() - 1)
                                  )
                                    ++ild;
                                break;
                            //----------------------------------------------------------------
                            case SDLK_LEFT: // display next frame
                                if(    ild.loop_animation
                                    || (ild.frame_index > 0)
                                  )
                                    --ild;
                                break;
                            //----------------------------------------------------------------
                            case SDLK_BACKSPACE:
                                ild.current_frame().remove_point(ild.current_frame().point_select_start);
                                ui_message = "delete start";
                                break;
                            //----------------------------------------------------------------
                            case '1':
                            //case '!':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.view_angle.x -= one_degree;
                                    ui_message = "- spin around X";
                                }
                                else
                                {
                                    ild.view_angle.x += one_degree;
                                    ui_message = "+ spin around X";
                                }
                                break;                    
                            //----------------------------------------------------------------
                            case '2':
                            //case '@':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.view_angle.y -= one_degree;
                                    ui_message = "- spin around Y";
                                }
                                else
                                {
                                    ild.view_angle.y += one_degree;
                                    ui_message = "+ spin around Y";
                                }
                                break;                    
                            //----------------------------------------------------------------
                            case '3':
                            //case '#':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.view_angle.z -= one_degree;
                                    ui_message = "- spin around Z";
                                }
                                else
                                {
                                    ild.view_angle.z += one_degree;
                                    ui_message = "+ spin around Z";
                                }
                                break;                    
                            //----------------------------------------------------------------
                            case '4':
                            //case '$':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.view_angle -= one_degree;
                                    ui_message = "- spin around XYZ";
                                }
                                else
                                {
                                    ild.view_angle += one_degree;
                                    ui_message = "+ spin around XYZ";
                                }
                                break;                    
                                //----------------------------------------------------------------
                            case '5':
                            //case '%':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.view_angle.x = 0;
                                    ild.view_angle.y = pi;
                                    ild.view_angle.z = 0;
                                    ui_message = "back";
                                }
                                else
                                {
                                    ild.view_angle   = 0;
                                    ui_message = "front";
                                }
                                break;                    
                            //----------------------------------------------------------------
                            case '6':
                            //case '^':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.view_angle.x = 0;
                                    ild.view_angle.y = -half_pi;
                                    ild.view_angle.z = 0;
                                    ui_message = "right side";
                                }
                                else
                                {
                                    ild.view_angle.x = 0;
                                    ild.view_angle.y = half_pi;
                                    ild.view_angle.z = 0;
                                    ui_message = "left side";
                                }
                                break;                    
                            //----------------------------------------------------------------
                            case '7':
                            //case '&':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.view_angle.x = -half_pi;
                                    ild.view_angle.y = 0;
                                    ild.view_angle.z = 0;
                                    ui_message = "bottom";
                                }
                                else
                                {
                                    ild.view_angle.x = half_pi;
                                    ild.view_angle.y = 0;
                                    ild.view_angle.z = 0;
                                    ui_message = "top";
                                }
                                break;                    
                            //----------------------------------------------------------------
                            case '8':
                            //case '*':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.view_angle -= one_degree;
                                    if(    ild.loop_animation
                                        || (ild.frame_index > 0)
                                      )
                                        --ild;
                                    ui_message = "- show & spin";
                                }
                                else
                                {
                                    ild.view_angle += one_degree;
                                    if(    ild.loop_animation
                                        || (ild.frame_index < ild.number_of_frames() - 1)
                                      )
                                        ++ild;
                                    ui_message = "+ show & spin";
                                }
                                break;
                            //----------------------------------------------------
                            case '[':
                            //case '{':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.current_frame().deccrement_point_select_start();
                                    ui_message = "- start select";
                                }
                                else
                                {
                                    ild.current_frame().increment_point_select_start();
                                    ui_message = "+ start select";
                                }
                                break;                    
                            //----------------------------------------------------
                            case ']':
                            //case '}':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.current_frame().deccrement_point_select_end();
                                    ui_message = "- end select";
                                }
                                else
                                {
                                    ild.current_frame().increment_point_select_end();
                                    ui_message = "+ end select";
                                }
                                break;                    
                            //----------------------------------------------------
                            case'\\':
                            //case '|':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.current_frame().point_select_start = 0;
                                    ui_message = "first vertex select";
                                }
                                else
                                {
                                    ild.current_frame().point_select_end = (ild.current_frame().number_of_points() - 1);
                                    ui_message = "last vertex select";
                                }
                                break;                    
                            //----------------------------------------------------
                            case 'u':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.current_frame().previous_segment_select();
                                    ui_message = "- lit segment";
                                }
                                else
                                {
                                    ild.current_frame().next_segment_select();
                                    ui_message = "+ lit segment";
                                }
                                break;                    
                            //----------------------------------------------------
                            case 'i':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.current_frame().previous_segment_select_start();
                                    ui_message = "- start lit segment";
                                }
                                else
                                {
                                    ild.current_frame().next_segment_select_start();
                                    ui_message = "+ start lit segment";
                                }
                                break;                    
                            //----------------------------------------------------
                            case 'o':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.current_frame().previous_segment_select_end();
                                    ui_message = "- end lit segment";
                                }
                                else
                                {
                                    ild.current_frame().next_segment_select_end();
                                    ui_message = "+ end lit segment";
                                }
                                break;
                            //----------------------------------------------------
                            case '/': // show frame set stats
                            //case '?': // show frame stats
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                    show_frame_stats = !show_frame_stats;
                                else
                                    show_frame_set_stats = !show_frame_set_stats;
                                ui_message = "";
                                break;                    
                            //----------------------------------------------------
                            case 'q':
                                ild.current_frame().at(ild.current_frame().point_select_start).x = ild.display_prompt_and_echo_int("set value of start X");
                                   ui_message = "X set";
                                   break;
                            //----------------------------------------------------            
                            case 'w':
                                ild.current_frame().at(ild.current_frame().point_select_start).y = ild.display_prompt_and_echo_int("set value of start Y");
                                   ui_message = "Y set";
                                   break;
                            //----------------------------------------------------            
                            case 'e':
                                ild.current_frame().at(ild.current_frame().point_select_start).z = ild.display_prompt_and_echo_int("set value of start Z");
                                   ui_message = "Z set";
                                   break;
                            //----------------------------------------------------
                            case 'p':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.selected_palette_color--;
                                    if(ild.selected_palette_color <= 0)
                                        ild.selected_palette_color = ild.current_palette().number_of_colors() - 1;
                                    ui_message = "- color";
                                }
                                else
                                {
                                    ild.selected_palette_color++;
                                    if(ild.selected_palette_color >= ild.current_palette().number_of_colors())
                                        ild.selected_palette_color = 0;
                                    ui_message = "+ color";
                                }
                                break;                    
                            //----------------------------------------------------
                            case 'a':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.unblank_all_points();
                                    ui_message = "unblank all frames";
                                }
                                else
                                {
                                    ild.current_frame().unblank_all_points();
                                    ui_message = "unblank all points";
                                }
                                break;                    
                            //----------------------------------------------------                        
                            case 'b':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.current_frame().unblank_points();
                                    ui_message = "unblank selected points";
                                }
                                else
                                {
                                    ild.current_frame().blank_points();
                                    ui_message = "blank selected points";
                                }
                                break;                    
                            //----------------------------------------------------                        
                            case 'c':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.selected_palette_color = ild.current_frame().at(ild.current_frame().point_select_start).c;
                                    ui_message = "pick color at select-start";
                                }
                                else
                                {
                                    ild.current_frame().set_points_color_index();
                                    ui_message = "color selected points";
                                }
                                break;                    
                            //----------------------------------------------------
                            case '.':
                                ild.current_frame().add_new_point(ild.current_frame().point_select_start);
                                if(ild.current_frame().point_select_start < (ild.current_frame().number_of_points() - 1))
                                    ui_message = "split vector";
                                else
                                    ui_message = "add vector";
                                break;
                            //----------------------------------------------------            
                            case 't':
                                {
                                    float  x_rotation;
                                    string font,
                                           text;
                                    //list_ilda_files();
                                    font       = ild.display_prompt_and_echo_name("enter font (ilda) file name");
                                    ild.display(ui_message);
                                    if(show_help)
                                        ild.display_draw_color_blank_menu();                        
                                    x_rotation = ild.display_prompt_and_echo_float ("enter angle of rotation around vector");
                                    ild.display(ui_message);
                                    if(show_help)
                                        ild.display_draw_color_blank_menu();                        
                                    text       = ild.display_prompt_and_echo_name("text to render");

                                    ild.display("rendering text");
                                    ild.display_please_wait();
                                    ild.current_frame()
                                    += ez_ilda_segment(      ild.current_frame().at(ild.current_frame().point_select_start)
                                                        , ild.current_frame().at(ild.current_frame().point_select_end)
                                                        , x_rotation
                                                        , font
                                                        , text
                                                      );

                                    ild.current_frame().at(ild.current_frame().point_select_start).blank();
                                    ui_message = "text rendered";
                                }
                                break;
                            //----------------------------------------------------            
                            case 'k':
                                {
                                    int number_of_sides = ild.display_prompt_and_echo_int("enter integer number of sides");
                                    ild.current_frame()
                                    += ez_ilda_segment(      ild.current_frame().at(ild.current_frame().point_select_start)
                                                        , ild.current_frame().at(ild.current_frame().point_select_end)
                                                        , number_of_sides
                                                      );
                                }
                                break;
                            //----------------------------------------------------            
                            case 'd':
                                {
                                    float   arc_angle,
                                            radii_ratio;
                                    if(event.key.keysym.mod & KMOD_SHIFT)
                                    {
                                        arc_angle = ild.display_prompt_and_echo_float("enter elliptical arc angle in degrees, (0 = 360)");
                                        ild.display(ui_message);
                                        if(show_help)
                                            ild.display_draw_color_blank_menu();                        
                                        radii_ratio = ild.display_prompt_and_echo_float("enter minor / major radii ratio");
                                        ild.display(ui_message);
                                        if(show_help)
                                            ild.display_draw_color_blank_menu();                        
                                        ild.current_frame()
                                        += ez_ilda_segment(      ild.current_frame().at(ild.current_frame().point_select_start)
                                                            , ild.current_frame().at(ild.current_frame().point_select_end)
                                                            , arc_angle
                                                            , radii_ratio
                                                          );
                                        ui_message = "elliptical arc";
                                    }
                                    else
                                    {
                                        arc_angle = ild.display_prompt_and_echo_float("enter circular arc angle in degrees, (0 = 360)");
                                        ild.display(ui_message);
                                        if(show_help)
                                            ild.display_draw_color_blank_menu();                        
                                        ild.current_frame()
                                        += ez_ilda_segment(      ild.current_frame().at(ild.current_frame().point_select_start)
                                                            , ild.current_frame().at(ild.current_frame().point_select_end)
                                                            , arc_angle
                                                          );
                                        ui_message = "circular arc";
                                    }
                                }                                    
                                break;
                            //----------------------------------------------------            
                            case 'l':
                                ild.current_frame().connect_the_dots();
                                ui_message = "draw line";
                                break;                                
                            //----------------------------------------------------            
                            case 's':
                                ild.current_frame().break_segment();
                                ui_message = "break segment";
                                break;
                            //----------------------------------------------------
                            case 'x':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    temp_float = ild.current_frame().at(ild.current_frame().point_select_start).x - ild.move_points_per_key;
                                    if(temp_float >= EZ_MIN_SHORT)
                                        ild.current_frame().at(ild.current_frame().point_select_start).x = (u_short)temp_float;
                                    ui_message = "- X start";
                                }
                                else
                                {
                                    temp_float = ild.current_frame().at(ild.current_frame().point_select_start).x + ild.move_points_per_key;
                                    if(temp_float <= EZ_MAX_SHORT)
                                        ild.current_frame().at(ild.current_frame().point_select_start).x = (u_short)temp_float;
                                    ui_message = "+ X start";
                                }
                                break;                    
                            //----------------------------------------------------
                            case 'y':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    temp_float = ild.current_frame().at(ild.current_frame().point_select_start).y - ild.move_points_per_key;
                                    if(temp_float >= EZ_MIN_SHORT)
                                        ild.current_frame().at(ild.current_frame().point_select_start).y = (u_short)temp_float;
                                    ui_message = "- Y start";
                                }
                                else
                                {
                                    temp_float = ild.current_frame().at(ild.current_frame().point_select_start).y + ild.move_points_per_key;
                                    if(temp_float <= EZ_MAX_SHORT)
                                        ild.current_frame().at(ild.current_frame().point_select_start).y = (u_short)temp_float;
                                    ui_message = "+ Y start";
                                }
                                break;                    
                            //----------------------------------------------------
                            case 'z':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    temp_float = ild.current_frame().at(ild.current_frame().point_select_start).z - ild.move_points_per_key;
                                    if(temp_float >= EZ_MIN_SHORT)
                                        ild.current_frame().at(ild.current_frame().point_select_start).z = (u_short)temp_float;
                                    ui_message = "- Z start";
                                }
                                else
                                {
                                    temp_float = ild.current_frame().at(ild.current_frame().point_select_start).z + ild.move_points_per_key;
                                    if(temp_float <= EZ_MAX_SHORT)
                                        ild.current_frame().at(ild.current_frame().point_select_start).z = (u_short)temp_float;
                                    ui_message = "+ Z start";
                                }
                                break;                    
                            //----------------------------------------------------
                            case 'f':
                                if(ild.current_frame().point_select_start) // is not the origin
                                {

                                    if(event.key.keysym.mod & KMOD_SHIFT)
                                    {
                                        temp_3D_float
                                        = rotate_point_around_point_z(   ild.current_frame().at(ild.current_frame().point_select_start)
                                                                       , ild.current_frame().at(ild.current_frame().point_select_start - 1)
                                                                       , -one_degree
                                                                     );
                                        if(!ez_ilda_bounds_check(temp_3D_float, EZ_ILDA_CUBE))
                                            ild.current_frame().at(ild.current_frame().point_select_start) = temp_3D_float;
                                        ui_message = "- spin start Z axis";
                                    }
                                    else
                                    {
                                        temp_3D_float
                                        = rotate_point_around_point_z(   ild.current_frame().at(ild.current_frame().point_select_start)
                                                                       , ild.current_frame().at(ild.current_frame().point_select_start - 1)
                                                                       , one_degree
                                                                     );
                                        if(!ez_ilda_bounds_check(temp_3D_float, EZ_ILDA_CUBE))
                                            ild.current_frame().at(ild.current_frame().point_select_start) = temp_3D_float;
                                        ui_message = "+ spin start Z axis";
                                    }
                                }
                                else
                                    ui_message = "not at zero";
                                break;                    
                            //----------------------------------------------------
                            case 'g':
                                if(ild.current_frame().point_select_start) // is not the origin
                                {

                                    if(event.key.keysym.mod & KMOD_SHIFT)
                                    {
                                        temp_3D_float
                                        = rotate_point_around_point_y(   ild.current_frame().at(ild.current_frame().point_select_start)
                                                                       , ild.current_frame().at(ild.current_frame().point_select_start - 1)
                                                                       , -one_degree
                                                                     );
                                        if(!ez_ilda_bounds_check(temp_3D_float, EZ_ILDA_CUBE))
                                            ild.current_frame().at(ild.current_frame().point_select_start) = temp_3D_float;
                                        ui_message = "- spin start Y axis";
                                    }
                                    else
                                    {
                                        temp_3D_float
                                        = rotate_point_around_point_y(   ild.current_frame().at(ild.current_frame().point_select_start)
                                                                       , ild.current_frame().at(ild.current_frame().point_select_start - 1)
                                                                       , one_degree
                                                                     );
                                        if(!ez_ilda_bounds_check(temp_3D_float, EZ_ILDA_CUBE))
                                            ild.current_frame().at(ild.current_frame().point_select_start) = temp_3D_float;
                                        ui_message = "+ spin start Y axis";
                                    }
                                }
                                else
                                    ui_message = "not at zero";
                                break;                    
                            //----------------------------------------------------
                            case 'm':
                                if(ild.current_frame().point_select_start) // is not the origin
                                {

                                    if(event.key.keysym.mod & KMOD_SHIFT)
                                    {
                                        temp_3D_float                                          
                                        = scale_point_around_point(   ild.current_frame().at(ild.current_frame().point_select_start)
                                                                    , ild.current_frame().at(ild.current_frame().point_select_start - 1)
                                                                    , ez_3D_float(0.99, 0.99, 0.99)
                                                                  );
                                        if(!ez_ilda_bounds_check(temp_3D_float, EZ_ILDA_CUBE))
                                            ild.current_frame().at(ild.current_frame().point_select_start) = temp_3D_float;
                                        ui_message = "- magnitude start vector";
                                    }
                                    else
                                    {
                                    temp_3D_float
                                    = scale_point_around_point(   ild.current_frame().at(ild.current_frame().point_select_start)
                                                                , ild.current_frame().at(ild.current_frame().point_select_start - 1)
                                                                , ez_3D_float(1.01, 1.01, 1.01)
                                                              );
                                    if(!ez_ilda_bounds_check(temp_3D_float, EZ_ILDA_CUBE))
                                        ild.current_frame().at(ild.current_frame().point_select_start) = temp_3D_float;
                                        ui_message = "+ magnitude start vector";
                                    }
                                }
                                else
                                    ui_message = "not at zero";
                                break;                    
                            //----------------------------------------------------                            
                            case SDLK_RETURN:
                                show_help = !show_help;
                            //----------------------------------------------------------------
                        } // end switch(event.key.keysym.sym)
                        ild.display(ui_message);
                        if(show_help)
                            ild.display_draw_color_blank_menu();                        
                        if(show_frame_stats)
                            ild.current_frame().display_stats();
                        if(show_frame_set_stats)
                            ild.display_stats();
                    } // end else if(event.type == SDL_KEYDOWN)
                } // end while(in_sub_menu)
                break;
            }   // end local scope
            //----------------------------------------------------------------








            //----------------------------------------------------------------
            case 'l': // move size spin
            {   // local scope
                u_short offset;
                ild.show_cursors         = true;
                ild.show_arbitrary_point = true;
                in_sub_menu              = true;
                
                ild.display("move size spin");
                if(show_frame_stats)
                    ild.current_frame().display_stats();
                if(show_frame_set_stats)
                    ild.display_stats();
                while(in_sub_menu)
                {
                    SDL_WaitEvent(&event);
                    if(event.type == SDL_QUIT)
                    {
                        in_sub_menu = false;
                        running     = false;
                    }
                    else if(event.type == SDL_KEYDOWN)
                    {
                        switch(event.key.keysym.sym)
                        {
                            //----------------------------------------------------------------
                            case SDLK_ESCAPE:
                                in_sub_menu              = false;
                                ild.show_cursors         = false;
                                ild.show_arbitrary_point = false;
                                ui_message = "exit move size spin";
                                break;
                            //----------------------------------------------------------------
                            case SDLK_UP: // display next palette
                                ild.target_palette_index++;
                                if(ild.target_palette_index >= ild.number_of_palettes())
                                    ild.target_palette_index = 0;
                                ui_message = "";
                                break;
                            //----------------------------------------------------------------
                            case SDLK_DOWN: // display previous palette
                                ild.target_palette_index--;
                                if(ild.target_palette_index < 0)
                                    ild.target_palette_index = ild.number_of_palettes() - 1;
                                ui_message = "";
                                break;                    
                            //----------------------------------------------------------------
                            case SDLK_RIGHT: // display previous frame
                                if(    ild.loop_animation
                                    || (ild.frame_index < ild.number_of_frames() - 1)
                                  )
                                    ++ild;
                                break;
                            //----------------------------------------------------------------
                            case SDLK_LEFT: // display next frame
                                if(    ild.loop_animation
                                    || (ild.frame_index > 0)
                                  )
                                    --ild;
                                break;
                            //----------------------------------------------------------------
                            case '1':
                            //case '!':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.view_angle.x -= one_degree;
                                    ui_message = "- spin around X";
                                }
                                else
                                {
                                    ild.view_angle.x += one_degree;
                                    ui_message = "+ spin around X";
                                }
                                break;                    
                            //----------------------------------------------------------------
                            case '2':
                            //case '@':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.view_angle.y -= one_degree;
                                    ui_message = "- spin around Y";
                                }
                                else
                                {
                                    ild.view_angle.y += one_degree;
                                    ui_message = "+ spin around Y";
                                }
                                break;                    
                            //----------------------------------------------------------------
                            case '3':
                            //case '#':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.view_angle.z -= one_degree;
                                    ui_message = "- spin around Z";
                                }
                                else
                                {
                                    ild.view_angle.z += one_degree;
                                    ui_message = "+ spin around Z";
                                }
                                break;                    
                            //----------------------------------------------------------------
                            case '4':
                            //case '$':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.view_angle -= one_degree;
                                    ui_message = "- spin around XYZ";
                                }
                                else
                                {
                                    ild.view_angle += one_degree;
                                    ui_message = "+ spin around XYZ";
                                }
                                break;                    
                                //----------------------------------------------------------------
                            case '5':
                            //case '%':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.view_angle.x = 0;
                                    ild.view_angle.y = pi;
                                    ild.view_angle.z = 0;
                                    ui_message = "back";
                                }
                                else
                                {
                                    ild.view_angle   = 0;
                                    ui_message = "front";
                                }
                                break;                    
                            //----------------------------------------------------------------
                            case '6':
                            //case '^':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.view_angle.x = 0;
                                    ild.view_angle.y = -half_pi;
                                    ild.view_angle.z = 0;
                                    ui_message = "right side";
                                }
                                else
                                {
                                    ild.view_angle.x = 0;
                                    ild.view_angle.y = half_pi;
                                    ild.view_angle.z = 0;
                                    ui_message = "left side";
                                }
                                break;                    
                            //----------------------------------------------------------------
                            case '7':
                            //case '&':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.view_angle.x = -half_pi;
                                    ild.view_angle.y = 0;
                                    ild.view_angle.z = 0;
                                    ui_message = "bottom";
                                }
                                else
                                {
                                    ild.view_angle.x = half_pi;
                                    ild.view_angle.y = 0;
                                    ild.view_angle.z = 0;
                                    ui_message = "top";
                                }
                                break;                    
/*
                            //----------------------------------------------------------------
                            case '8':
                            //case '*':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.view_angle -= one_degree;
                                    if(    ild.loop_animation
                                        || (ild.frame_index > 0)
                                      )
                                        --ild;
                                    ui_message = "- show & spin";
                                }
                                else
                                {
                                    ild.view_angle += one_degree;
                                    if(    ild.loop_animation
                                        || (ild.frame_index < ild.number_of_frames() - 1)
                                      )
                                        ++ild;
                                    ui_message = "+ show & spin";
                                }
                                break;
*/                                
                            //----------------------------------------------------
                            case '[':
                            //case '{':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.current_frame().deccrement_point_select_start();
                                    ui_message = "- start select";
                                }
                                else
                                {
                                    ild.current_frame().increment_point_select_start();
                                    ui_message = "+ start select";
                                }
                                break;                    
                            //----------------------------------------------------
                            case ']':
                            //case '}':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.current_frame().deccrement_point_select_end();
                                    ui_message = "- end select";
                                }
                                else
                                {
                                    ild.current_frame().increment_point_select_end();
                                    ui_message = "+ end select";
                                }
                                break;                    
                            //----------------------------------------------------
                            case'\\':
                            //case '|':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.current_frame().point_select_start = 0;
                                    ui_message = "first vertex select";
                                }
                                else
                                {
                                    ild.current_frame().point_select_end = (ild.current_frame().number_of_points() - 1);
                                    ui_message = "last vertex select";
                                }
                                break;                    
                            //----------------------------------------------------
                            case 'u':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.current_frame().previous_segment_select();
                                    ui_message = "- lit segment";
                                }
                                else
                                {
                                    ild.current_frame().next_segment_select();
                                    ui_message = "+ lit segment";
                                }
                                break;                    
                            //----------------------------------------------------
                            case 'i':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.current_frame().previous_segment_select_start();
                                    ui_message = "- start lit segment";
                                }
                                else
                                {
                                    ild.current_frame().next_segment_select_start();
                                    ui_message = "+ start lit segment";
                                }
                                break;                    
                            //----------------------------------------------------
                            case 'o':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.current_frame().previous_segment_select_end();
                                    ui_message = "- end lit segment";
                                }
                                else
                                {
                                    ild.current_frame().next_segment_select_end();
                                    ui_message = "+ end lit segment";
                                }
                                break;                    
                                
                            //----------------------------------------------------
                            case '/': // show frame set stats
                            //case '?': // show frame stats
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                    show_frame_stats = !show_frame_stats;
                                else
                                    show_frame_set_stats = !show_frame_set_stats;
                                ui_message = "";
                                break;                    
                            //----------------------------------------------------
                            case 'c':
                                ild.current_frame() += ild.current_frame().selected_segment();
                                sprintf(formatted_string, "copied %d points to end of frame", ild.current_frame().selected_segment().size());
                                ui_message = formatted_string;
                                clear_string(formatted_string);
                                break;
                            //----------------------------------------------------                            
                            case 'j':
                                ild.arbitrary_point.x = ild.display_prompt_and_echo_float("set value of arbitrary point X");
                                ui_message = "point X set";
                                break;
                            //----------------------------------------------------            
                            case 'k':
                                ild.arbitrary_point.y = ild.display_prompt_and_echo_float("set value of arbitrary point Y");
                                ui_message = "point Y set";
                                break;
                            //----------------------------------------------------            
                            case 'l':
                                ild.arbitrary_point.z = ild.display_prompt_and_echo_float("set value of arbitrary point Z");
                                ui_message = "point Z set";
                                break;
                            //----------------------------------------------------            
                            case 'q':
                                offset = ild.display_prompt_and_echo_int("set value of selected center X");
                                ild.current_frame().move_selection(ez_3D_float(   offset - ild.current_frame().selected_segment().find_center().x
                                                                                , 0
                                                                                , 0
                                                                              )
                                                                  );
                                break;
                            //----------------------------------------------------            
                            case 'w':
                                offset = ild.display_prompt_and_echo_int("set value of selected center Y");
                                ild.current_frame().move_selection(ez_3D_float(   0
                                                                                , offset - ild.current_frame().selected_segment().find_center().y
                                                                                , 0
                                                                              )
                                                                  );
                                break;
                            //----------------------------------------------------            
                            case 'e':
                                offset = ild.display_prompt_and_echo_int("set value of selected center Z");
                                ild.current_frame().move_selection(ez_3D_float(   0
                                                                                , 0
                                                                                , offset - ild.current_frame().selected_segment().find_center().z
                                                                              )
                                                                  );
                                break;
                            //----------------------------------------------------
                            case 'x':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.current_frame().move_selection(ez_3D_float(-ild.move_points_per_key, 0, 0));
                                    ui_message = "- X move";
                                }
                                else
                                {
                                    ild.current_frame().move_selection(ez_3D_float(ild.move_points_per_key, 0, 0));
                                    ui_message = "+ X move";
                                }
                                break;                    
                            //----------------------------------------------------
                            case 'y':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.current_frame().move_selection(ez_3D_float(0, -ild.move_points_per_key, 0));
                                    ui_message = "- Y move";
                                }
                                else
                                {
                                    ild.current_frame().move_selection(ez_3D_float(0, ild.move_points_per_key, 0));
                                    ui_message = "+ Y move";
                                }
                                break;                    
                            //----------------------------------------------------
                            case 'z':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.current_frame().move_selection(ez_3D_float(0, 0,-ild.move_points_per_key));
                                    ui_message = "- Z move";
                                }
                                else
                                {
                                    ild.current_frame().move_selection(ez_3D_float(0, 0, ild.move_points_per_key));
                                    ui_message = "+ Z move";
                                }
                                break;                    
                            //----------------------------------------------------
                            case 'f':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.current_frame().rotate_selection(ez_3D_float(0, 0, -one_degree));
                                    ui_message = "- spin Z";
                                }
                                else
                                {
                                    ild.current_frame().rotate_selection(ez_3D_float(0, 0, one_degree));
                                    ui_message = "+ spin Z";
                                }
                                break;                    
                            //----------------------------------------------------
                            case 'g':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.current_frame().rotate_selection(ez_3D_float(0,-one_degree, 0));
                                    ui_message = "- spin y";
                                }
                                else
                                {
                                    ild.current_frame().rotate_selection(ez_3D_float(0, one_degree, 0));
                                    ui_message = "+ spin y";
                                }
                                break;                    
                            //----------------------------------------------------
                            case 'r':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.current_frame().rotate_selection_around_arbitrary_point(ez_3D_float(0, 0, -one_degree));
                                    ui_message = "- spin Z around aribitrary point";
                                }
                                else
                                {
                                    ild.current_frame().rotate_selection_around_arbitrary_point(ez_3D_float(0, 0, one_degree));
                                    ui_message = "+ spin Z around aribitrary point";
                                }
                                break;                    
                            //----------------------------------------------------
                            case 't':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.current_frame().rotate_selection_around_arbitrary_point(ez_3D_float(0, -one_degree, 0));
                                    ui_message = "- spin Y around aribitrary point";
                                }
                                else
                                {
                                    ild.current_frame().rotate_selection_around_arbitrary_point(ez_3D_float(0, one_degree, 0));
                                    ui_message = "+ spin Y around aribitrary point";
                                }
                                break;                    
                            //----------------------------------------------------
                            case '8':
                            //case '*':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.current_frame().scale_selection(ez_3D_float(0.99, 1, 1));
                                    ui_message = "- size X";
                                }
                                else
                                {
                                    ild.current_frame().scale_selection(ez_3D_float(1.01, 1, 1));
                                    ui_message = "+ size X";
                                }
                                break;                    
                            //----------------------------------------------------
                            case '9':
                            //se '(':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.current_frame().scale_selection(ez_3D_float(1, 0.99, 1));
                                    ui_message = "- size Y";
                                }
                                else
                                {
                                    ild.current_frame().scale_selection(ez_3D_float(1, 1.01, 1));
                                    ui_message = "+ size Y";
                                }
                                break;                    
                            //----------------------------------------------------
                            case '0':
                            //case ')':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.current_frame().scale_selection(ez_3D_float(1, 1, 0.99));
                                    ui_message = "- size Z";
                                }
                                else
                                {
                                    ild.current_frame().scale_selection(ez_3D_float(1, 1, 1.01));
                                    ui_message = "+ size Z";
                                }
                                break;                    
                            //----------------------------------------------------
                            case '-':
                            //case '_':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.current_frame().scale_selection(ez_3D_float(0.99, 0.99, 0.99));
                                    ui_message = "- size XYZ";
                                }
                                else
                                {
                                    ild.current_frame().scale_selection(ez_3D_float(1.01, 1.01, 1.01));
                                    ui_message = "+ size XYZ";
                                }
                                break;                    
                            //----------------------------------------------------
                            case 'v':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.current_frame().scale_selection_around_arbitrary_point(ez_3D_float(0.99, 1, 1));
                                    ui_message = "- size X per aribitrary point";
                                }
                                else
                                {
                                    ild.current_frame().scale_selection_around_arbitrary_point(ez_3D_float(1.01, 1, 1));
                                    ui_message = "+ size X per aribitrary point";
                                }
                                break;                    
                            //----------------------------------------------------
                            case 'b':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.current_frame().scale_selection_around_arbitrary_point(ez_3D_float(1, 0.99, 1));
                                    ui_message = "- size Y per aribitrary point";
                                }
                                else
                                {
                                    ild.current_frame().scale_selection_around_arbitrary_point(ez_3D_float(1, 1.01, 1));
                                    ui_message = "+ size Y per aribitrary point";
                                }
                                break;                    
                            //----------------------------------------------------
                            case 'n':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.current_frame().scale_selection_around_arbitrary_point(ez_3D_float(1, 1, 0.99));
                                    ui_message = "- size Z per aribitrary point";
                                }
                                else
                                {
                                    ild.current_frame().scale_selection_around_arbitrary_point(ez_3D_float(1, 1, 1.01));
                                    ui_message = "+ size Z per aribitrary point";
                                }
                                break;                    
                            //----------------------------------------------------
                            case 'm':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.current_frame().scale_selection_around_arbitrary_point(ez_3D_float(0.99, 0.99, 0.99));
                                    ui_message = "- size XYZ per aribitrary point";
                                }
                                else
                                {
                                    ild.current_frame().scale_selection_around_arbitrary_point(ez_3D_float(1.01, 1.01, 1.01));
                                    ui_message = "+ size XYZ per aribitrary point";
                                }
                                break;                    
                            //----------------------------------------------------
                            case 'a':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.arbitrary_point.x -= ild.move_points_per_key;
                                    ui_message = "- X arbitrary point";
                                }
                                else
                                {
                                    ild.arbitrary_point.x += ild.move_points_per_key;
                                    ui_message = "+ X arbitrary point";
                                }
                                break;                    
                            //----------------------------------------------------
                            case 's':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.arbitrary_point.y -= ild.move_points_per_key;
                                    ui_message = "- Y arbitrary point";
                                }
                                else
                                {
                                    ild.arbitrary_point.y += ild.move_points_per_key;
                                    ui_message = "+ Y arbitrary point";
                                }
                                break;                    
                            //----------------------------------------------------
                            case 'd':
                                if(event.key.keysym.mod & KMOD_SHIFT)
                                {
                                    ild.arbitrary_point.z -= ild.move_points_per_key;
                                    ui_message = "- Z arbitrary point";
                                }
                                else
                                {
                                    ild.arbitrary_point.z += ild.move_points_per_key;
                                    ui_message = "+ Z arbitrary point";
                                }
                                break;                    
                            //----------------------------------------------------
                              case SDLK_RETURN:
                                show_help = !show_help;
                            //----------------------------------------------------------------
                        } // end switch(event.key.keysym.sym)
                        ild.display(ui_message);
                        if(show_help)
                            ild.display_move_size_spin_menu();                
                        if(show_frame_stats)
                            ild.current_frame().display_stats();
                        if(show_frame_set_stats)
                            ild.display_stats();
                    } // end else if(event.type == SDL_KEYDOWN)
                } // end while(in_sub_menu)
                break;
            }   // end local scope
            //----------------------------------------------------------------








            //----------------------------------------------------------------
            case 'i': // open data from file
                ild.display();
                ild.display_in_file_type_menu();
                do
                {
                    SDL_WaitEvent(&event);
                } while(    event.type != SDL_KEYDOWN
                         && event.type != SDL_QUIT
                       );
                if(event.type == SDL_QUIT)
                    running = false;
                else
                {
                    switch(event.key.keysym.sym)
                    {
                        //----------------------------------------------------------------
                        case SDLK_ESCAPE:
                            in_sub_menu = false;
                            ui_message = "action canceled";
                            break;
                        //----------------------------------------------------------------
                        case '1': // open ILDA
                            ild.display();
                            ild.list_files(EZ_ILDA_FRAME_SHARE);
                            file_name = ild.display_prompt_and_echo_name("enter ILDA file name");
                            ild.display();
                            ild.display_in_ild_action_menu();            
                            do
                            {
                                SDL_WaitEvent(&event);
                            } while(    event.type != SDL_KEYDOWN
                                     && event.type != SDL_QUIT
                                   );
                            if(event.type == SDL_QUIT)
                                running = false;
                            else
                            {
                                switch(event.key.keysym.sym)
                                {
                                    //----------------------------------------------------------------
                                    case SDLK_ESCAPE:
                                        ui_message = "action canceled";
                                        break;
                                    //----------------------------------------
                                    case '1': // replace
                                        ild.display("opening ILDA");
                                        ild.display_please_wait();
                                        ild.from_file(file_name);
                                        ui_message = "replace";
                                        break;
                                    //----------------------------------------
                                    case '2': // prepend
                                        ild.display("opening ILDA");
                                        ild.display_please_wait();
                                        {
                                            ez_ilda_frame_set prepend;
                                            prepend.from_file(file_name);
                                            prepend += ild;
                                            ild = prepend;
                                        }                          
                                        ui_message = "prepend";
                                        break;
                                    //----------------------------------------
                                    case '3': // insert
                                        ild.display("opening ILDA");
                                        ild.display_please_wait();
                                        {
                                            int                i;
                                            ez_ilda_frame_set  insert,
                                                               result;
                                            insert.from_file(file_name);                                                    
                                            for(i = 0; i <= ild.frame_index; i++)
                                                result += ild[i];
                                            for(i = 0; i < insert.number_of_frames(); i++)
                                                result += insert[i];
                                            for(i = ild.frame_index + 1; i < ild.number_of_frames(); i++)
                                                result += ild[i];
                                            ild = result;
                                        }
                                        ild.from_file(file_name, true);
                                        ui_message = "insert";
                                        break;
                                    //----------------------------------------
                                    case '4': // append
                                        ild.display("opening ILDA");
                                        ild.display_please_wait();
                                        ild.from_file(file_name, true);
                                        ui_message = "append";
                                        break;
                                    //----------------------------------------
                                    case '5': // superimpose
                                        ild.display("opening ILDA");
                                        ild.display_please_wait();
                                        ild.superimpose_from_file(file_name);
                                        ui_message = "superimpose";
                                        break;
                                    //----------------------------------------
                                }
                            }
                            break;
                        //--------------------------------------------------------
                        case '2': // open DXF
                            ild.display();
                            ild.display_in_dxf_action_menu();
                            do
                            {
                                SDL_WaitEvent(&event);
                            } while(    event.type != SDL_KEYDOWN
                                     && event.type != SDL_QUIT
                                   );
                            if(event.type == SDL_QUIT)
                                running = false;
                            else
                            {
                                switch(event.key.keysym.sym)
                                {
                                    //----------------------------------------------------------------
                                    case SDLK_ESCAPE:
                                        ui_message = "action canceled";
                                        break;
                                    //----------------------------------------
                                    case '1': // open dxf
                                        ild.list_files(EZ_ILDA_DXF_SHARE);
                                        file_name = ild.display_prompt_and_echo_name("enter DXF file name");
                                        ild.display("opening DXF");
                                        ild.display_please_wait();
                                        ild.from_dxf_file(file_name);
                                        ui_message = "open dxf";
                                        break;
                                    //----------------------------------------
                                    case '2': // append dxf
                                        ild.list_files(EZ_ILDA_DXF_SHARE);
                                        file_name = ild.display_prompt_and_echo_name("enter DXF file name");            
                                        ild.display("opening DXF");
                                        ild.display_please_wait();
                                        ild.from_dxf_file(file_name, true);
                                        ui_message = "append dxf file";
                                        break;
                                    //----------------------------------------
                                    case '3': // open dxf directory
                                        ild.list_directories(EZ_ILDA_DXF_SHARE);
                                        file_name = ild.display_prompt_and_echo_name("enter DXF directory name");
                                        ild.display("opening DXF");
                                        ild.display_please_wait();
                                        ild.from_dxf_directory(file_name);
                                        ui_message = "open dxf directory";
                                        break;
                                    //----------------------------------------
                                    case '4': // append dxf directory
                                        ild.list_directories(EZ_ILDA_DXF_SHARE);
                                        file_name = ild.display_prompt_and_echo_name("enter DXF directory name");
                                        ild.display("opening DXF");
                                        ild.display_please_wait();
                                        ild.from_dxf_directory(file_name, true);
                                        ui_message = "open dxf directory";
                                        break;
                                    //----------------------------------------
                                }
                            }
                            break;
                        //--------------------------------------------------------
                        case '3': // open wave                                
                            ild.display();
                            ild.list_files(EZ_ILDA_WAVE_SHARE);
                            file_name = ild.display_prompt_and_echo_name("enter wave file name");
                            ild.display_in_wave_action_menu();
                            do
                            {
                                SDL_WaitEvent(&event);
                            } while(    event.type != SDL_KEYDOWN
                                     && event.type != SDL_QUIT
                                   );
                            if(event.type == SDL_QUIT)
                                running = false;
                            else
                            {
                                switch(event.key.keysym.sym)
                                {
                                    //----------------------------------------------------------------
                                    case SDLK_ESCAPE:
                                        ui_message = "action canceled";
                                        break;
                                    //----------------------------------------
                                    case '1': // open wave
                                        ild.display("opening wave");
                                        ild.display_please_wait();
                                        ild.from_wave_file(file_name);
                                        ui_message = "open wave";
                                        break;
                                    //----------------------------------------
                                    case '2': // append wave
                                        ild.display("opening wave");
                                        ild.display_please_wait();
                                        ild.from_wave_file(file_name, true);
                                        ui_message = "append wave";
                                        break;
                                //----------------------------------------
                                }
                            }
                            break;
                        //--------------------------------------------------------
                        case '4': // open bitmap
                            ild.display();
                            ild.display_in_bmp_action_menu();
                            do
                            {
                                SDL_WaitEvent(&event);
                            } while(    event.type != SDL_KEYDOWN
                                     && event.type != SDL_QUIT
                                   );
                            if(event.type == SDL_QUIT)
                                running = false;
                            else
                            {
                                switch(event.key.keysym.sym)
                                {
                                    //----------------------------------------------------------------
                                    case SDLK_ESCAPE:
                                        ui_message = "action canceled";
                                        break;
                                    //----------------------------------------
                                    case '1': //
                                        ild.list_files(EZ_ILDA_BMP_SHARE);
                                        file_name = ild.display_prompt_and_echo_name("enter bitmap file name");
                                        ild.display("loading bmp");
                                        ild.display_please_wait();
                                        ild.color_over_ezfb_bitmap(file_name);
                                        ui_message = "color over bmp";
                                        break;
                                    //----------------------------------------
                                    case '2': //
                                        ild.list_files(EZ_ILDA_BMP_SHARE);
                                        file_name = ild.display_prompt_and_echo_name("enter bitmap file name");
                                        ild.display("loading bmp");
                                        ild.display_please_wait();
                                        ild.load_background_bitmap(file_name);
                                        ui_message = "background bmp";
                                        break;
                                    //----------------------------------------
                                    case '3': //
                                        ild.clear_background_bitmap();
                                        ui_message = "clear background";
                                        break;
                                //----------------------------------------
                                }
                            }                                
                            break;
                        //--------------------------------------------------------
                        case '5': // ascii palette
                            ild.list_files(EZ_ILDA_PALETTE_SHARE);
                            file_name = ild.display_prompt_and_echo_name("enter Palette file name");
                            ild.display("opening palette");
                            ild.display_please_wait();
                            ild.add_palette_from_file(file_name);
                            ui_message = "open palette";
                            break;
                        //--------------------------------------------------------
                    } // end switch(event.key.keysym.sym)
                    break;
                } // end else of if(event.type == SDL_QUIT)
                //----------------------------------------------------------------




            //----------------------------------------------------------------
            case 'o': // save contents of memory to various file types
                ild.display("save as file");
                ild.display_out_file_type_menu();
                do
                {
                    SDL_WaitEvent(&event);
                } while(    event.type != SDL_KEYDOWN
                         && event.type != SDL_QUIT
                       );
                if(event.type == SDL_QUIT)
                    running = false;
                else
                {
                    switch(event.key.keysym.sym)
                    {
                        //----------------------------------------------------------------
                        case SDLK_ESCAPE:
                            in_sub_menu = false;
                            ui_message = "action canceled";
                            break;
                        //----------------------------------------------------------------
                        case '1':   
                            file_name = ild.display_prompt_and_echo_name("enter new ILDA file name");
                            ild.display("saving ILDA");
                            ild.display_please_wait();
                            ild.current_frame().save_as_ilda(file_name);
                            ui_message = "ILDA saved";
                            break;
                        //--------------------------------------------------------
                        case '2':   
                            file_name = ild.display_prompt_and_echo_name("enter new ILDA file name");
                            ild.display("saving ILDA");
                            ild.display_please_wait();
                            ild.selected_frames().save_as_ilda(file_name);
                            ui_message = "ILDA saved";
                            break;
                        //--------------------------------------------------------
                        case '3':   
                            file_name = ild.display_prompt_and_echo_name("enter new ILDA file name");
                            ild.display("saving ILDA");
                            ild.display_please_wait();
                            ild.save_as_ilda(file_name);
                            ui_message = "ILDA saved";
                            break;
                        //--------------------------------------------------------
                        case '4':   
                            ild.display_frame_effects();
                            file_name = ild.display_prompt_and_echo_name("enter frame effect name");
                            ild.display("saving ILDA");
                            ild.display_please_wait();
                            ild.call_frame_effect(file_name);
                            ui_message = "ILDA saved";
                            break;
                        //--------------------------------------------------------
                        case '5':   
                            ild.display_frame_set_effects();
                            file_name = ild.display_prompt_and_echo_name("enter frame set effect name");
                            ild.display("saving ILDA");
                            ild.display_please_wait();
                            ild.call_frame_set_effect(file_name);
                            ui_message = "ILDA saved";
                            break;
                        //--------------------------------------------------------
                        case '6':   
                            file_name = ild.display_prompt_and_echo_name("enter new DXF file name");
                            ild.display("saving DXF");
                            ild.display_please_wait();
                            ild.current_frame().save_as_dxf(file_name);
                            ui_message = "DXF saved";
                            break;
                        //--------------------------------------------------------
                        case '7':
                            file_name = ild.display_prompt_and_echo_name("enter new DXF directory name");
                            ild.display("saving DXF");
                            ild.display_please_wait();
                            ild.selected_frames().save_as_dxf_directory(file_name);
                            ui_message = "DXF directory saved";
                            break;
                        //--------------------------------------------------------
                        case '8':
                            file_name = ild.display_prompt_and_echo_name("enter new DXF directory name");
                            ild.display("saving DXF");
                            ild.display_please_wait();
                            ild.save_as_dxf_directory(file_name);
                            ui_message = "DXF directory saved";
                            break;
                        //--------------------------------------------------------
                        case '9':
                            file_name = ild.display_prompt_and_echo_name("enter new wave file name");
                            ild.display("saving wave");
                            ild.display_please_wait();
                            ild.current_frame().save_as_wave(file_name);
                            ui_message = "wave saved";
                            break;
                        //--------------------------------------------------------
                        case '0':
                            file_name = ild.display_prompt_and_echo_name("enter new wave file name");
                            ild.display("saving wave");
                            ild.display_please_wait();
                            ild.save_as_wave(file_name);
                            ui_message = "wave saved";
                            break;
                        //--------------------------------------------------------
                        case 'a':   
                            file_name = ild.display_prompt_and_echo_name("enter new C++ file name");
                            ild.display("saving C++");
                            ild.display_please_wait();
                            ild.current_frame().save_as_code(file_name);
                            ui_message = "C++ saved";
                            break;
                        //--------------------------------------------------------
                        case 'b':   
                            file_name = ild.display_prompt_and_echo_name("enter new C++ file name");
                            ild.display("saving C++");
                            ild.display_please_wait();
                            ild.selected_frames().save_as_code(file_name);
                            ui_message = "C++ saved";
                            break;
                        //--------------------------------------------------------
                        case 'c':
                            file_name = ild.display_prompt_and_echo_name("enter new C++ file name");
                            ild.display("saving C++");
                            ild.display_please_wait();
                            ild.save_as_code(file_name);
                            ui_message = "C++ saved";
                            break;
                        //--------------------------------------------------------
                        case 'd':
                            file_name = ild.display_prompt_and_echo_name("enter new bitmap file name");
                            ild.display("saving bitmap");
                            ild.display_please_wait();
                            ild.current_frame().save_as_bmp(file_name);
                            ui_message = "bitmap saved";
                            break;
                        //--------------------------------------------------------
                        case 'e':
                            file_name = ild.display_prompt_and_echo_name("enter new bitmap directory file name");
                            ild.display("saving bitmap directory");
                            ild.display_please_wait();
                            ild.selected_frames().save_as_bmp_directory(file_name);
                            ui_message = "bitmap directory saved";
                            break;
                        //--------------------------------------------------------
                        case 'f':
                            file_name = ild.display_prompt_and_echo_name("enter new bitmap directory file name");
                            ild.display("saving bitmap directory");
                            ild.display_please_wait();
                            ild.save_as_bmp_directory(file_name);
                            ui_message = "bitmap directory saved";
                            break;
                        //--------------------------------------------------------
                        case 'g':
                            file_name = ild.display_prompt_and_echo_name("enter new palette file name");
                            ild.display("saving ascii palette");
                            ild.display_please_wait();
                            ild.palette_picker(ild.current_frame().palette_index).save_as_ascii(file_name);
                            ui_message = "ascii palette saved";
                            break;
                        //--------------------------------------------------------
                        case 'h':
                            file_name = ild.display_prompt_and_echo_name("enter new palette file name");
                            ild.display("saving ascii palette");
                            ild.display_please_wait();
                            ild.palette_picker(ild.target_palette_index).save_as_ascii(file_name);
                            ui_message = "ascii palette saved";
                            break;
                        //--------------------------------------------------------
                    } // end switch(event.key.keysym.sym)
                    break;
                } // end else of if(event.type == SDL_QUIT)
                //----------------------------------------------------------------






            //----------------------------------------------------------------
            case SDLK_RETURN:
                show_help = !show_help;
                if(!show_help)
                    ui_message = "[Enter] help";
                break;
            //----------------------------------------------------------------
            default:
                ui_message = "[Enter] help";
                break;
            //----------------------------------------------------------------
        } // end switch(event.key.keysym.sym)
        if(show_help)
        {
            ui_message = "";
            ild.display();
            ild.display_program_menu();
        }
        else 
            ild.display(ui_message);
        if(show_frame_stats)
            ild.current_frame().display_stats();
        if(show_frame_set_stats)
            ild.display_stats();
        if(teminate_requested)
        {
            ild.display("terminating");
            if(ild.display_prompt_and_echo_bool("Do you really want to terminate laserboy? y|Y"))
                running = false;
            else
            {
                teminate_requested = false;
                ui_message = "terminate canceled";
                ild.display(ui_message);
            }
        }
    } // end while(running)
    ild.display("terminating");
    if(!ild.display_prompt_and_echo_bool("Exit without saving OUT.ild? y|Y"))
    {   
        ild.display("saving OUT.ild");
        if(ild.auto_optimize)
            ild.minimize_points();
        ild.save_as_ilda("OUT.ild");
    }
    //----------------------------------------------------------------------------
    ild.display("LaserBoy terminated");
    //----------------------------------------------------------------------------    
    SDL_Delay(800);
    return(0);
}

//############################################################################
//////////////////////////////////////////////////////////////////////////////
//############################################################################
